Package: BiocPkgTools
Type: Package
Title: Collection of simple tools for learning about Bioconductor
        Packages
Version: 1.30.0
Date: 2026-03-13
Authors@R: c( 
    person("Shian", "Su", role=c("aut", "ctb"), email = "su.s@wehi.edu.au"),
    person("Lori", "Shepherd", role="ctb", email = "Lori.Shepherd@roswellpark.org"),
    person("Marcel", "Ramos", , "marcel.ramos@sph.cuny.edu", c("aut", "ctb"),
      c(ORCID = "0000-0002-3242-0582")),
    person("Felix G.M.", "Ernst", role="ctb", email = "felix.gm.ernst@outlook.com"),
    person("Jennifer","Wokaty", role="ctb", email="jennifer.wokaty@gmail.com"),
    person("Charlotte", "Soneson", role="ctb", email = "charlottesoneson@gmail.com"),
    person("Martin", "Morgan", role="ctb", email = "martin.morgan@roswellpark.org"),
    person("Vince", "Carey", role="ctb", email="stvjc@channing.harvard.edu"),
    person("Sean", "Davis", email = "seandavi@gmail.com",role = c("aut", "cre"))
    )
Description: Bioconductor has a rich ecosystem of metadata around
        packages, usage, and build status. This package is a simple
        collection of functions to access that metadata from R. The
        goal is to expose metadata for data mining and value-added
        functionality such as package searching, text mining, and
        analytics on packages.
Depends: htmlwidgets, R (>= 4.1.0)
Imports: BiocFileCache, BiocManager, biocViews, tibble, methods, rlang,
        stringr, stats, rvest, dplyr, xml2, readr, httr, httr2,
        htmltools, DT, tools, utils, igraph (>= 2.0.0), jsonlite, gh,
        RBGL, graph, curl, glue, lubridate, purrr, tidyr, yaml
VignetteBuilder: knitr
Suggests: BiocStyle, knitr, rmarkdown, testthat, tm, networkD3,
        visNetwork, clipr, blastula, kableExtra, DiagrammeR,
        SummarizedExperiment
License: MIT + file LICENSE
BugReports: https://github.com/seandavi/BiocPkgTools/issues/new
URL: https://github.com/seandavi/BiocPkgTools
Encoding: UTF-8
RoxygenNote: 7.3.3
Roxygen: list(markdown = TRUE)
SystemRequirements: mailsend-go
biocViews: Software, Infrastructure
Config/pak/sysreqs: cmake git libglpk-dev make libicu-dev libuv1-dev
        libxml2-dev libssl-dev libx11-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:48:41 UTC
RemoteUrl: https://github.com/bioc/BiocPkgTools
RemoteRef: RELEASE_3_23
RemoteSha: 9b3d701c47cee85923a1d8e60bfe12ca1cf83453
NeedsCompilation: no
Packaged: 2026-04-29 16:24:25 UTC; root
Author: Shian Su [aut, ctb],
  Lori Shepherd [ctb],
  Marcel Ramos [aut, ctb] (ORCID:
    <https://orcid.org/0000-0002-3242-0582>),
  Felix G.M. Ernst [ctb],
  Jennifer Wokaty [ctb],
  Charlotte Soneson [ctb],
  Martin Morgan [ctb],
  Vince Carey [ctb],
  Sean Davis [aut, cre]
Maintainer: Sean Davis <seandavi@gmail.com>
Built: R 4.6.0; ; 2026-04-29 16:26:41 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,repositoryStats)
S3method(summary,biocrevdeps)
export(CRANstatus)
export(activitySince)
export(anacondaDownloadStats)
export(biocBuildEmail)
export(biocBuildReport)
export(biocBuildReportDB)
export(biocBuildStatusDB)
export(biocDownloadStats)
export(biocExplore)
export(biocMaintained)
export(biocPkgList)
export(biocPkgRanges)
export(biocRevDepEmail)
export(biocVIEWSdb)
export(buildClassDepData)
export(buildClassDepFromPackage)
export(buildClassDepGraph)
export(buildPkgDependencyDataFrame)
export(buildPkgDependencyIgraph)
export(firstInBioc)
export(getBiocVignette)
export(getPackageInfo)
export(getPkgYearsInBioc)
export(get_bioc_data)
export(get_cre_orcids)
export(githubDetails)
export(githubURLParts)
export(hasBiocMaint)
export(inducedSubgraphByPkgs)
export(latestPkgStats)
export(orcid_table)
export(pkgBiocDeps)
export(pkgBiocRevDeps)
export(pkgCombDependencyGain)
export(pkgDepImports)
export(pkgDepMetrics)
export(pkgDownloadRank)
export(pkgDownloadStats)
export(pkgToolsCache)
export(plotClassDep)
export(plotClassDepData)
export(plotClassDepGraph)
export(problemPage)
export(repositoryStats)
export(sentHistory)
export(setCache)
export(subgraphByDegree)
export(templatePath)
import(biocViews)
import(tools)
import(utils)
import(xml2)
importFrom(BiocFileCache,BiocFileCache)
importFrom(BiocFileCache,bfcdownload)
importFrom(BiocFileCache,bfcneedsupdate)
importFrom(BiocFileCache,bfcnew)
importFrom(BiocFileCache,bfcquery)
importFrom(BiocFileCache,bfcrpath)
importFrom(BiocFileCache,bfcupdate)
importFrom(BiocManager,containerRepository)
importFrom(BiocManager,repositories)
importFrom(BiocManager,version)
importFrom(DT,datatable)
importFrom(RBGL,transitive.closure)
importFrom(curl,curl)
importFrom(curl,has_internet)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,collect)
importFrom(dplyr,desc)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,top_n)
importFrom(gh,gh)
importFrom(glue,glue)
importFrom(grDevices,extendrange)
importFrom(graph,edges)
importFrom(graph,nodes)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(htmltools,a)
importFrom(htmlwidgets,createWidget)
importFrom(httr,HEAD)
importFrom(httr,POST)
importFrom(httr,PUT)
importFrom(httr,VERB)
importFrom(httr,accept)
importFrom(httr,add_headers)
importFrom(httr,authenticate)
importFrom(httr,content)
importFrom(httr,content_type)
importFrom(httr,headers)
importFrom(httr,message_for_status)
importFrom(httr,status_code)
importFrom(httr,timeout)
importFrom(httr2,oauth_client)
importFrom(httr2,oauth_flow_client_credentials)
importFrom(httr2,obfuscated)
importFrom(httr2,req_auth_bearer_token)
importFrom(httr2,req_body_form)
importFrom(httr2,req_headers)
importFrom(httr2,req_method)
importFrom(httr2,req_options)
importFrom(httr2,req_perform)
importFrom(httr2,req_template)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(igraph,"E<-")
importFrom(igraph,"V<-")
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,degree)
importFrom(igraph,delete_edges)
importFrom(igraph,delete_vertices)
importFrom(igraph,distances)
importFrom(igraph,edge_attr)
importFrom(igraph,ego)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,induced_subgraph)
importFrom(igraph,is_igraph)
importFrom(igraph,layout_with_kk)
importFrom(igraph,subcomponent)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(lubridate,mdy)
importFrom(methods,getClass)
importFrom(methods,is)
importFrom(purrr,list_rbind)
importFrom(purrr,map2)
importFrom(readr,read_lines)
importFrom(rlang,is_call)
importFrom(rlang,is_pairlist)
importFrom(rlang,is_symbol)
importFrom(rlang,is_syntactic_literal)
importFrom(rvest,html_nodes)
importFrom(rvest,html_table)
importFrom(rvest,html_text)
importFrom(rvest,html_text2)
importFrom(stats,setNames)
importFrom(stringr,str_match)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_split)
importFrom(stringr,str_squish)
importFrom(stringr,str_to_upper)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,tibble)
importFrom(tidyr,unnest_longer)
importFrom(utils,available.packages)
importFrom(utils,combn)
importFrom(utils,download.file)
importFrom(utils,read.table)
importFrom(xml2,read_html)
importFrom(yaml,read_yaml)

# BiocPkgTools

<!-- badges: start -->
[![R-CMD-check](https://github.com/seandavi/BiocPkgTools/actions/workflows/R-CMD-check.yaml/badge.svg?branch=RELEASE_3_21)](https://github.com/seandavi/BiocPkgTools/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

Bioconductor has a rich ecosystem of metadata around packages, usage, and build status. 
This package is a simple collection of functions to access that metadata from R in a tidy data format. 
The goal is to expose metadata for data mining and value-added functionality such as package searching, text mining, and analytics on packages. 

Functionality includes access to ***computable versions of***:

- Download statistics
- Detailed package information
- Package dependendencies (and reverse dependencies)
- Package BiocViews categories
- Build reports
- Vignettes (including examples of text mining)

library(BiocPkgTools)
pl = biocPkgList()
testing = TRUE # change to FALSE to actually update/create records
z = mapply(
    FUN = BiocPkgTools:::generateBiocPkgDOI,
    pl$Package,
    pl$Author,
    2017,
    testing=testing
)

#' Extract GitHub user and repo name from GitHub URL
#'
#' @param urls `character()` A vector of URLs
#'
#' @return A `data.frame` with four columns:
#'   * url: The original GitHub URL
#'   * user_repo: The GitHub "username/repo", combined
#'   * user:  The GitHub username
#'   * repo: The GitHub repo name
#'
#' @importFrom stringr str_match
#'
#' @examples
#' # find GitHub URL details for
#' # Bioconductor packages
#' bpkgl = biocPkgList()
#' urldetails = githubURLParts(bpkgl$URL)
#' urldetails = urldetails[!is.na(urldetails$url),]
#' head(urldetails)
#'
#' @export
githubURLParts <- function(urls) {
    tmp = str_match(urls,'.*http[s]?://github.com/(([^/]+)/([^/]+)).*')
    tmp = data.frame(tmp, stringsAsFactors = FALSE)
    colnames(tmp) = c('url', 'user_repo', 'user', 'repo')
    ## remove ".git" from tail of GitHub repo
    tmp$user_repo = sub('\\.git$','', tmp$user_repo)
    tmp$repo = sub('\\.git$','', tmp$repo)
    tmp
}


.gh_pkg_info <- function(user_repo) {
    f = tryCatch(gh(sprintf('/repos/%s', user_repo)),
                 error=function(e) {
                     warning(sprintf("package %s not found", user_repo))
                     NA
                 })
    f[is.null(f)]=""
    f
}


#' Get package details from GitHub
#'
#' For packages that live on GitHub, we can mine
#' further details. This function returns the
#' GitHub details for the listed packages.
#'
#' @details
#' The [gh::gh()] function is used to
#' do the fetching. If the number of packages supplied
#' to this function is large (>40 or so), it is possible
#' to run into problems with API rate limits. The `gh`
#' package uses the environment variable "GITHUB_PAT"
#' (for personal access token) to authenticate and then
#' provide higher rate limits. If you run into problems
#' with rate limits, set sleep to some small positive
#' number to slow queries. Alternatively, create a Personal
#' Access Token on GitHub and register it. See the `gh`
#' package for details.
#'
#' @importFrom gh gh
#'
#' @param pkgs a `character()` vector of username/repo
#' for one or more GitHub repos, such as `seandavi/GEOquery`.
#'
#' @param sleep `numeric()` denoting the number of seconds to
#' sleep between GitHub API calls. Since GitHub rate limits
#' its APIs, it might be necessary to either use small
#' chunks of packages iteratively or to supply a non-zero
#' argument here. See the `details` section for a better
#' solution using GitHub tokens.
#'
#' @examples
#' pkglist = biocPkgList()
#'
#' # example of "pkgs" format.
#' head(pkglist$URL)
#'
#' gh_list = githubURLParts(pkglist$URL)
#' gh_list = gh_list[!is.null(gh_list$user_repo),]
#'
#' head(gh_list$user_repo)
#'
#' ghd = githubDetails(gh_list$user_repo[1:5])
#' lapply(ghd, '[[', "stargazers")
#'
#' @export
githubDetails <- function(pkgs, sleep=0) {
    vals = lapply(pkgs, function(pkg) {
        Sys.sleep(sleep)
        .gh_pkg_info(pkg)
    })
    names(vals) = pkgs
    vals = vals[!is.na(vals)]
    vals
}

context("githubURLParts")
library(BiocPkgTools)

urls = c("https://github.com/kevinrue/TVTB",
         "https://github.com/ronammar/zFPKM/",
         "https://github.com/ronammar/zFPKM/extrastuff/abc")
ghp = githubURLParts(urls)


test_that("returns data frame and columns correctly", {
    expect_equal(nrow(ghp), 3)
    expect_equal(ncol(ghp), 4)
})

test_that("handles extra stuff in URL appropriately", {
    expect_equivalent(ghp[3, 4, drop=TRUE], 'zFPKM')
    expect_equivalent(ghp[2, 4, drop=TRUE], 'zFPKM')
})

utils::globalVariables(
    c(".data", "biocViewsVocab", "Nb_of_distinct_IPs", "Nb_of_downloads")
)

.cache_read <- function(uri) {
    ufile <- .cache_url_file(uri)
    read.table(ufile, header = TRUE)
}

.repos_short_names <- function() {
    data.frame(
        code.names = c(
            "software", "data-experiment", "workflows", "data-annotation",
            "books"
        ),
        stat.url = c(
            "bioc", "data-experiment", "workflows", "data-annotation",
            NA_character_
        ),
        stat.file = c(
            "bioc", "experiment", "workflows", "annotation", NA_character_
        ),
        pkg.url = c(
            "bioc", "data/experiment", "workflows", "data/annotation",
            NA_character_
        ),
        biocmanager.names = c(
            "BioCsoft", "BioCexp", "BioCworkflows", "BioCann", "BioCbooks"
        )
    )
}

.get_all_biocpkgs <- function() {
    db <- available.packages(
        repos = BiocManager:::.repositories_bioc(version = version)
    )
    rownames(db)
}

#' @importFrom httr2 request req_options req_perform
.filter_http_error <- function(urls) {
    check_url <- function(url) {
        res <- try({
            request(url) |>
                req_options(followlocation = FALSE) |>
                req_perform()
        }, silent = TRUE)
        !inherits(res, "try-error")
    }
    is_valid <- vapply(urls, check_url, logical(1L))

    if (any(!is_valid)) {
        inv_urls <- paste(names(urls)[!is_valid], collapse = ", ")
        warning(
            "Download stats for these resources are currently down:\n ",
            inv_urls,
            call. = FALSE
        )
    }
    urls[is_valid]
}

.STATS_BASE_URL <- "https://bioconductor.org/packages/stats/"

#' Get Bioconductor download statistics
#'
#' @details Note that Bioconductor package download
#' stats are not version-specific.
#'
#' @param pkgType `character()` All, some, or one of 'software',
#'   'data-experiment', 'workflows', or 'data-annotation' (defaults to all
#'   types)
#'
#' @importFrom dplyr mutate
#' @importFrom utils read.table
#' @importFrom tibble as_tibble
#' @importFrom BiocFileCache BiocFileCache bfcupdate bfcneedsupdate bfcrpath
#'     bfcquery bfcnew bfcdownload
#'
#' @return A `tibble` of download statistics for all Bioconductor packages
#'
#' @examples
#' biocDownloadStats()
#' @export
biocDownloadStats <-
    function(
        pkgType = c(
            "software", "data-experiment",
            "workflows", "data-annotation"
        )
    )
{
    if (identical(pkgType, "all"))
        .Defunct(
            msg = "Value 'all' is defunct as 'pkgType' option"
        )
    pkgType <- match.arg(pkgType, several.ok = TRUE)
    linkPkg <- .matchGetShortName(pkgType, "stat.url")
    fnameType <- .matchGetShortName(pkgType, "stat.file")
    repo_name <- .matchGetShortName(pkgType, "biocmanager.names")
    pkgs <- biocPkgList(repo = repo_name)[["Package"]]

    base_url <- paste0(.STATS_BASE_URL, "%s/%s_pkg_stats.tab")

    stats_urls <- sprintf(base_url,
        linkPkg, fnameType
    )
    names(stats_urls) <- pkgType
    stats_urls <- .filter_http_error(stats_urls)

    tlist <- lapply(stats_urls, .cache_read)
    tbl <- as_tibble(
        cbind(
            pkgType = rep(names(stats_urls), vapply(tlist, nrow, numeric(1L))),
            dplyr::bind_rows(tlist)
        )
    )

    tbl <- filter(tbl, .data$Month != "all") |>
        dplyr::mutate(
            Date = as.Date(
                paste(.data$Year, .data$Month, '01'), '%Y %b %d'
            )
        )

    tbl <- tbl[tbl$Package %in% pkgs, ]

    class(tbl) <- c('bioc_downloads', class(tbl))
    tbl
}

.try.read.table <- function(...) {
    withCallingHandlers({
        tryCatch({
            read.table(...)
        }, error = function(err) {
            data.frame(
                Year = integer(0L), Month = character(0L),
                Nb_of_distinct_IPs = integer(0L), Nb_of_downloads = integer(0L)
            )
        })
    }, warning = function(w) {
        w <- conditionMessage(w)
        if (grepl("404 Not Found", w, fixed = TRUE))
            warning(
                "No data for year: ", gsub(".*_([0-9]{4})_.*", "\\1", w),
                call. = FALSE
            )
        invokeRestart("muffleWarning")
    })
}

#' Get Bioconductor download statistics for a package
#'
#' @param pkg `character(1)` The name of a Bioconductor package
#'
#' @param pkgType `character(1)` One of 'software', 'data-experiment',
#'     'workflows', or 'data-annotation' (defaults to 'software')
#'
#' @param years numeric(), `character()` A vector of years from which to
#'     obtain download statistics (defaults to current year)
#'
#' @return A `tibble` of download statistics
#'
#' @examples
#' pkgDownloadStats("GenomicRanges")
#' @export
pkgDownloadStats <-
    function(
        pkg,
        pkgType = c(
            "software", "data-experiment", "workflows", "data-annotation"
        ),
        years = format(Sys.time(), "%Y")
    )
{

    pkgType <- match.arg(pkgType)
    pkgType <- .matchGetShortName(pkgType, "stat.url")

    base_url <- paste0(.STATS_BASE_URL, "%s/%s/%s_%s_stats.tab")

    stats_urls <- sprintf(base_url,
        pkgType, pkg, pkg, years
    )

    tlist <- lapply(stats_urls, .try.read.table, header = TRUE)
    tbl <- as_tibble(dplyr::bind_rows(tlist))
    filter(tbl,
        .data$Month != "all" &
            (Nb_of_distinct_IPs != 0 & Nb_of_downloads != 0)
    )

}

#' When did a package enter Bioconductor?
#'
#' This function uses the biocDownloadStats
#' data to *approximate* when a package entered
#' Bioconductor. Note that the download stats
#' go back only to 2009.
#'
#' @importFrom dplyr filter group_by top_n collect
#'
#' @param download_stats a data.frame from [biocDownloadStats()]
#'
#' @examples
#' dls <- biocDownloadStats()
#' tail(firstInBioc(dls))
#' @export
firstInBioc <- function(download_stats) {
  download_stats |>
    dplyr::filter(.data$Month!='all') |>
    dplyr::group_by(.data$Package) |>
    # thanks: https://stackoverflow.com/questions/43832434/arrange-within-a-group-with-dplyr
    dplyr::top_n(1, dplyr::desc(.data$Date)) |>
    dplyr::collect()
}

.matchGetShortName <- function(pkgType, colName) {
    rsn <- .repos_short_names()
    rsn[match(pkgType, rsn[["code.names"]]), colName]
}

#' What is a package's download rank?
#'
#' This function uses `available.packages` to calculate the download rank
#' _percentile_ of a given package. It approximates what is observed
#' in the Bioconductor landing page.
#'
#' @inheritParams pkgDownloadStats
#' @inheritParams repositoryStats
#'
#' @return The package's percentile rank, in terms of download statistics, and
#'   proportion in the name
#'
#' @examples
#' ## Percentile rank for BiocGenerics (top 1%)
#' pkgDownloadRank("BiocGenerics", "software")
#' @export
pkgDownloadRank <-
    function(
        pkg,
        pkgType = c(
            "software", "data-experiment", "workflows", "data-annotation"
        ),
        version = BiocManager::version()
    )
{
    pkgType <- match.arg(pkgType)
    jsonPath <- .matchGetShortName(pkgType, "pkg.url")
    rankURL <-
        "https://www.bioconductor.org/packages/json/%s/%s/packages.json"
    ufile <- .cache_url_file(sprintf(rankURL, version, jsonPath))

    pkgs <- jsonlite::fromJSON(ufile)
    rank <- pkgs[[pkg]]$Rank

    repoType <- .matchGetShortName(pkgType, "biocmanager.names")
    viewsdb <- .get_VIEWS(version = version, repoType)

    pct <- round(rank*100 / nrow(viewsdb), 2)
    names(pct) <- paste(rank, nrow(viewsdb), sep = "/")
    pct
}

#' Summary of the latest package statistics
#'
#' The `latestPkgStats` function combines outputs from several functions to
#' generate a table of relevant statistics for a given package.
#'
#' @inheritParams pkgDownloadRank
#' @inheritParams activitySince
#'
#' @examplesIf interactive()
#' latestPkgStats("Bioconductor/BiocGenerics", "2021-05-05")
#' @export
latestPkgStats <-
    function(
        gh_repo,
        Date,
        pkgType = c(
            "software", "data-experiment", "workflows", "data-annotation"
        )
    )
{
    if (!requireNamespace("lubridate", quietly = TRUE))
        stop("Install the 'lubridate' package to work with dates")
    y1 <- lubridate::year(Date)
    y2 <- lubridate::year(Sys.Date())
    m1 <- lubridate::month(Date, label = TRUE, abbr = TRUE)

    pkg <- strsplit(gh_repo, "/", fixed = TRUE)[[1]][[2]]
    stattab <- pkgDownloadStats(pkg, years = y1:y2)
    afterDate <- seq_len(nrow(stattab)) >=
      with(stattab, which(Year == y1 & Month == m1))
    csince <- nrow(activitySince(gh_repo, "commits", Date = Date))
    stattab[afterDate, ] |>
      dplyr::summarize(
          avg_mo_distict_IPs = mean(Nb_of_distinct_IPs),
          avg_mo_downloads = mean(Nb_of_downloads)
      ) |>
      dplyr::mutate(Package = pkg, .before = 1, Date = Date) |>
      dplyr::mutate(
          downloadRank = pkgDownloadRank(pkg, pkgType),
          issuesClosedSince =
              nrow(activitySince(gh_repo, "issues", "closed", Date)),
          commitsSince = csince
      )
}

.commit_table <- function(jsonlist) {
    parents <- jsonlist[["parents"]][["sha"]]
    if (is.null(parents)) parents <- NA_character_
    cdata <- jsonlist[["commit"]]
    author <- cdata[["author"]][["name"]]
    commit <- cdata[["tree"]][["sha"]]
    committer_date <- cdata[["committer"]][["date"]]
    message <- cdata[["message"]]
    tibble::tibble(committer_date, commit, parents, author, message)
}

.gh_data_process <- function(x, fields, commits = FALSE) {
    if (commits)
      .commit_table(x)
    else
      x[fields]
}

#' What are the issues, pulls, commits created since a date?
#'
#' This function uses the `gh` package to get a list of either issues, pull
#' requests, or GitHub commits since the specified date for a particular GitHub
#' repository. The repository must have both the username / organization and the
#' name, e.g., "Bioconductor/S4Vectors".
#'
#' @details The `tibble` returned by the commits activity report contains five
#'   columns:
#'   * 'committer_date'
#'   * 'commit' - hash
#'   * 'parents' - hash of parent for merge commits
#'   * 'author'
#'   * 'message'
#'
#'   For information on other columns, refer to the GitHub API under repository
#'   issues or pulls (e.g., `/repos/:repo/issues`).
#'
#' @param gh_repo `character(1)` The GitHub repository location including the
#'   username / organization and the repository name, e.g.,
#'   "Bioconductor/S4Vectors"
#'
#' @param activity `character(1)` The type of repository activity to pull from the
#'   GitHub API. It can be one of "issues" (default), "pulls", or "commits".
#'
#' @param Date `character(1)` The date cutoff from which to analyze closed issues
#'   in the YYYY-MM-DD or YYYY-MM-DDTHH:MM:SSZ format (ISO 8601).
#'
#' @param status `character(1)` One of 'closed', 'open', or 'all' corresponding to
#'   the issue state desired from the GitHub API (Default: "closed"). This
#'   argument is ignored for the "commits" activity report.
#'
#' @param issue_metadata `character()` The metadata labels to extract from the
#'   `gh::gh` response. See `?gh::gh` for more details. Defaults to
#'   'created_at', 'number', and 'title'. This argument is ignored for the
#'   "commits" activity report.
#'
#' @param token `character(1)` For big requests, e.g., commit history, you may be
#'   prompted to use a GitHub Personal Access Token. Enter the token as plain
#'   text.
#'
#' @return A `tibble` with three columns corresponding to issue metadata (i.e.,
#'   "created_at", "number", "title")
#'
#' @examplesIf interactive()
#' activitySince("Bioconductor/S4Vectors", "issues", "closed", "2021-05-01")
#' activitySince("Bioconductor/S4Vectors", "issues", "open", "2022-05-01")
#' activitySince("Bioconductor/S4Vectors", "commits", Date = "2022-05-01")
#' @export
activitySince <- function(
    gh_repo,
    activity = c("issues", "pulls", "commits"),
    status = c("closed", "open", "all"),
    Date,
    issue_metadata = c("created_at", "number", "title"),
    token = NULL
) {
    activity <- match.arg(activity)
    isCommits <- identical(activity, "commits")
    if (!isCommits)
      status <- match.arg(status)
    else
      status <- NULL
    if (!requireNamespace("lubridate", quietly = TRUE))
        stop("Install the 'lubridate' package to work with dates")
    accept <- "application/vnd.github.v3+json"
    act_report <- gh(
        endpoint = paste0("/repos/:repo/", activity),
        .token = token,
        repo = gh_repo,
        .limit = Inf,
        state = status,
        since = Date,
        .accept = accept
    )
    dplyr::bind_rows(
        lapply(act_report, .gh_data_process,
            fields = issue_metadata, commits = isCommits)
    )
}

context("biocDownloadStats")

test_that("biocDownloadStats works", {
    res <- biocDownloadStats("software")
    expect_s3_class(res, "bioc_downloads")
    expect_s3_class(res, "tbl_df")
    expect_true(
        all(
            c(
                "Package",
                "Year",
                "Month",
                "Nb_of_distinct_IPs",
                "Nb_of_downloads",
                "Date"
            ) %in%
                names(res)
        )
    )
})

test_that("pkgDownloadStats works", {
    res <- pkgDownloadStats("BiocGenerics")
    expect_s3_class(res, "tbl_df")
    expect_true(
        all(
            c("Year", "Month", "Nb_of_distinct_IPs", "Nb_of_downloads") %in%
                names(res)
        )
    )
})

test_that("pkgDownloadRank works", {
    res <- pkgDownloadRank("BiocGenerics", "software")
    expect_type(res, "double")
    expect_true(!is.na(res))
})

test_that("activitySince works", {
    res <- activitySince(
        "Bioconductor/BiocGenerics",
        "issues",
        "closed",
        "2021-05-01"
    )
    expect_s3_class(res, "tbl_df")
    expect_true(
        all(c("created_at", "number", "title") %in% names(res))
    )
})

.pkgsLinks <- function(pkgs, version, md = TRUE) {
    db <- available.packages(repos = BiocManager::repositories())
    pkgsdb <- db[rownames(db) %in% pkgs, , drop = FALSE]
    isbioc <- grepl("bioconductor", pkgsdb[, "Repository"], fixed = TRUE)
    c(
        .biocPkgsLinks(pkgs[isbioc], version = version, md = md),
        .CRANpkgsLinks(pkgs[!isbioc], md = md)
    )
}

.biocPkgsLinks <- function(pkgs, version, md = TRUE) {
    if (!length(pkgs))
        return(NULL)

    buildrepURL <- paste0(
        "https://bioconductor.org/checkResults/", version, "/bioc-LATEST/", pkgs
    )
    if (md)
        paste0("[", pkgs, "]", "(", buildrepURL, ")")
    else
        buildrepURL
}

.CRANpkgsLinks <- function(pkgs, md = TRUE) {
    if (!length(pkgs))
        return(NULL)

    cranrepURL <- paste0(
        .CRAN_WEB_URL, "package=", pkgs
    )
    if (md)
        paste0("[", pkgs, "]", "(", cranrepURL, ")")
    else
        cranrepURL
}

#' Notify downstream maintainers of changes in upstream packages
#'
#' @description
#' The `biocRevDepEmail` function collects all the emails of the reverse
#' dependencies and sends a notification that upstream package(s) have been
#' deprecated or removed. It uses a template found in `inst/resources` with the
#' `templatePath()` function.
#'
#' @param packages `character()` A vector of CRAN and/or Bioconductor packages
#'   for whose reverse dependencies are to be checked and notified.
#'
#' @param cc `character()` A vector of email addresses for sending the message
#'   as a carbon copy.
#'
#' @param pkg `character(1)` DEPRECATED. The name of a single package whose
#'   reverse dependencies are to be checked and notified.
#'
#' @param \ldots Additional inputs to internal functions (not used).
#'
#' @inheritParams biocBuildEmail
#' @inheritParams tools::package_dependencies
#'
#' @examples
#'
#' biocRevDepEmail(
#'     "FindMyFriends", version = "3.13", dry.run = TRUE, textOnly = TRUE
#' )
#'
#' @export
biocRevDepEmail <-
    function(packages, which = c("strong", "most", "all"),
        PS = character(1L), version = BiocManager::version(),
        dry.run = TRUE,  cc = NULL, emailTemplate = templatePath("revdepnote"),
        core.name = NULL, core.email = NULL, core.id = NULL,
        textOnly = FALSE, verbose = FALSE, credFile = "~/.blastula_creds",
        ..., pkg)
{
    stopifnot(
        is.character(packages),
        is.character(PS), identical(length(PS), 1L),
        !is.na(packages), !is.na(PS), !is.na(core.name), !is.na(core.email),
        !is.na(core.id)
    )
    if (!file.exists(emailTemplate))
        stop("'emailTemplate' file not found.")
    which <- match.arg(which)

    core.list <- .getUserInfo(core.name, core.email, core.id)
    core.name <- core.list[["core.name"]]
    core.email <- core.list[["core.email"]]
    core.id <- core.list[["core.id"]]

    if (!missing(pkg))
        .Defunct(msg = "'pkg' argument is defunct. Use 'packages'.")

    db <- available.packages(
        repos = BiocManager:::.repositories_bioc(version)["BioCsoft"]
    )
    revdeps <- tools::package_dependencies(
        packages, db, reverse = TRUE, which = which
    )
    revdeps <- Filter(length, revdeps)
    revdeps <- unlist(revdeps, use.names = FALSE)

    if (!length(revdeps))
        stop("No reverse dependencies on ", packages)

    listall <- biocPkgList(version = version)
    pkgMeta <- listall[listall[["Package"]] %in% revdeps, "Maintainer"]
    mainInfo <- vapply(
        unlist(pkgMeta, use.names = FALSE), .emailCut, character(1L)
    )

    mainEmails <- unname(vapply(mainInfo, .emailCut, character(1L)))

    repolinks <- .pkgsLinks(packages, version = version, md = TRUE)
    bioclinks <- .biocPkgsLinks(revdeps, version = version)

    if (nchar(PS))
        PS <- paste0("**P.S.** ", PS)

    revdeps <- paste(bioclinks, collapse = "\n\n")
    mail <- paste0(readLines(emailTemplate), collapse = "\n")
    maildate <- format(Sys.time(), "%B %d, %Y")
    firstname <- vapply(strsplit(core.name, "\\s"), `[`, character(1L), 1L)
    cpackages <- paste(packages, collapse = ", ")
    crepolinks <- paste(repolinks, collapse = ", ")
    send <- sprintf(
        mail, core.name, maildate, cpackages, crepolinks,
        revdeps, PS, firstname
    )
    title <- "Package(s) Deprecation Notification"

    if (dry.run)
        message("Message not sent: Set 'dry.run=FALSE'")

    if (textOnly) {
        send <- strsplit(send, "---")[[1L]][[4L]]
        mainEmails <- paste(mainEmails, collapse = ", ")
        send <- paste(mainEmails, title, send, sep = "\n")
        if (requireNamespace("clipr", quietly=TRUE) &&
            clipr::clipr_available())
        {
            clipr::write_clip(send)
            message("Message copied to clipboard")
        }
    } else {
        tfile <- tempfile(fileext = ".Rmd")
        writeLines(send, tfile)
        send <- blastula::render_email(tfile)
        if (!dry.run) {
            blastula::smtp_send(email = send,
                from = core.email, to = core.email, bcc = mainEmails, cc = cc,
                subject = title,
                credentials =
                    if (file.exists(credFile)) {
                        blastula::creds_file(credFile)
                    } else {
                        blastula::creds(
                            user = paste0(core.id, "@roswellpark.org"),
                            host = "smtp.office365.com",
                            port = 587,
                            use_ssl = TRUE
                        )
                    }, verbose = verbose
            )
        }
    }
    send
}

#' BiocPkgTools: Examine and analyze Bioconductor package metadata
#'
#' Bioconductor has a rich ecosystem of metadata around
#' packages, usage, and build status. This package is a simple
#' collection of functions to access that metadata from R. The goal
#' is to expose metadata for data mining and value-added functionality
#' such as package searching, text mining, and analytics on
#' packages.
#'
#' @section For developers:
#' The [biocBuildReport()] function returns a computable
#' form of the Bioconductor Build Report.
#'
#' @section For users:
#' The [biocDownloadStats()] function gets Bioconductor
#' download stats, allowing users to quickly find commonly used
#' packages. The [biocPkgList()] is useful for getting
#' a complete listing of all Bioconductor packages.
#'
#' @section Infrastructure:
#' Bioconductor packages all have Digital Object Identifiers (DOIs).
#' This package contains basic infrastructure for creating, updating,
#' and de-referencing DOIs.
#'
#' @name BiocPkgTools
"_PACKAGE"

utils::globalVariables(c(
    "Release", "approx_years_in", "category",
    "first_version_available",
    "first_version_release_date", "last_version_available",
    "last_version_release_date", "major", "minor",
    "package", "packages", "release", "years_before_rm"
))

#' Calculate the years in Bioconductor
#'
#' @description This function determines the number of years a package has been
#'   in Bioconductor. Available information includes first Bioconductor version
#'   a package appeared and the current length of time in Bioconductor. If a
#'   package has been removed from Bioconductor, information on the last
#'   Bioconductor version and approximate time in Bioconductor before removal is
#'   available.
#'
#' @param pkglist List of packages to retrieve information. If default NULL,
#'     returns a tibble of all Bioconductor packages.
#'
#' @author Lori Shepherd Kern, Robert Shear
#'
#' @return 'tibble' with the following columns:
#'   * package: name of Bioconductor package
#'   * category: bioc, data/experiment, data/annotation, workflow
#'   * first_version_available: Bioconductor version (e.g. 1.9, 3.21) the
#'     package first became available
#'   * first_version_release_date: Equivalent calendar date of given
#'     Bioconductor release
#'   * approx_years_in: Numeric indicator of years in Bioconductor. If empty,
#'     indicates package was removed. See final three columns for more
#'     information.
#'   * last_version_available: If package was removed from Bioconductor, the
#'     last Bioconductor version (e.g. 1.9, 3.21) the package was able to be
#'     installed
#'   * last_version_release_date: Equivalent calendar date of given Bioconductor
#'     release
#'   * years_before_rm: If removed, how many years it was in Bioconductor
#'
#' @importFrom yaml read_yaml
#' @importFrom dplyr filter mutate select group_by summarize arrange
#' @importFrom purrr list_rbind map2
#' @importFrom glue glue
#' @importFrom curl curl has_internet
#' @importFrom tidyr tibble unnest_longer
#' @importFrom lubridate mdy
#' @importFrom BiocFileCache bfcquery bfcnew bfcrpath
#'
#' @examples
#' \dontrun{
#'    ## full table all Bioconductor packages
#'    tbl <- getPkgYearsInBioc()
#'
#'    ## example of package list. Packages active in Bioconductor
#'    tbl <- getPkgYearsInBioc(c("BiocFileCache", "BiocPkgTools"))
#'
#'    ## example of a package that has been removed from Bioconductor
#'    tbl <- getPkgYearsInBioc("ensemblVEP")
#' }
#'
#' @export
getPkgYearsInBioc <- function(pkglist = NULL) {
    bfc <- .get_cache()
    rid <- BiocFileCache::bfcquery(
        bfc, "bioc.pkg.years.info", "rname", exact = TRUE
    )[["rid"]]
    if (!length(rid)) {
        pkginfo <-
            BiocFileCache::bfcnew(bfc, "bioc.pkg.years.info", ext = ".Rdata")
    } else {
        pkginfo <- BiocFileCache::bfcrpath(bfc, rids = rid)
    }

    if (!file.exists(pkginfo)) {
        NeededYearsData <- .generateBiocYearsInData(pkginfo)
    } else {
        load(pkginfo)
        if (has_internet()) {
            message("Checking for Bioc Release Update")
            NeededYearsData <- .updateYearsInBioc(NeededYearsData, pkginfo)
        } else {
            message("Cannot check for release update without internet.")
        }
    }
    YearsInBiocDF <- .formatBiocYearsDF(NeededYearsData)
    if (is.null(pkglist)) {
        return(YearsInBiocDF)
    } else {
        return(filter(YearsInBiocDF, package %in% pkglist))
    }
}

## Internal: reformat information to be per package instead of per release.
## This will also calculate out when package first appears and when if
## applicable last appears (if it was removed).
.formatBiocYearsDF <- function(NeededYearsData) {
    config <- NeededYearsData[["config"]]
    manifestDF <- NeededYearsData[["manifestDF"]]

    manifestDFUnlist <- manifestDF |>
        unnest_longer(packages) |>
        mutate(
            release = Release, version = major * 100 + minor,
            category = packages$category, package = packages$package,
            .keep = "none"
        )

    current_version <- max(manifestDFUnlist$version)

    manifestByPkg <- manifestDFUnlist |>
        group_by(package) |>
        summarize(
            category = category[which.max(version)],
            first_version_available = release[which.min(version)],
            last_version_available = release[which.max(version)]
        ) |>
        arrange(package)

    release_dates <- unlist(config[["release_dates"]])
    release_dates_date <- c(mdy(release_dates), NA)
    names(release_dates_date) <-
        c(names(release_dates), config[["devel_version"]])
    since <-
        as.numeric(round(((Sys.Date() - release_dates_date) / 365.25) * 2) / 2)
    names(since) <- c(names(release_dates), config[["devel_version"]])

    YearsInBiocDF <- manifestByPkg |>
        mutate(
            first_version_release_date =
                release_dates_date[first_version_available],
            last_version_release_date =
                release_dates_date[last_version_available],
            approx_years_in = ifelse(
                last_version_available == config[["devel_version"]],
                since[first_version_available],
                NA
            ),
            years_before_rm = ifelse(
                last_version_available == config[["devel_version"]],
                NA,
                as.numeric(
                    round(
                        (
                            (
                                (
                                    last_version_release_date -
                                        first_version_release_date
                                ) / 365.25
                            ) * 2
                        ) / 2
                    )
                )
            ),
            last_version_available =
                ifelse(
                    last_version_available == config[["devel_version"]],
                    NA_character_,
                    last_version_available
                )
        ) |>
        select(
            package, category, first_version_available,
            first_version_release_date, approx_years_in,
            last_version_available, last_version_release_date,
            years_before_rm
        )
    return(YearsInBiocDF)
}

## Internal: Check if the current data frame is up-to-date or needs to be
## updated with a new Bioconductor release. Avoids regenerating entire data
## frame and only adds and saves update
.updateYearsInBioc <- function(NeededYearsData, cachepathinfo) {
    config <- NeededYearsData[["config"]]
    manifestDF <- NeededYearsData[["manifestDF"]]

    file_path <- "https://bioconductor.org/config.yaml"
    current_config <- yaml::read_yaml(file_path)
    current_bioc_ver <- names(current_config$r_ver_for_bioc_ver)
    ## no info for 1.6/1.7 despite listing in config
    ## remove so that it does not try to generate information
    neededReleases <-
        setdiff(current_bioc_ver, c("1.6", "1.7", manifestDF$Release))
    if (length(neededReleases) > 0) {
        bioc_ver <- data.frame(Release = neededReleases)
        newEntries <- cbind(
            bioc_ver,
            list_rbind(lapply(
                strsplit(bioc_ver$Release, ".", fixed = TRUE),
                (\(u) data.frame(
                    major = as.integer(u[1]),
                    minor = as.integer(u[2])
                ))
            ))
        ) |>
            filter(major > 1 | (major == 1 & minor > 7)) |>
            mutate(packages = map2(major, minor, .genBiocManifestDF))
        manifestDF <- rbind(manifestDF, newEntries)
        NeededYearsData[["manifestDF"]] <- manifestDF
        NeededYearsData[["config"]] <- current_config
        save(NeededYearsData, file = cachepathinfo)
    }
    return(NeededYearsData)
}

## Internal: Determine all Bioconductor releases and loop over
## .genBiocManfiestDF to determine list of package per release
## Save data frame to given [cache] path to avoid lengthy regeneration.
.generateBiocYearsInData <- function(cachefilepath) {
    message(
        "Internet Access Required For Data Generation \n",
        "  First time may take a few minutes, \n",
        "  Subsequent runs should be instantaneous."
    )

    ## get all packages for each release

    file_path <- "https://bioconductor.org/config.yaml"
    config <- yaml::read_yaml(file_path)
    bioc_ver <- data.frame(Release = names(config$r_ver_for_bioc_ver))
    manifestDF <- cbind(
        bioc_ver,
        list_rbind(lapply(
            strsplit(bioc_ver$Release, ".", fixed = TRUE),
            (\(u) data.frame(
                major = as.integer(u[1]),
                minor = as.integer(u[2])
            ))
        ))
    ) |>
        filter(major > 1 | (major == 1 & minor > 7)) |>
        mutate(packages = map2(major, minor, .genBiocManifestDF))
    NeededYearsData <- list(config = config, manifestDF = manifestDF)
    save(NeededYearsData, file = cachefilepath)
    return(NeededYearsData)
}

## Internal: Loop over versions to determine package list of packages in a given
## release
.genBiocManifestDF <- function(major, minor) {
    manifest_template <- paste0(
        "https://www.bioconductor.org/packages/",
        "{version}/{category}/src/contrib/PACKAGES"
    )
    categories <- c("bioc", "data/annotation", "data/experiment")
    ## the workflows category first appeared in 2.13
    if (major > 2 || (major == 2 && minor > 12)) {
        categories <- append(categories, "workflows")
    }
    list_rbind(mapply(
        \(category, version) {
            print(glue("{category} / {version}"))

            con <- curl(glue(manifest_template))
            result <- read.dcf(con)
            close(con)
            tibble(category, package = result[, 1])
        },
        categories, glue("{major}.{minor}"),
        SIMPLIFY = FALSE
    ))
}

#' Get download statistics for Bioconductor packages distributed via
#' Anaconda.
#'
#' @details Anaconda provide daily download counts for all software packages
#' they distribute.  These are summarised into monthly tables of counts and made
#' available from https://github.com/grimbough/anaconda-download-stats
#' This function provides a mechanism to download these monthly counts for
#' Bioconductor packages distributed through Anaconda.
#'
#' @importFrom dplyr mutate select arrange desc
#' @importFrom utils read.table download.file
#' @importFrom tibble as_tibble
#'
#' @return A \code{data.frame} of download statistics for
#' all Bioconductor packages distributed by Anaconda, in tidy format.
#' Note: Anaconda do not provide counts for unique IP addresses.  This column
#' is listed as `NA` for all packages to provide continuity with data from
#' Bioconductor.org obtained by [biocDownloadStats()].  The counts are
#' updated monthly, so do not expect to see counts for the current month.
#'
#' @author Mike L. Smith
#'
#' @examples
#' anacondaDownloadStats()
#'
#' @export
anacondaDownloadStats <- function() {
    temp_file <- tempfile()
    url <- paste0(
        "https://github.com/grimbough/anaconda-download-stats",
        "/raw/master/rdata/bioc_counts.rds"
    )
    download.file(url, destfile = temp_file, quiet = TRUE, mode = "wb")

    tmp <- readRDS(temp_file)
    tmp$repo <- "Anaconda"
    tmp$Nb_of_distinct_IPs <- NA_integer_

    tmp <- as_tibble(tmp) |>
        dplyr::mutate(Date = as.Date(
            paste(.data$Year, .data$Month, "01"),
            "%Y %b %d"
        )) |>
        select(
            "Package", "Year", "Month", "Nb_of_distinct_IPs",
            "Nb_of_downloads", "repo", "Date"
        ) |>
        ## put this into the same order as the BioC table
        arrange(tmp, .data$Package, desc(.data$Year), .data$Date)

    class(tmp) <- c("bioc_downloads", class(tmp))
    tmp
}

checkMe <- function(ver="devel", authorPattern="V.*Carey", includeOK = FALSE) {
 rep = biocBuildReport(ver)
 kp = grep(authorPattern, rep$author)
 mine = rep[kp,]
 if(nrow(mine) == 0L) stop("Used authorPattern return zero results.",
                           call. = FALSE)
 if(includeOK)
     bad = mine
 else
     bad = mine[mine$result != "OK",]
 if (nrow(bad)>0) return(bad[,-c(3,4,5)]) else return(NULL)
}

chkURL <- function(ver="devel", result = "ERROR", pack="BiocOncoTK",
   node="malbec1", stage="buildsrc"){
  urls <- rep("",length(ver))
  skipped <- result %in% c("skipped","NA")
  urls[!skipped] <- sprintf("https://bioconductor.org/checkResults/%s/bioc-LATEST/%s/%s-%s.html",
                            ver[!skipped], pack[!skipped], node[!skipped], stage[!skipped])
  urls[skipped] <- sprintf("https://bioconductor.org/checkResults/%s/bioc-LATEST/%s/",
                           ver[skipped], pack[skipped])
  urls
}

checkDeps <- function(dependsOn, ver="devel", includeOK = FALSE) {

  if(missing(dependsOn)) {
    stop("You must provide a package name to the 'dependsOn' argument")
  }

  rep = biocBuildReport(ver)
  all_pkg_deps = buildPkgDependencyDataFrame()
  pkg_deps <- all_pkg_deps[ all_pkg_deps$dependency == dependsOn, 1]

  mine = rep[rep$pkg %in% pkg_deps$Package, ]

  if(nrow(mine) == 0L) stop("No dependent packages found.",
                            call. = FALSE)
  if(includeOK)
    bad = mine
  else
    bad = mine[mine$result != "OK",]

  if (nrow(bad)>0) return(bad[,-c(3,4,5)]) else return(NULL)
}

#' generate hyperlinked HTML for build reports for Bioc packages
#'
#' This is a quick way to get an HTML report of packages maintained by a specific developer
#' or which depend directly on a specified package. The function is keyed to filter based on either
#' the maintainer name or by using the 'Depends', 'Suggests' and 'Imports' fields in package descriptions.
#'
#' @importFrom htmltools a
#' @importFrom DT datatable
#'
#' @param authorPattern `character(1)` regexp used with grep() to filter author field of package DESCRIPTION for listing
#' @param dependsOn `character(1)` name of a Bioconductor package. The function will return the status of packages
#' that directly depend on this package  Can only be used when 'authorPattern' is the empty string.
#' @param ver `character(1)` version tag for Bioconductor
#' @param includeOK `logical(1)` include entries from the build report that are listed
#'     as "OK". Default FALSE will result in only those entries that
#'     are in WARNING or ERROR state.
#'
#' @return DT::datatable call; if assigned to a variable, must evaluate to get the page to appear
#'
#' @author Vince Carey, Mike L. Smith
#'
#' @examples
#' if (interactive()) {
#'   problemPage()
#'   problemPage(dependsOn = "limma")
#' }
#'
#' @export
problemPage <- function(
    authorPattern="V.*Carey", dependsOn, ver="devel", includeOK = FALSE
) {

    if (!requireNamespace("htmltools")) stop("install htmltools to use this function")
    if (!requireNamespace("DT")) stop("install DT to use this function")

    if(nchar(authorPattern) > 0 && !missing(dependsOn)) {
      warning("Both 'authorPattern' and 'dependsOn' arguments provided\n",
              "Listing results for 'authorPattern'.")
      dependsOn = ""
    }

    ver = as.character(ver)

    if(nchar(authorPattern) > 0)
      mm = checkMe(authorPattern=authorPattern, ver=ver, includeOK = includeOK)
    else
      mm = checkDeps(dependsOn = dependsOn, ver=ver, includeOK = includeOK)

    nn = nrow(mm)
    if (is.null(nn)) stop("all packages fine")
    cc = chkURL(mm[["bioc_version"]], mm[["result"]], mm[["pkg"]], mm[["node"]],
                mm[["stage"]])
    hr = lapply(seq_len(nrow(mm)), function(x)
        htmltools::a(mm[x, "pkg"], href=cc[[x]]))
    col1 = unlist(lapply(hr, as.character))
    DT::datatable(data.frame(package = col1, mm[,-1], stringsAsFactors=FALSE),
                  escape=FALSE)
}

#' Bioconductor Binary Repository Statistics
#'
#' @aliases print.repositoryStats
#'
#' @description Summarize binary packages compatible with the Bioconductor or
#'   Terra container in use.
#'
#' @inheritParams biocBuildReport
#'
#' @param binary_repository `character(1)` location of binary repository as
#'   given by `BiocManager::containerRepository` (default)
#'
#' @param local `logical(1)` whether to check the local file system for the
#'   `PACKAGES` file's last modified date (default: `FALSE`).
#'
#' @param ... further arguments passed to or from other methods (not used).
#'
#' @return a list of class `repositoryStats` with the following fields:
#' * container: `character(1)` container label, e.g.,
#'   `bioconductor_docker`, or NA if not evaluated on a supported container
#' * bioconductor_version: `package_version` the
#'   Bioconductor version provided by the user.
#' * repository_exists: `logical(1)` TRUE if a binary repository
#'   exists for the container and Bioconductor_Version version.
#' * bioconductor_binary_repository: `character(1)` repository
#'   location, if available, or NA if the repository does not exist.
#' * n_software_packages: `integer(1)` number of software packages
#'   in the Bioconductor source repository.
#' * n_binary_packages: `integer(1)` number of binary packages
#'   available. When a binary repository exists, this number is likely
#'   to be larger than the number of source software packages, because
#'   it includes the binary version of the source software packages, as
#'   well as the (possibly CRAN) dependencies of the binary packages
#' * n_binary_software_packages: `integer(1)` number of binary
#'   packages derived from Bioconductor source packages. This number is
#'   less than or equal to `n_software_packages`.
#' * missing_binaries: `integer(1)` the number of Bioconductor
#'   source software packages that are not present in the binary
#'   repository.
#' * out_of_date_binaries: `integer(1)` the number of Bioconductor
#'   source software packages that are newer than their binary
#'   counterpart. A newer source software package
#'   might occur when the main Bioconductor build system has
#'   updated a package after the most recent run of the binary
#'   build system.
#'
#' @importFrom utils available.packages
#' @importFrom httr HEAD headers
#' @importFrom BiocManager containerRepository
#'
#' @details For local repositories, use the `local = TRUE` argument. Local
#'   repositories will typically start with the `file://` URI. The function
#'   checks the `mtime` of the output of `file.info` on the `PACKAGES` file in
#'   the local repository. Otherwise, by default, it will check the
#'   `last-modified` header of the `PACKAGES` file via `httr::HEAD()`.
#'
#' @author M. Morgan
#'
#' @examples
#' stats <- repositoryStats() # obtain statistics
#' stats                       # display a summary
#' stats$container             # access an element for further computation
#'
#' @export
repositoryStats <- function(
    version = BiocManager::version(),
    binary_repository = BiocManager::containerRepository(version),
    local = FALSE
) {
    if (version %in% c("release", "devel"))
        version <- BiocManager:::.version_bioc(version)

    version <- try(as.package_version(version), silent = TRUE)

    if (is(version, "try-error"))
        stop(
            "'version' is not 'release', 'devel', or a valid 'package_version'"
        )

    platform_docker <- BiocManager:::.repository_container_version()
    container <- platform_docker$platform
    bioc_repository <- suppressMessages({
        BiocManager::repositories()[["BioCsoft"]]
    })
    db_bioc <- available.packages(repos = bioc_repository)
    if (length(binary_repository)) {
        db_binary <- available.packages(repos = binary_repository)
        packages <- paste0(contrib.url(binary_repository), "/PACKAGES")
        if (!local) {
            response <- HEAD(packages)
            last_modified <- headers(response)$`last-modified`
        } else {
            last_modified <- file.info(packages)$mtime
        }
        PACKAGES_mtime <-
            format(strptime(
                last_modified, "%a, %d %b %Y %H:%M", tz = "UTC"
            ), usetz = TRUE)
    } else {
        db_binary <- db_bioc[NULL,]
        PACKAGES_mtime <- NA_character_
    }

    missing_binaries <- setdiff(rownames(db_bioc), rownames(db_binary))
    found_binaries <- intersect(rownames(db_bioc), rownames(db_binary))

    bioc_versions <- package_version(db_bioc[found_binaries, "Version"])
    binary_versions <- package_version(db_binary[found_binaries, "Version"])
    binary_out_of_date <- bioc_versions > binary_versions
    n_out_of_date_binaries <- sum(binary_out_of_date)
    out_of_date_binaries <- found_binaries[binary_out_of_date]

    query_timestamp = format(
        Sys.time(), "%Y-%m-%d %H:%M", tz = "UTC", usetz = TRUE
    )

    result <- list(
        container = if (nzchar(container)) container else NA_character_,
        bioconductor_version = version,
        bioconductor_binary_repository =
            if (length(binary_repository)) binary_repository else NA_character_,
        PACKAGES_mtime = PACKAGES_mtime,
        query_timestamp = query_timestamp,
        repository_exists = length(binary_repository) > 0L,
        n_software_packages = nrow(db_bioc),
        n_binary_packages = nrow(db_binary),
        n_binary_software_packages = length(found_binaries),
        missing_binaries = missing_binaries,
        out_of_date_binaries = out_of_date_binaries
    )
    class(result) <- c("repositoryStats", class(result))
    result
}

.repositoryStats_package_format <-
    function(x)
{
    msg <- paste(sort(x), collapse = " ")
    msg <- paste(strwrap(msg, indent = 2L, exdent = 2L), collapse = "\n")
    paste0(msg, "\n")
}

#' @describeIn repositoryStats Print a summary of package
#'     availability in binary repositories.
#'
#' @param x the object returned by `repositoryStats()`.
#'
#' @export
print.repositoryStats <-
    function(x, ...)
{
    bioconductor_binary_repository <- ifelse(
        is.na(x$bioconductor_binary_repository),
        paste(" ", x$bioconductor_binary_repository),
        paste("\n  ", x$bioconductor_binary_repository)
    )
    cat(
        "Container: ", x$container, "\n",
        "Bioconductor version: ", as.character(x$bioconductor_version), "\n",
        "Bioconductor binary repos:", bioconductor_binary_repository, "\n",
        "PACKAGES timestamp: ", x$PACKAGES_mtime, "\n",
        "Query timestamp: ", x$query_timestamp, "\n",
        "Bioconductor software packages: ", x$n_software_packages, "\n",
        "Binary packages: ", x$n_binary_packages, "\n",
        "Binary software packages: ", x$n_binary_software_packages, "\n",
        "Missing binary software packages: ", length(x$missing_binaries), "\n",
        if (x$repository_exists)
            .repositoryStats_package_format(x$missing_binaries),
        "Out-of-date binary software packages: ",
            length(x$out_of_date_binaries), "\n",
        if (x$repository_exists)
            .repositoryStats_package_format(x$out_of_date_binaries),
        sep = ""
    )
}

#' Work with Bioconductor package dependencies
#'
#' Bioconductor is built using an extensive set of
#' core capabilities and data structures. This leads
#' to package developers depending on other packages
#' for interoperability and functionality. This
#' function extracts package dependency information
#' from [biocPkgList()] and returns a tidy
#' `data.frame` that can be used for analysis
#' and to build graph structures of package dependencies.
#'
#' @param dependencies `character()` a vector listing the types of dependencies, a
#'   subset of c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances").
#'   Character string "all" is shorthand for that vector, character string
#'   "most" for the same vector without "Enhances", character string "strong"
#'   (default) for the first three elements of that vector.
#'
#' @param ... parameters passed along to [biocPkgList()]
#'
#' @importFrom dplyr bind_rows
#'
#' @seealso See [buildPkgDependencyIgraph()], [biocPkgList()].
#'
#' @note This function requires network access.
#'
#' @return A `data.frame` (also a `tbl_df`) of
#' S3 class "biocDepDF" including columns "Package", "dependency",
#' and "edgetype".
#'
#' @examples
#' # performs a network call, so must be online.
#' library(BiocPkgTools)
#' depdf <- buildPkgDependencyDataFrame()
#' head(depdf)
#' library(dplyr)
#' # filter to include only "Imports" type
#' # dependencies
#' imports_only <- depdf |> filter(edgetype=='Imports')
#'
#' # top ten most imported packages
#' imports_only |> select(dependency) |>
#'   group_by(dependency) |> tally() |>
#'   arrange(desc(n))
#'
#' # The Bioconductor packages with the
#' # largest number of imports
#' largest_importers <- imports_only |>
#'   select(Package) |>
#'   group_by(Package) |> tally() |>
#'   arrange(desc(n))
#'
#' # not sure what these packages do. Join
#' # to their descriptions
#' biocPkgList() |> select(Package, Description) |>
#'   left_join(largest_importers) |> arrange(desc(n)) |>
#'   head()
#' @export
buildPkgDependencyDataFrame <-
    function(
        dependencies = c("strong", "most", "all"),
        ...
    )
{
    dadeps <- c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")
    if (missing(dependencies))
        dependencies <- match.arg(dependencies)
    if (identical(length(dependencies), 1L))
        dependencies <- switch(
            dependencies,
            strong = dadeps[1:3], most = dadeps[1:4], all = dadeps[1:5],
            dependencies
        )
    else
        dependencies <- match.arg(
            dependencies, choices = dadeps, several.ok = TRUE
        )
    x <- biocPkgList(...)
    reslist <- lapply(dependencies, function(dep) {
        resframe <- cbind.data.frame(
            Package = rep(x[["Package"]], lengths(x[[dep]])),
            dependency = stripVersionString(unlist(x[[dep]])),
            edgetype = dep
        )
        noNAorR <- with(
            resframe, !is.na(dependency) & !dependency %in% c("", "R")
        )
        resframe[noNAorR, ]
    })
    df <- dplyr::bind_rows(reslist)
    class(df) <- c('biocDepDF', class(df))
    df
}

#' Work with package dependencies as a graph
#'
#' Package dependencies represent a directed
#' graph (though Bioconductor dependencies are
#' not an acyclic graph). This function simply
#' returns an igraph graph from the package
#' dependency data frame from a call to
#' [buildPkgDependencyDataFrame()] or
#' any tidy data frame with rows of (Package, dependency)
#' pairs. Additional columns are added as igraph edge
#' attributes (see [igraph::graph_from_data_frame()]).
#'
#' @importFrom igraph graph_from_data_frame
#'
#' @param pkgDepDF a tidy data frame. See description for
#' details.
#'
#' @seealso See [buildPkgDependencyDataFrame()],
#' [igraph::graph_from_data_frame()],
#' [inducedSubgraphByPkgs()], [subgraphByDegree()],
#' [igraph::igraph-es-indexing],
#' [igraph::igraph-vs-indexing]
#'
#' @return An igraph directed graph. See the igraph
#' package for details of what can be done.
#'
#' @examples
#'
#' library(igraph)
#'
#' pkg_dep_df = buildPkgDependencyDataFrame()
#'
#' # at this point, filter or join to manipulate
#' # dependency data frame as you see fit.
#'
#' g = buildPkgDependencyIgraph(pkg_dep_df)
#' g
#'
#' # Look at nodes and edges
#' head(V(g)) # vertices
#' head(E(g)) # edges
#'
#' # subset graph by attributes
#'
#'
#' head(sort(degree(g, mode='in'), decreasing=TRUE))
#' head(sort(degree(g, mode='out'), decreasing=TRUE))
#'
#' @export
buildPkgDependencyIgraph <- function(pkgDepDF) {
    g = igraph::graph_from_data_frame(pkgDepDF)
    class(g) = c('biocDepGraph',class(g))
    g
}

#' Return a minimal subgraph based on package name(s)
#'
#' Find the subgraph induced by including
#' specific packages. The induced subgraph is
#' the graph that includes the named packages
#' and all edges connecting them. This is useful
#' for a developer, for example, to examine her packages
#' and their intervening dependencies.
#'
#' @param g an igraph graph, typically created by
#' [buildPkgDependencyIgraph()]
#'
#' @param pkgs `character()` vector of packages to
#' include. Package names not included in
#' the graph are ignored.
#'
#' @param pkg_color `character(1)` giving color of named
#' packages. Other packages in the graph that fall in
#' connecting paths will be colored as the igraph default.
#'
#' @importFrom igraph induced_subgraph V
#'
#' @examples
#' library(igraph)
#' g <- buildPkgDependencyIgraph(buildPkgDependencyDataFrame())
#' ## subgraph of only the first 10 packages maintained by Bioconductor
#' biocmaintained <- head(biocMaintained()[["Package"]], 10L)
#' g2 <- inducedSubgraphByPkgs(g, pkgs = biocmaintained)
#' g2
#' V(g2)
#'
#' plot(g2)
#'
#' ## subgraph of a package's strong Bioconductor package dependencies
#' maedeps <- unlist(pkgBiocDeps(
#'     "MultiAssayExperiment", which = "strong",
#'     recursive = TRUE, only.bioc = TRUE
#' ), use.names = FALSE)
#' g3 <- inducedSubgraphByPkgs(g, pkgs = maedeps)
#' plot(g3)
#'
#' ## same subgraph with networkD3::forceNetwork
#' library(networkD3)
#' wt <- cluster_walktrap(g3)
#' members <- membership(wt)
#' ndg3 <- igraph_to_networkD3(g3, group = members)
#' forceNetwork(
#'     Links = ndg3$links, Nodes = ndg3$nodes, Source = 'source',
#'     Target = 'target', NodeID = 'name', Group = 'group', zoom = TRUE,
#'     linkDistance = 200, fontSize = 20, opacity = 0.9, opacityNoHover = 0.9
#' )
#' @export
inducedSubgraphByPkgs <- function(g, pkgs, pkg_color='red') {
    pkgs = intersect(pkgs, names(igraph::V(g)))
    g2 = igraph::induced_subgraph(graph=g, vids=pkgs)
    igraph::V(g2)[pkgs]$color='red'
    g2
}

#' Subset graph by degree
#'
#' While the [inducedSubgraphByPkgs()]
#' returns the subgraph with the minimal connections
#' between named packages, this function takes a vector of
#' package names, a degree (1 or more) and returns the
#' subgraph(s) that are within `degree` of the
#' package named.
#'
#' @param g an igraph graph, typically created by
#' [buildPkgDependencyIgraph()]
#'
#' @param pkg `character(1)` package name from which to
#' measure degree.
#'
#' @param degree integer(1) degree, limit search for
#' adjacent vertices to this degree.
#'
#' @param ... passed on to [igraph::distances()]
#'
#' @importFrom igraph distances induced_subgraph V is_igraph
#'
#' @return An igraph graph, with only nodes and their
#' edges within degree of the named package
#'
#' @examples
#'
#' g <- buildPkgDependencyIgraph(buildPkgDependencyDataFrame())
#' g2 <- subgraphByDegree(g, 'GEOquery')
#' plot(g2)
#'
#' @export
subgraphByDegree <- function(g, pkg, degree=1, ...) {
    stopifnot(is.character(pkg) & pkg %in% names(igraph::V(g)) & length(pkg)==1)
    stopifnot(is_igraph(g))
    d = igraph::distances(graph=g, v=pkg, ...)
    d2 = d[1,d[1,]<= degree]
    igraph::induced_subgraph(graph=g, vids=names(d2))
}

#' Grab build report results from BUILD_STATUS_DB for a particular package range
#'
#' @param start `character(1)` alphabetically first package name in range
#'
#' @param end `character(1)` alphabetically last package name in range
#'
#' @param condition `character(1)` condition string, typically 'ERROR' or
#'     'WARNING'
#'
#' @param phase `character(1)` string for phase of event:
#'     'install', 'checksrc', or 'buildsrc' (default)
#'
#' @param version `character(1)` string indication Bioconductor version, either
#'     'devel' (default) or 'release'
#'
#' @author Vincent J. Carey
#'
#' @examples
#' \dontrun{
#' biocPkgRanges(
#'     start = "a4", end = "CMA",
#'     condition = "ERROR", version = "devel"
#' )
#' }
#' @export
biocPkgRanges <- function(
    start,
    end,
    condition = c("ERROR", "WARNINGS"),
    phase = "buildsrc",
    version = c("devel", "release")
) {
    condition <- match.arg(condition)
    version <- match.arg(version)

    sdat <- biocBuildStatusDB(version = version, pkgType = "software")

    cur <- which(sdat[["pkg"]] >= start & sdat[["pkg"]] <= end)
    if (!length(cur))
        stop("no packages in range")

    mine <- sdat[cur, ]

    crit <- mine[grep("OK|skipped", mine[["result"]], invert = TRUE), ]
    if (!nrow(crit))
        stop("all packages OK or skipped!")

    flag <- crit[grep(condition, crit[["result"]]), ]
    flagged <- flag[grep(phase, flag[["stage"]]), ]
    split(flagged, flagged[["pkg"]])
}

#' Tidy Bioconductor build report results
#'
#' The online Bioconductor build reports
#' are great for humans to look at, but
#' they are not easily computable. This function
#' scrapes HTML and text files available
#' from the build report online pages to generate
#' a tidy data frame version of the build report.
#'
#' @param version `character(1)` or `package_version` indicating the
#'   _Bioconductor_ version (e.g., "3.22"). Defaults to the "current version"
#'   provided by `BiocManager::version()`. The keywords "release" and "devel"
#'   are also accepted.
#'
#' @param pkgType `character()` The type of packages for which to get
#'   information for. Values can be one or more of the following:
#'   * `software`: Software packages
#'   * `data-experiment`: Experiment data packages
#'   * `data-annotation`: Annotation data packages
#'   * `workflows`: Workflow packages
#'
#' @param stage.timings `logical(1)` Whether to include the start, end, and
#' elapsed time for each build, check, install stage from each building in
#' the result (default: FALSE)
#'
#' @return A `tbl_df` object with columns pkg, version,
#' author, commit, date, node, stage, and result.
#'
#' @importFrom readr read_lines
#' @importFrom tibble as_tibble
#' @importFrom rvest html_text html_text2 html_nodes
#' @importFrom xml2 read_html
#' @importFrom dplyr left_join
#' @importFrom BiocManager version
#'
#' @examples
#'
#' # Set the stage--what version of Bioc am I using?
#' BiocManager::version()
#'
#' latest_build <- biocBuildReport()
#' head(latest_build)
#'
#' @export
biocBuildReport <- function(
    version = BiocManager::version(),
    pkgType = c(
        "software", "data-experiment", "data-annotation", "workflows"
    ),
    stage.timings = FALSE
) {
  stopifnot(is.logical(stage.timings), is.character(pkgType))
  if (version %in% c("release", "devel"))
    version <- BiocManager:::.version_bioc(version)

  version <- try(as.package_version(version), silent = TRUE)

  if (is(version, "try-error"))
    stop(
        "'version' is not 'release', 'devel', or a valid 'package_version'"
    )

  z <- biocBuildStatusDB(version, pkgType)

  if (version >= package_version("3.14")) {
    y <- biocBuildReportDB(version, pkgType)

    if (stage.timings) {
      typeTimings <- lapply(attr(y, "dcf_folder"), .read_summary_dcfs)
      timings <- do.call(rbind, unname(typeTimings))
      z <- merge(
          z, timings,
          by.x = c("pkg", "node", "stage"),
          by.y = c("Package", "node", "stage")
      )
    }
  } else {
    warning("Only 'software' provided for Bioconductor < 3.14")
    url <- .get_build_report_url(version, pkgType = "bioc")
    dat <- xml2::read_html(url)

    rowspan <- length(html_text(html_nodes(
        dat,xpath='/html/body/table[@class="node_specs"]/tr[@class!=""]'
    )))
    if(rowspan > 5L || rowspan < 2L){
      warning("Detected an unusual number of builders == ",rowspan," ... ")
    }
    res <- html_nodes(dat,
      xpath = '/html/body/table[@id="THE_BIG_GCARD_LIST"]') |>
      html_nodes("tr") |> html_nodes("td") |> html_nodes("b") |>
      html_nodes("a")
    pkgnames <- html_text(res)
    versions <- html_nodes(dat,
      xpath = '/html/body/table[@id="THE_BIG_GCARD_LIST"]') |>
      html_nodes(xpath = "//td[@rowspan=3]") |> html_nodes("b") |>
      html_text2()
    versions <- vapply(strsplit(versions, "\\s"), `[`, character(1L), 2L)
    maints <- html_nodes(dat,
      xpath = '//*[@id="THE_BIG_GCARD_LIST"]/tbody/tr/td[@rowspan=3]/text()'
    ) |> html_text2()
    # Account for packages with malformed maintainer fields in page
    idx <- which(rle(maints)$lengths > 1)
    if (length(idx)) {
        off_set <- seq(0, length(idx) - 1)
        idx <- idx + off_set
        for (i in off_set) {
          maints <- append(maints, values = " ", after = idx[i+1] + 1)
        }
    }
    maints <- maints[c(FALSE, TRUE)]
    stopifnot(identical(length(pkgnames), length(maints)))

    meta <- html_nodes(dat,
      xpath = '//*[@id="THE_BIG_GCARD_LIST"]/tbody/tr/td[@rowspan=3]')
    values <- meta |> html_nodes("table") |> html_text2()

    if (version >= package_version("3.13")) {
      values <- trimws(gsub("[\n]*git_last_commit[_date]*:", "", values))
      splitter <- "\\s"
    } else {
      values <- trimws(gsub("Last.Commit:|.Last.Changed.Date", "", values))
      splitter <- ": "
    }

    commitdate <- do.call(rbind.data.frame, strsplit(values, splitter))
    names(commitdate)[1:2] <- c("git_last_commit", "git_last_commit_date")
    commitdate[["git_last_commit_date"]] <-
        as.POSIXct(commitdate[["git_last_commit_date"]])

    y <- data.frame(
      pkg = pkgnames,
      author = maints,
      version = versions,
      commitdate[, 1:2]
    )
    y <- y[!is.na(y$pkg),]
  }

  depdf <- get_deprecated_status_df(version)
  isEmpty <- all(
      vapply(depdf, function(x) identical(length(x), 0L), logical(1L))
  )
  if (!isEmpty)
      y <- merge(y, depdf, by.x = "pkg", by.y = "Package")

  df <- suppressMessages(left_join(y, z)) # just suppress "Joining by...."
  df <- as_tibble(df)
  if (!nrow(df)) {
    warning("No Bioconductor build report found.")
    return(df)
  }
  df[['bioc_version']] <- as.character(version)
  attr(df,'git_last_commit_date') <- as.POSIXct(df[['git_last_commit_date']][1])
  attr(df,'class') = c('biocBuildReport',class(df))
  df
}

context("biocBuildReport")

# test release build report
test_that("columns are consistent in report output", {
    columnames <- c("pkg", "author", "version", "git_last_commit",
        "git_last_commit_date", "node", "stage", "PackageStatus")

    bioc3.12_build <- biocBuildReport("3.12")
    expect_true(all(columnames %in% colnames(bioc3.12_build)))
    expect_true(inherits(bioc3.12_build[["git_last_commit_date"]], "POSIXct"))
    expect_true(tibble::is_tibble(bioc3.12_build))

    bioc3.13_build <- biocBuildReport("3.13")
    expect_true(all(columnames %in% colnames(bioc3.13_build)))
    expect_true(inherits(bioc3.13_build[["git_last_commit_date"]], "POSIXct"))
    expect_true(tibble::is_tibble(bioc3.13_build))

    bioc3.14_build <- biocBuildReport("3.14")
    expect_true(all(columnames %in% colnames(bioc3.14_build)))
    expect_true(inherits(bioc3.14_build[["git_last_commit_date"]], "POSIXct"))
    expect_true(tibble::is_tibble(bioc3.14_build))
})

.get_cache <- function() {
    cache <- getOption("pkgToolsCache", setCache(verbose = FALSE))
    BiocFileCache::BiocFileCache(cache)
}

.getAnswer <- function(msg, allowed)
{
    if (interactive()) {
        repeat {
            cat(msg)
            answer <- readLines(n = 1)
            if (answer %in% allowed)
                break
        }
        tolower(answer)
    } else {
        "n"
    }
}

#' @name BiocPkgTools-cache
#'
#' @title Manage cache for BiocPkgTools
#'
#' @description Managing user data is important to allow use of email functions
#' such as `biocBuildEmail` and made easy with `BiocFileCache`.
#'
#' @section pkgToolsCache:
#' Get the directory location of the cache. It will prompt the user to create
#' a cache if not already created. A specific directory can be used via
#' `setCache`.
#'
#' @section setCache:
#' Specify the directory location of the data cache. By default, it will
#' got to the user's home/.cache/R and "appname" directory as specified by
#' \code{tools::R_user_dir} (with package="BiocPkgTools" and which="cache").
#'
#' @param directory The file location where the cache is located. Once set
#' future downloads will go to this folder.
#' @param verbose Whether to print descriptive messages
#' @param ask logical (default TRUE when interactive session) Confirm the file
#' location of the cache directory
#'
#' @param ... For `pkgToolsCache`, arguments are passed to `setCache`
#'
#' @export
setCache <-
    function(directory = tools::R_user_dir("BiocPkgTools", "cache"),
        verbose = TRUE,
        ask = interactive())
{
    stopifnot(is.character(directory),
        length(directory) == 1L, !is.na(directory))

    if (!dir.exists(directory)) {
        if (ask) {
            qtxt <- sprintf(
                "Create BiocPkgTools cache at \n    %s? [y/n]: ",
                directory
            )
            answer <- .getAnswer(qtxt, allowed = c("y", "Y", "n", "N"))
            if ("n" == answer)
                stop("'BiocPkgTools' directory not created. Use 'setCache'")
        }
        dir.create(directory, recursive = TRUE, showWarnings = FALSE)
    }
    options("pkgToolsCache" = directory)

    if (verbose)
        message("BiocPkgTools cache directory set to:\n    ",
            directory)
    invisible(directory)
}

#' @rdname BiocPkgTools-cache
#' @export
pkgToolsCache <- function(...) {
    getOption("pkgToolsCache", setCache(..., verbose = FALSE))
}

#' Look up a package's Bioconductor dependencies
#'
#' The function uses the `pkgType` argument to restrict the look up to only the
#' relevant Bioconductor repository. It works for multiple packages of the same
#' type.
#'
#' @inheritParams tools::package_dependencies
#' @inheritParams BiocManager::repositories
#'
#' @param pkg `character(1)` The package for which to look up dependencies.
#'
#' @param pkgType `character()` Any of 'software', 'data-experiment',
#'   'workflows', and / or 'data-annotation' (defaults to all)
#'
#' @param only.bioc `logical(1)` Whether to only return Bioconductor
#'   dependencies in the list (default `TRUE`)
#'
#' @examples
#'
#' pkgBiocDeps("MultiAssayExperiment", only.bioc = TRUE)
#'
#' pkgBiocDeps("MultiAssayExperiment", only.bioc = FALSE)
#'
#' @export
pkgBiocDeps <- function(
    pkg,
    pkgType = c(
        "software", "data-experiment", "workflows", "data-annotation"
    ),
    which = "strong",
    only.bioc = TRUE,
    recursive = FALSE,
    version = BiocManager::version()
) {
    pkgType <- match.arg(pkgType, several.ok = TRUE)
    repo.name <- .matchGetShortName(pkgType, "biocmanager.names")
    all_db <- utils::available.packages(repos = BiocManager::repositories())
    repo <- BiocManager:::.repositories_bioc(version)[repo.name]
    biocdb <- utils::available.packages(repos = repo)
    res <- tools::package_dependencies(
        pkg, db = all_db, which = which, recursive = recursive
    )
    if (only.bioc)
        lapply(res, function(pkglist) pkglist[pkglist %in% rownames(biocdb)])
    else
        res
}

#' Obtain all the reverse dependencies for a Bioconductor package
#'
#' The function returns a slightly upgraded list with dependency types as
#' elements and package names in each of those elements, if any. The
#' types of dependencies can be seen in the `which` argument documentation.
#'
#' @details The summary method of the `biocrevdeps` class given by
#'   `pkgBiocRevDeps` provides a tally in each dependency field.
#'
#' @inheritParams pkgBiocDeps
#'
#' @return A `biocrevdeps` list class object
#'
#' @examples
#' rdeps <- pkgBiocRevDeps("MultiAssayExperiment", which = "all")
#' rdeps
#' summary(rdeps)
#' @export
pkgBiocRevDeps <- function(
    pkg,
    pkgType = c(
        "software", "data-experiment", "workflows", "data-annotation"
    ),
    which = "strong",
    only.bioc = TRUE,
    version = BiocManager::version(),
    recursive = FALSE
) {
    whiches <- c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")
    names(whiches) <- whiches

    stopifnot(is.character(pkg), !is.na(pkg), nzchar(pkg))

    if (!all(which %in% c(whiches, "all", "strong", "most")))
        stop("'which' dependency type is invalid; see ?package_dependencies")

    lwhich <- switch(
        which,
        all = TRUE,
        strong = c(TRUE, TRUE, TRUE, FALSE, FALSE),
        most = c(TRUE, TRUE, TRUE, TRUE, FALSE),
        whiches %in% which
    )
    which <- whiches[lwhich]

    pkgType <- match.arg(pkgType, several.ok = TRUE)
    repo.name <- .matchGetShortName(pkgType, "biocmanager.names")
    all_db <- utils::available.packages(repos = BiocManager::repositories())
    repo <- BiocManager:::.repositories_bioc(version)[repo.name]
    biocdb <- utils::available.packages(repos = repo)

    if (isFALSE(recursive))
        res <- lapply(which, function(ofwhich) {
            tools::package_dependencies(
                pkg, all_db, reverse = TRUE, which = ofwhich, recursive = recursive
            )[[pkg]]
        })
    else
        res <- tools::package_dependencies(
            pkg, all_db, reverse = TRUE, which = which, recursive = recursive,
        )[[pkg]]

    if (only.bioc && (is.character(recursive) || recursive))
        res <- list(recursive = res[res %in% rownames(biocdb)])
    else if (only.bioc)
        res <- lapply(
            res, function(pkglist) pkglist[pkglist %in% rownames(biocdb)]
        )

    attributes(res) <- list(
        package = pkg, class = "biocrevdeps", which = which,
        only.bioc = only.bioc, recursive = recursive,
        names = names(res), version = version, reverse = TRUE
    )
    res
}

#' @rdname pkgBiocRevDeps
#'
#' @inheritParams base::summary
#'
#' @export
summary.biocrevdeps <- function(object, ...) {
    totals <- lapply(object, length)
    sum.df <- data.frame(
        totals,
        row.names = attr(object, "package"),
        only.bioc = attr(object, "only.bioc")
    )
    if (!"recursive" %in% names(object)) {
        tot.df <- data.frame(Total = sum(unlist(totals)))
        sum.df <- cbind.data.frame(sum.df, tot.df)
    }
    cat(
        "tools::package_dependencies(",
        "\n  ", dQuote(attr(object, "package"), FALSE), ",",
        "\n  db=utils::available.packages(repos=BiocManager::repositories()),",
        "\n  which=", capture.output(dput(unname(attr(object, "which")))), ",",
        "\n  recursive=", attr(object, "recursive"), ",",
        "\n  reverse=", attr(object, "reverse"), "\n)\n",
        sep = ""
    )
    sum.df
}

test_that("pkgBiocRevDeps(recursive = TRUE) returns only reverse deps", {
    skip_if_offline()
    ## Regression test for issue #81: Recursing reverse dependencies over "all"
    ## dependencies will return nearly the entire repository, including some
    ## forward dependencies of the queried package.
    pkg <- "Rarr"

    db <- utils::available.packages(repos = BiocManager::repositories())
    skip_if_not(pkg %in% rownames(db), "Rarr not available in repositories")

    ## Forward dependencies that are not themselves reverse dependencies should
    ## never appear in the recursive reverse-dependency result.
    fwd <- tools::package_dependencies(pkg, db, which = "all")[[pkg]]
    rev1 <- tools::package_dependencies(
        pkg, db, reverse = TRUE, which = "all"
    )[[pkg]]
    fwd_only <- setdiff(fwd, rev1)

    res <- pkgBiocRevDeps(
        pkg, which = "all", recursive = TRUE, only.bioc = FALSE
    )
    rec <- as.character(res)

    ## check that some forward dependencies are present in the recursive result
    expect_true(any(fwd_only %in% rec))

    ## check that the recursive result is not too small
    expect_gt(length(rec), nrow(db) / 2)
})

.CRAN_WEB_URL <- "https://cran.r-project.org/"

#' Check the CRAN build report page and email a notification
#'
#' The `CRANstatus` function allows users to check the status of a package
#' and send an email report of any failures.
#'
#' @inheritParams biocBuildEmail
#'
#' @importFrom rvest html_table
#'
#' @param to.mail The email of the CRAN report recipient
#'
#' @export
CRANstatus <-
    function(pkg, core.name = NULL, core.email = NULL, core.id = NULL,
        to.mail = "maintainer@bioconductor.org", dry.run = TRUE,
        emailTemplate = templatePath("cranreport")) {

    stopifnot(is.character(pkg), identical(length(pkg), 1L), !is.na(pkg))

    if (!requireNamespace("blastula", quietly = TRUE))
        stop("Install the 'blastula' package to send HTML emails or use\n",
            "  'textOnly=TRUE'")
    if (!requireNamespace("kableExtra", quietly = TRUE))
        stop("Install the 'kableExtra' package to create HTML tables")

    core.list <- .getUserInfo(core.email = core.email, core.name = core.name,
        core.id = core.id)
    core.name <- core.list[["core.name"]]
    core.email <- core.list[["core.email"]]
    core.id <- core.list[["core.id"]]


    URL <- paste0(.CRAN_WEB_URL, "/web/checks/check_results_", pkg, ".html")

    htmlfile <- .cache_url_file(URL)
    html <- xml2::read_html(htmlfile)
    cran_results <- rvest::html_table(html)[[1L]]

    status <- cran_results[["Status"]]

    mail <- paste0(readLines(emailTemplate), collapse = "\n")
    maildate <- format(Sys.time(), "%B %d, %Y")
    send <- sprintf(mail, pkg, maildate, pkg, pkg, "cran_results", URL)

    if ("ERROR" %in% toupper(status)) {

        titletext <- paste0(pkg, " has at least one ERROR")
        tfile <- tempfile(fileext = ".Rmd")
        writeLines(send, tfile)
        print(repmail <- blastula::render_email(tfile))
        if (!dry.run) {
            blastula::smtp_send(email = repmail,
                from = core.email, to = to.mail, subject = titletext,
                credentials = blastula::creds(
                    user = paste0(core.id, "@roswellpark.org"),
                    provider = "office365",
                    sender_name = core.name,
                    use_ssl = FALSE
                )
            )
            message("Email sent to: ", to.mail)
        }
        c(OK = FALSE)

    } else {

        c(OK = TRUE)
    }
}

context("CRANstatus")

test_that("CRANstatus works", {
    ## test a package that is likely to be OK
    res <- CRANstatus("BiocManager", dry.run = TRUE)
    expect_type(res, "logical")
    expect_true(res)
})

#' Explore Bioconductor packages interactively
#'
#' @description Explore Bioconductor packages through an interactive bubble
#'     plot. Click on bubbles to bring up additional information about the
#'     package. Size and proximity to center of a bubble is based on the
#'     downloads the package has in the past month.
#'
#' @param top maximum number of packages displayed in any biocView
#'
#' @param ... parameters passed to \code{htmlwidgets::createWidget()}
#'
#' @importFrom htmlwidgets createWidget
#'
#' @return A bubble plot of Bioconductor packages
#'
#' @export
biocExplore <- function(top = 500L, ...) {
    if(!is.numeric(top) || top < 1){
        stop("top must be >= 1")
    }
    # instruction messages
    message("- Hover over bubbles to see full name and lifetime downloads")
    message("- Click on bubbles to see more information about package")
    message("- Use filter to filter by biocViews")

    data <- list(
        data = get_bioc_data(),
        top = top
    )

    settings <- list(
    )

    x <- list(
        data = data,
        settings = settings
    )

    # create the widget
    # create widget
    htmlwidgets::createWidget(
        name = 'bioc_explore',
        package = 'BiocPkgTools',
        x = x,
        ...
    )
}

#' Generate needed information to create DOI from a package
#' directory.
#'
#' @param dir `character(1)` Path to package
#'
#' @import utils
#' @import tools
#'
#' @return A \code{data.frame}
#'
#' @keywords Internal
#'
#' @export
getPackageInfo <- function(dir){
    if(!dir.exists(file.path(dir)))
        stop('dir "',dir,'" does not exist.')
    if(!file.exists(file.path(dir, "DESCRIPTION")))
        stop('dir "',dir,'" does not contain a DESCRIPTION file.')
    # the next line avoids warnings about unexported
    # functions. see: https://github.com/rstudio/learnr/commit/8cf2ad60993108079f38123300b23d96ef58bd4c
    db <- asNamespace('tools')$.read_description(file.path(dir, "DESCRIPTION"))
    Package <- unname(db["Package"])
    aar <- db["Authors@R"]
    if (is.na(aar)) {
        Author <- unname(db["Author"])
    } else {
        # the next line avoids warnings about unexported
        # functions. see: https://github.com/rstudio/learnr/commit/8cf2ad60993108079f38123300b23d96ef58bd4c
        aut <- asNamespace('utils')$.read_authors_at_R_field(aar)
        vl <- format(aut, include = c("given", "family"))
        Author <- paste(unlist(vl), collapse=", ")
    }
    cbind(Package, Author)
}

context("getPackageInfo")
library(BiocPkgTools)

test_that("getPackageInfo", {
    expect_error(getPackageInfo("../test"),
                 'dir "../test" does not exist')
    expect_error(getPackageInfo("../"),
                 'dir "../" does not contain')
    actual <- getPackageInfo(system.file(package = "BiocManager"))
    expect_true(is.matrix(actual))
    expect_type(actual,"character")
    expect_equal(colnames(actual),c("Package","Author"))
})

test_that("getPackageInfo() parses Authors@R", {
    fl <- tempfile()
    dir.create(fl)

    writeLines(c(
        "Package: test",
        "Authors@R: person('Im A.', 'Author')"
    ), file.path(fl, "DESCRIPTION"))
    expect_identical(getPackageInfo(fl)[,"Author"], c(Author = "Im A. Author"))

    writeLines(c(
        "Package: test",
        "Authors@R: person('Iman', 'Author')"
    ), file.path(fl, "DESCRIPTION"))
    expect_identical(getPackageInfo(fl)[,"Author"], c(Author = "Iman Author"))

    writeLines(c(
        "Package: test",
        "Authors@R: c(person('Iman', 'Author'), person('Im A.', 'Author'))"
    ), file.path(fl, "DESCRIPTION"))
    aut <- getPackageInfo(fl)[,"Author"]
    expect_identical(aut, c(Author = "Iman Author, Im A. Author"))
})

test_that(".extract_fnd() returns NA for no funder", {
    expect_identical(
        BiocPkgTools:::.extract_fnd(NA_character_),
        NA_character_
    )
    expect_identical(
        BiocPkgTools:::.extract_fnd(""),
        NA_character_
    )
    expect_identical(
        BiocPkgTools:::.extract_fnd("Jane Doe [aut, cre]"),
        NA_character_
    )
})

test_that(".extract_fnd() extracts single funder", {
    result <- BiocPkgTools:::.extract_fnd(
        "Jane Doe [aut], Big Funder [fnd]"
    )
    expect_identical(result, "Big Funder [fnd]")
})

test_that(".extract_fnd() extracts multiple funders", {
    result <- BiocPkgTools:::.extract_fnd(
        "Jane Doe [aut], Funder One [fnd], Funder Two [fnd]"
    )
    expect_length(result, 2L)
    expect_identical(result, c("Funder One [fnd]", "Funder Two [fnd]"))
})

test_that(".extract_fnd() extracts funders from multiline Author field", {
    author <- paste0(
        "Martin Morgan [aut, cre],\n",
        "    Chan Zuckerberg Initiative DAF CZF2019-002443 [fnd],\n",
        "    NIH NCI ITCR U24CA180996 [fnd]"
    )
    result <- BiocPkgTools:::.extract_fnd(author)
    expect_length(result, 2L)
    expect_identical(result[[1L]], "Chan Zuckerberg Initiative DAF CZF2019-002443 [fnd]")
    expect_identical(result[[2L]], "NIH NCI ITCR U24CA180996 [fnd]")
})

test_that(".extract_fnd() returns NA when no [fnd] tag present", {
    expect_identical(
        BiocPkgTools:::.extract_fnd("some text without role brackets"),
        NA_character_
    )
})

get_build_status_db_url <- function(version, pkgType) {
  db_prefix <- if (package_version(version) >= '3.13') "BUILD_" else ""
  db_file <- paste0(db_prefix, "STATUS_DB.txt")
  sprintf(
    'https://bioconductor.org/checkResults/%s/%s-LATEST/%s',
    version, pkgType, db_file
  )
}

.get_build_report_tgz_url <- function(version, pkgType) {
    url <- .get_build_report_url(version = version, pkgType = pkgType)
    paste0(url, "/report.tgz")
}

.get_build_report_url <- function(version, pkgType) {
    sprintf(
        'https://bioconductor.org/checkResults/%s/%s-LATEST',
        version, pkgType
    )
}

get_deprecated_status_df <- function(version) {
    viewsfile <- .get_VIEWS(version = version, type = "BioCsoft")

    if (nrow(viewsfile))
        depdf <- cbind.data.frame(
            Package = viewsfile[, "Package"],
            Deprecated = viewsfile[, "PackageStatus"] == "Deprecated" &
                !is.na(viewsfile[, "PackageStatus"]),
            PackageStatus = viewsfile[, "PackageStatus"]
        )
    else
        depdf <- data.frame(
            Package = character(0L),
            Deprecated = logical(0L),
            PackageStatus = character(0L)
        )

    depdf
}

.read_summary_dcfs <- function(dcf_location) {
    summary_dcfs <- list.files(dcf_location, pattern="-summary\\.dcf$",
        full.names = TRUE, recursive = TRUE)
    fields <- c("Package", "StartedAt", "EndedAt", "EllapsedTime")
    summaries <- lapply(summary_dcfs, .import_dcf_stage_node, fields = fields)
    as.data.frame(do.call(rbind, summaries))
}

.import_dcf_stage_node <- function(filepath, fields) {
    stage <- head(strsplit(basename(filepath), "-", fixed = TRUE)[[1L]], 1L)
    node <- basename(dirname(filepath))
    dcf_pkg <- read.dcf(filepath, fields = fields)
    dcf_chr <- structure(as.character(dcf_pkg), .Names = fields)
    append(dcf_chr, c(node = node, stage = stage), after = 1L)
}

.read_info_dcfs <- function(info_files_location) {
    dcffiles <- list.files(path = info_files_location,
        pattern = "info\\.dcf$", full.names = TRUE, recursive = TRUE)
    meta <- do.call(rbind.data.frame, lapply(dcffiles, read.dcf))
    y <- meta[,
        c("Package", "Maintainer", "Version",
        "git_last_commit", "git_last_commit_date")
    ]
    y[["git_last_commit_date"]] <-
        as.POSIXct(gsub("^(.*)\\s\\(.*", "\\1", y[["git_last_commit_date"]]))
    names(y)[1:3] <- c("pkg", "author", "version")
    y
}

.cache_url_file <- function(url) {
    bfc <- BiocFileCache()
    bquery <- bfcquery(bfc, url, "rname", exact = TRUE)
    needsUpdate <- bfcneedsupdate(bfc, bquery[["rid"]])
    if (
        identical(nrow(bquery), 1L) && (is.na(needsUpdate) || needsUpdate)
    )
        bfcdownload(
            x = bfc, rid = bquery[["rid"]], ask = FALSE
        )

    bfcrpath(
        bfc, rnames = url, exact = TRUE, download = TRUE, rtype = "web"
    )
}

.url_exists <- function(url) {
    response <- httr::HEAD(url)
    status <- httr::status_code(response) < 300
    if (!status)
        warning(url, " cannot be reached", call. = FALSE)
    status
}

.nameCut <- function(string) {
    gsub("\n", " ",
        gsub("(.*)\\s+<.*", "\\1", string),
    fixed = TRUE
    )
}

.emailCut <- function(string) {
    gsub(" at ", "@",
        gsub("(.*)<(.*)>.*", "\\2", string)
    )
}

#' Obtain the location of available email templates
#'
#' These templates are used with `biocBuildEmail` to notify maintainers
#' regarding package errors and final deprecation warning.
#'
#' @param type `character(1)` Either one of "buildemail", "deprecation",
#'   "deprecguide", "cranreport", or "revdepnote". See the templates in the
#'   `resources` folder.
#'
#'
#' @export
templatePath <- function(
    type = c(
        "buildemail", "deprecation", "deprecguide", "cranreport", "revdepnote"
    )
) {
    type <- match.arg(type)
    type <- switch(
        type,
        buildemail = "BiocBuildEmail_Template.Rmd",
        cranreport = "CRANReport_Template.Rmd",
        deprecation = "Deprecation_Template.Rmd",
        deprecguide = "Deprecation_Guidelines.Rmd",
        revdepnote = "Deprecation_RevDeps.Rmd"
    )
    system.file(
        package = "BiocPkgTools", "resources", type,
        mustWork = TRUE
    )
}

.getUserInfo <- function(core.name, core.email, core.id) {
    bfc <- .get_cache()
    rid <- BiocFileCache::bfcquery(bfc, "user.info", "rname", exact = TRUE)$rid
    if (!length(rid))
        userfile <- BiocFileCache::bfcnew(bfc, "user.info", ext = ".txt")
    else
        userfile <- BiocFileCache::bfcrpath(bfc, rids = rid)

    if (!file.exists(userfile)) {
        if (is.null(core.name))
            core.name <- readline("Provide your full name: ")
        if (is.null(core.email))
            core.email <- readline("What is your core-team email? ")
        if (is.null(core.id))
            core.id <- readline(paste("What is your employee ID",
                "(e.g., AB12345, leave blank if N/A)? "))

        writeLines(c(
            as.character(core.name), as.character(core.email),
            as.character(core.id)
        ), con = userfile)

        message("Saved data to: ", pkgToolsCache())
    } else {
        devinfo <- readLines(userfile)
        core.name <- devinfo[[1L]]
        core.email <- devinfo[[2L]]
        core.id <- devinfo[[3L]]
    }
    list(core.name = core.name, core.email = core.email, core.id = core.id)
}

.getMailLog <- function() {
    bfc <- .get_cache()
    rid <- BiocFileCache::bfcquery(bfc, "email.log", "rname", exact = TRUE)$rid
    if (!length(rid)) {
        logpath <- BiocFileCache::bfcnew(bfc, "email.log", ext = ".rda")
        metainfo <- data.frame(maintainer = character(), email = character(),
            package = character(), dateSent = character(), times = integer(),
            stringsAsFactors = FALSE)
        save(metainfo, file = logpath)
    } else {
        logpath <- BiocFileCache::bfcrpath(bfc, rids = rid)
    }
    logpath
}

.checkEntry <- function(logloc, mainname, mainemail, pkg, date, dry.run) {
    if (dry.run) return(FALSE)
    mailinfo <- data.frame(maintainer = mainname, email = mainemail,
        package = pkg, dateSent = date, times = 1L, stringsAsFactors = FALSE)

    dataenv <- new.env(parent = emptyenv())
    load(logloc, dataenv)
    metainfo <- dataenv[["metainfo"]]
    newinfo <- do.call(rbind.data.frame, list(metainfo, mailinfo))

    anyDuplicated(newinfo[, names(newinfo) != "times"])
}

.addEntry <- function(logloc, mainname, mainemail, pkg, date, resend) {
    mailinfo <- data.frame(maintainer = mainname, email = mainemail,
        package = pkg, dateSent = date, times = 1L, stringsAsFactors = FALSE)

    dataenv <- new.env(parent = emptyenv())
    load(logloc, dataenv)
    metainfo <- dataenv[["metainfo"]]

    impcols <- names(metainfo)[names(metainfo) != "times"]
    bound <- do.call(rbind.data.frame, list(metainfo, mailinfo))

    if (resend) {
        dup <- duplicated(bound[, impcols], fromLast = TRUE)
        if (!any(dup))
            stop("'resend' used in the wrong context")
        dup <- dup[-length(dup)]
        metainfo[dup, "times"] <- metainfo[dup, "times"] + 1L
    } else {
        metainfo <- do.call(rbind.data.frame, list(metainfo, mailinfo))
    }

    save(metainfo, file = logloc)
}

#' @rdname biocBuildEmail
#'
#' @title Create and copy e-mail package notification template to clipboard
#'
#' @description The `biocBuildEmail` function provides a template for notifying
#' maintainers of errors in the Bioconductor Build System (BBS). This
#' convenience function returns the body of the email from a template within the
#' package and provides a copy in the clipboard.
#'
#' @param pkg `character(1)` The name of the package in trouble
#'
#' @param version `character()` A vector indicating which version of Bioconductor
#'     the package is failing in (either 'release' or 'devel'; defaults to both)
#'
#' @param PS `character(1)` Postscript, an additional note to the recipient of
#'     the email (i.e., the package maintainer)
#'
#' @param dry.run `logical(1)` Display the email without sending to the recipient.
#'     It only works for HTML email reports and ignored when `textOnly=TRUE`
#'
#' @param to `character()` A vector of email addresses serving as primary
#'   recipients for the message. For secondary recipients, use the `cc` and
#'   `bcc` arguments.
#'
#' @param cc,bcc `character()` A vector of email addresses for sending the message
#'   as a carbon copy or blind carbon copy.
#'
#' @param emailTemplate `character(1)` The path to the email template Rmd file as
#'     obtained by `templatePath()`. A custom template can be provided as file
#'     path.
#'
#' @param core.name `character(1)` The full name of the core team member
#'
#' @param core.email `character(1)` The Roswell Park email of the core team
#'     member
#'
#' @param core.id `character(1)` The internal identifier for the Roswell employee.
#'     This ID usually matches `^[A-Z]{2}[0-9]{5}` for more recent identifiers.
#'
#' @param textOnly `logical(1)` Whether to return the text of the email only.
#'     This avoids the use of the 'blastula' package and adds the text to the
#'     system clipboard if the `clipr` package is installed (default: FALSE)
#'
#' @param resend `logical(1)` Whether to force a resend of the email
#'
#' @param verbose `logical(1)` Whether to output full email information from
#'     'smtp_send' (when `dry.run` is `FALSE` and 'blastula' is installed)
#'
#' @param credFile `character(1)` An optional file generated by the
#'     `blastula::create_smtp_creds_file` function containing email
#'      authentication information (default: "~/.blastula_creds"). See
#'      `?biocBuildEmail` details.
#'
#' @details The `credFile` argument is a convenience for avoiding password entry
#'     at every instance an email is sent. If the default file
#'     `~/.blastula_creds` does not exist, the user will be prompted for
#'     authorization information. Currently it is configured to emails for the
#'     core-team:
#'     \preformatted{
#'         blastula::create_smtp_creds_file(
#'             file = "~/.blastula_creds",
#'             user = "user.email@domain.org",
#'             host = "smtp.office365.com",
#'             port = 587,
#'             use_ssl = TRUE
#'         )
#'     }
#'
#' @return A character string of the email
#'
#' @export
biocBuildEmail <-
    function(pkg, version = c("release", "devel"), PS = character(1L),
        dry.run = TRUE, to = NULL, cc = NULL, bcc = NULL,
        emailTemplate = templatePath(),
        core.name = NULL, core.email = NULL, core.id = NULL,
        textOnly = FALSE, resend = FALSE, verbose = FALSE,
        credFile = "~/.blastula_creds")
{
    stopifnot(
        is.character(pkg), identical(length(pkg), 1L),
        is.character(PS), identical(length(PS), 1L),
        !is.na(pkg), !is.na(PS), !is.na(core.name), !is.na(core.email),
        !is.na(core.id)
    )
    if (!file.exists(emailTemplate))
        stop("'emailTemplate' file not found.")

    if (!textOnly) {
        if (!requireNamespace("blastula"))
            stop("Install the 'blastula' package to send HTML emails or use\n",
                "  'textOnly=TRUE'")
    }

    core.list <- .getUserInfo(core.name, core.email, core.id)
    core.name <- core.list[["core.name"]]
    core.email <- core.list[["core.email"]]
    core.id <- core.list[["core.id"]]

    info_version <- tail(version, 1L)
    info_version <- BiocManager:::.version_bioc(info_version)
    listall <- biocPkgList(version = info_version)
    pkgMeta <- listall[listall[["Package"]] %in% pkg, ]
    if (!nrow(pkgMeta))
        stop("No pkg '",pkg,"' found on Bioconductor for ",
            "version '",version,"'")

    mainInfo <- pkgMeta[["Maintainer"]][[1L]]

    mainName <- unname(vapply(mainInfo, .nameCut, character(1L)))
    if (is.null(to))
        to <- unname(vapply(mainInfo, .emailCut, character(1L)))

    if (length(mainName) > 1L)
        mainName <- paste0(mainName, collapse = " & ")

    if (length(version) == 2L)
        vers <- paste(version, collapse = " and ")
    else
        vers <- version

    repolink <- vapply(version, function(vername) {
        sprintf("https://bioconductor.org/checkResults/%s/bioc-LATEST/%s/",
            vername, pkg)
    }, character(1L))
    repolink <- paste0(repolink, collapse = "\n")

    if (nchar(PS))
        PS <- paste0("**P.S.** ", PS)

    firstname <- vapply(strsplit(core.name, "\\s"), `[`, character(1L), 1L)
    mail <- paste0(readLines(emailTemplate), collapse = "\n")
    maildate <- format(Sys.time(), "%B %d, %Y")
    send <- sprintf(mail, pkg, core.name, maildate, pkg, mainName, pkg,
        vers, repolink, firstname, PS)

    title <- sprintf("%s Bioconductor package", pkg)
    logfile <- .getMailLog()
    sent_status <- .checkEntry(
        logfile, mainName, to, pkg, maildate, dry.run
    )
    sendagain <- (sent_status && resend)

    if (dry.run)
        message("Message not sent: Set 'dry.run=FALSE'")

    if (textOnly) {
        send <- strsplit(send, "---")[[1L]][[4L]]
        send <- paste(to, title, send, sep = "\n")
        if (requireNamespace("clipr", quietly=TRUE) &&
            clipr::clipr_available())
        {
            clipr::write_clip(send)
            message("Message copied to clipboard")
        }
    } else {
        tfile <- tempfile(fileext = ".Rmd")
        writeLines(send, tfile)
        send <- blastula::render_email(tfile)
        if (!dry.run && (!sent_status || sendagain)) {
            blastula::smtp_send(email = send,
                from = core.email, to = to, cc = cc, bcc = bcc, subject = title,
                credentials =
                    if (file.exists(credFile)) {
                        blastula::creds_file(credFile)
                    } else {
                        blastula::creds(
                            user = paste0(core.id, "@roswellpark.org"),
                            host = "smtp.office365.com",
                            port = 587,
                            use_ssl = TRUE
                        )
                    }, verbose = verbose
            )
            .addEntry(logfile, mainName, to, pkg, maildate, sendagain)
        }
    }
    send
}

#' @name biocBuildEmail
#'
#' @section sentHistory: Check the history of emails sent
#'
#' @export
sentHistory <- function() {
    bfc <- .get_cache()
    ## first build data.frame logger
    rid <- BiocFileCache::bfcquery(bfc, "email.log", "rname", exact = TRUE)$rid
    if (!length(rid))
        stop("No log available. Send some emails.")

    minienv <- new.env(parent = emptyenv())
    load(BiocFileCache::bfcrpath(bfc, rids = rid), envir = minienv)
    minienv[["metainfo"]]
}

#' Parse the Build Report tarball for a Bioconductor release
#'
#' @description This function parses the Build Report tarball for a Bioconductor
#'   release. By default it will pull all the `report.tgz` files for each
#'   Bioconductor package type. The Bioconductor Build System (BBS) Build Report
#'   tarball contains build status information for all packages in a
#'   Bioconductor release. This function is mainly used by [biocBuildReport()].
#'
#' @inheritParams biocBuildStatusDB
#' @inheritParams biocBuildReport
#'
#' @export
biocBuildReportDB <- function(
    version = BiocManager::version(),
    pkgType = c("software", "data-experiment", "data-annotation", "workflows"),
    stage.timings = FALSE
) {
    if (version %in% c("release", "devel"))
        version <- BiocManager:::.version_bioc(version)

    version <- try(as.package_version(version), silent = TRUE)

    if (is(version, "try-error"))
        stop(
            "'version' is not 'release', 'devel', or a valid 'package_version'"
        )

    pkgType <- match.arg(pkgType, several.ok = TRUE)
    pkgType <- .matchGetShortName(pkgType, "stat.url")

    urls <- .get_build_report_tgz_url(version, pkgType)
    names(urls) <- pkgType
    urls <- Filter(.url_exists, urls)

    tempfolders <- replicate(length(urls), tempfile())

    dcf_tables <- Map(
        function(url, tempfolder) {
            treport <- .cache_url_file(url)
            untar(treport, exdir = tempfolder)
            .read_info_dcfs(tempfolder)
        }, url = urls, tempfolder = tempfolders
    )
    report_table <- do.call(rbind.data.frame, unname(dcf_tables))
    report_table[["pkgType"]] <- rep(names(urls), vapply(dcf_tables, nrow, 1L))
    attr(report_table, "dcf_folder") <- tempfolders
    report_table
}

utils::globalVariables(".from")

#' Retrieve Class relationships
#'
#' As the title says it should do something with class relationships
#'
#' @param class a single `character` value defining a \sQuote{S4} class
#'   name
#' @param includeUnions `TRUE` or `FALSE`: Should union definitions
#'   included in the result? (default: `FALSE`)
#' @param data a \code{data.frame} with compatible columns. See output of
#'   `buildClassDepData`
#' @param g an `igraph` object with compatible edge attributes. See output
#'   of `buildClassDepGraph`
#' @param pkg a single `character` value defining a package name
#'
#' @name class-dependencies
#'
#' @importFrom methods getClass is
#' @importFrom igraph graph_from_data_frame
#'
#' @export
#'
#' @examples
#' library("SummarizedExperiment")
#' depData <- buildClassDepData("RangedSummarizedExperiment")
#' depData
#' g <- buildClassDepGraph("RangedSummarizedExperiment")
#' plotClassDepGraph(g)
buildClassDepGraph <- function(class, includeUnions = FALSE) {
    # input check
    if(!is.logical(includeUnions) && length(includeUnions) != 1L &&
       !is.na(includeUnions)){
        stop("'includeUnions' must be TRUE or FALSE.", call. = FALSE)
    }
    if(!is.character(class) && length(class) != 1L &&
       !is.na(class)){
        stop("'class' must be a single character value.", call. = FALSE)
    }
    #
    data <- buildClassDepData(class = class, includeUnions = includeUnions)
    g <- igraph::graph_from_data_frame(data)
    g
}

#' @rdname class-dependencies
#'
#' @export
buildClassDepData <- function(class, includeUnions = FALSE) {
    # input check
    if(!is.logical(includeUnions) && length(includeUnions) != 1L &&
       !is.na(includeUnions)){
        stop("'includeUnions' must be TRUE or FALSE.", call. = FALSE)
    }
    if(!is.character(class) && length(class) != 1L &&
       !is.na(class)){
        stop("'class' must be a single character value.", call. = FALSE)
    }
    #
    classData <- .get_s_class_data(class)
    slotClassData <- .get_s_slot_class_data(class)
    data <- rbind(classData,slotClassData)
    if(!includeUnions){
        data <- data[data$childUnion == FALSE,]
    }
    data
}

.get_s_class_data <- function(class){
    classDef <- getClass(class)
    classData <- .get_s_extends_data(classDef@contains)
    classData <- classData[!vapply(classData,is.null,logical(1))]
    if(!is.null(classData)){
        classData$type <- "object"
    }
    classData
}

.get_s_slot_class_data <- function(class){
    classDef <- getClass(class)
    slotClassDefs <- lapply(classDef@slots, getClass)
    # check which slot is a union
    isUnion <- vapply(slotClassDefs,is,logical(1),"ClassUnionRepresentation")
    #
    notUnionSlotClassDefs <- lapply(slotClassDefs[!isUnion],
                                    function(scd){ scd@contains })
    FUN <- function(scd){
        scd <- getClass(scd)
        scd <- scd@subclasses
        scd <- scd[vapply(scd,function(e){e@distance == 1L},logical(1))]
        scd <- lapply(names(scd), getClass)
        scd <- scd[!vapply(scd,function(e){e@className == "NULL"},logical(1))]
        lapply(scd,function(e){e@contains})
    }
    unionSlotClassDefs <- lapply(classDef@slots[isUnion], FUN)
    slotClassDefs <- c(notUnionSlotClassDefs,
                       do.call(c,unionSlotClassDefs))
    if(length(slotClassDefs) == 0L){
        return(NULL)
    }
    # create class -> slot link
    child <- vapply(slotClassDefs,
                    function(scd){
                        if(length(scd) == 0L){
                            return(NA_character_)
                        }
                        scd[[1L]]@subClass
                    }, character(1))
    childVirtual <- .is_s_virtual(child)
    parentUnion <- is(classDef,"ClassUnionRepresentation")
    childUnion <- vapply(slotClassDefs,is,logical(1),"ClassUnionRepresentation")
    classData <- .get_s_extends_data.frame(class, child,
                                           parentVirtual = classDef@virtual,
                                           parentUnion = parentUnion,
                                           childVirtual = childVirtual,
                                           childUnion = childUnion)
    classData <- unique(classData)
    # get slot class data
    slotClassData <- lapply(slotClassDefs,.get_s_extends_data)
    # merge results
    slotClassData <- do.call(rbind,c(list(classData),slotClassData))
    rownames(slotClassData) <- NULL
    slotClassData$type <- "slot"
    slotClassData
}

.get_s_extends_data <- function(ext){
    if(length(ext) == 0L){
        return(NULL)
    }
    ext <- ext[vapply(ext,function(e){e@distance == 1L},logical(1))]
    if(length(ext) > 0L){
        res <- .get_s_extends_element(ext)
        res <- do.call(rbind,
                       c(list(res),
                         lapply(res$child, .get_s_class_data)))
    } else {
        res <- data.frame(parent = c(), child = c(), type = c(),
                          parentVirtual = c(), childVirtual = c(),
                          union = c())
    }
    res
}

.is_s_virtual <- function(x){
    unname(vapply(x,
                  function(e){
                      if(is.na(e)){
                          return(FALSE)
                      }
                      getClass(e)@virtual
                  },logical(1)))
}
.is_s_union <- function(x){
    unname(vapply(x,
                  function(e){
                      is(getClass(e),"ClassUnionRepresentation")
                  },logical(1)))
}

.get_s_extends_element <- function(ext){
    parent <- unname(vapply(ext,function(e){as.character(e@subClass)},character(1)))
    child <- unname(vapply(ext,function(e){as.character(e@superClass)},character(1)))
    parentVirtual <- .is_s_virtual(parent)
    childVirtual <- .is_s_virtual(child)
    parentUnion <- .is_s_union(parent)
    childUnion <- .is_s_union(child)
    .get_s_extends_data.frame(parent = parent, child = child,
                              parentVirtual = parentVirtual,
                              parentUnion = parentUnion,
                              childVirtual = childVirtual,
                              childUnion = childUnion)
}

.get_s_extends_data.frame <- function(parent = "", child = "", type = "",
                                      parentVirtual = FALSE,
                                      parentUnion = FALSE,
                                      childVirtual = FALSE,
                                      childUnion = FALSE){
    data.frame(parent = parent,
               child = child,
               type = type,
               parentVirtual = parentVirtual,
               parentUnion = parentUnion,
               childVirtual = childVirtual,
               childUnion = childUnion)
}

#' @rdname class-dependencies
#' @export
buildClassDepFromPackage <- function(pkg, includeUnions = FALSE){
    exports <- getNamespaceExports(pkg)
    classNames <- exports[grepl(".__C__",exports)]
    classNames <- gsub(".__C__","",classNames)
    data <- lapply(classNames, buildClassDepData, includeUnions = includeUnions)
    names(data) <- classNames
    data
}

#' @rdname class-dependencies
#' @export
plotClassDep <- function(class, includeUnions = FALSE) {
    plotClassDepGraph(buildClassDepGraph(class, includeUnions = includeUnions))
}
#' @rdname class-dependencies
#'
#' @importFrom igraph graph_from_data_frame
#'
#' @export
plotClassDepData <- function(data) {
    plotClassDepGraph(igraph::graph_from_data_frame(data))
}
#' @rdname class-dependencies
#'
#' @importFrom igraph E V degree layout_with_kk edge_attr V<- E<-
#' @importFrom graphics strheight strwidth
#' @importFrom grDevices extendrange
#'
#' @export
plotClassDepGraph <- function(g) {
    # input check
    if(!all(require_edge_attr %in% names(edge_attr(g)))){
        stop("'g' must have the following edge attributes: '",
             paste(require_edge_attr, collapse = "', '"),"'")
    }
    # common
    V(g)$size <- nchar(names(V(g)))
    V(g)$label.family <- "sans"
    V(g)$label <- V(g)$name
    # set vertex color
    from <- function(){}
    color <- vapply(names(V(g)),
                    function(name){
                        virtual <- unique(E(g)[ .from(name) ]$parentVirtual)
                        if(length(virtual) == 0 || virtual) {
                            "#1a81c2"
                        } else {
                            "#87b13f"
                        }
                    },
                    character(1))
    V(g)$color <- color
    label.color <- color
    label.color[label.color == "#1a81c2"] <- "white"
    label.color[label.color == "#87b13f"] <- "black"
    V(g)$label.color <- label.color
    # set edge color
    color <- E(g)$type
    color[color == "object"] <- "orange"
    color[color == "slot"] <- "gray"
    E(g)$color <- color
    # set edge type
    lty <- E(g)$type
    lty[lty == "object"] <- "solid"
    lty[lty == "slot"] <- "dotted"
    E(g)$lty <- lty
    # other edge settings
    E(g)$width <- 2
    # root
    root <- which(degree(g, v = V(g), mode = "in")==0, useNames = T)
    V(g)[root]$color <- "red"
    # plot
    co <- layout_with_kk(g)
    plot(0, type="n", ann=FALSE, axes=FALSE, xlim = extendrange(co[,1]),
         ylim=extendrange(co[,2]))
    plot(g, layout = co, rescale=FALSE, add=TRUE,
         vertex.shape="rectangle",
         vertex.size = (strwidth(V(g)$label) + strwidth("oo")) * 100,
         vertex.size2 = strheight("I") * 2 * 100)
}

require_edge_attr <- c("type",
                      "parentVirtual", "parentUnion",
                      "childVirtual", "childUnion")

#' Generate a DOI for a Bioconductor package
#'
#' This function makes calls out to the REST API (v2) described
#' here: \url{https://support.datacite.org/docs/api-create-dois}.  This
#' function creates a new DOI for a Bioconductor package (cannot already
#' exist). The target URL for the DOI is the short Bioconductor
#' package URL.
#'
#' The login information for the "real" Bioconductor account
#' should be stored in the environment variables "EZID_USERNAME"
#' and "EZID_PASSWORD".
#'
#' The GUI is available here: \url{https://doi.datacite.org/}.
#'
#' @param pkg character(1) package name
#' @param authors character vector of authors (will be "pasted" together)
#' @param pubyear integer(1) publication year
#' @param testing logical(1) If true, will use the apitest
#'     user with the password apitest. These DOIs will expire.
#'     The same apitest:apitest combination can be used to
#'     login to the EZID website for doing things using the
#'     web interface. If false, the Bioconductor-specific
#'     user credentials should be in the correct environment
#'     variables
#'
#' @return The DOI as a character(1) vector.
#'
#' @importFrom httr POST status_code PUT authenticate timeout content_type accept content
#'
#' @keywords Internal
#'
#' @examples
#' \dontrun{
#'   x = generateBiocPkgDOI('RANDOM_TEST_PACKAGE','Sean Davis',1972)
#' }
#
# draft concept
#
# {
#   "data": {
#     "type": "dois",
#     "attributes": {
#       "doi": "10.18129/B9"
#     }
#   }
# }
# BIOC prefix: 10.18129/B9
# $ curl -X POST -H "Content-Type: application/vnd.api+json" --user YOUR_REPOSITORY_ID:YOUR_PASSWORD -d @my_draft_doi.json https://api.test.datacite.org/dois 
# 
generateBiocPkgDOI_REST = function(pkg, authors, pubyear, testing=TRUE) {
  if(testing) {
    # View results at: https://doi.test.datacite.org
    # The testing piece does not work with new API?
    username='TESTING_USERNAME'
    password='TESTING_PASSWORD'
    bioc_shoulder='doi:10.5072/FK2'
    base_url = 'https://ez.test.datacite.org/id'
  } else {
    username=Sys.getenv('EZID_USERNAME')
    password=Sys.getenv('EZID_PASSWORD')
    bioc_shoulder='doi:10.18129/B9'
    base_url = "https://ez.datacite.org/id"
  }
  bioc_doi_namespace = ".bioc."
  pkg_doi = paste0(bioc_shoulder,bioc_doi_namespace,pkg)
  url0 = file.path(base_url,pkg_doi)
  body = paste(c(sprintf("datacite.title: %s",pkg),
                 sprintf("_target: https://bioconductor.org/packages/%s",pkg),
                 sprintf("datacite.creator: %s",gsub('\n','',paste(authors,collapse=", "))),
                 "datacite.publisher: Bioconductor",
                 sprintf("datacite.publicationyear: %d",pubyear),
                 sprintf("datacite.resourcetype: %s","Software")),collapse="\n")
  res = httr::POST(url0,
                  content_type('text/plain'),
                  accept("text/plain"),
                  httr::authenticate(username,password),
                  body=body,timeout(30))
  if(status_code(res)>=400) {
    res = httr::PUT(url0,
                     content_type('text/plain'),
                     accept("text/plain"),
                     httr::authenticate(username,password),
                     body=body,timeout(30))
  }
  message(res)
  return(res)
  tmp = strsplit(content(res),' ')[[1]][c(2,4)]
  return(tmp[1])
}

#' Bioconductor Maintained Packages
#'
#' List all the packages associated with a maintainer. By default, it will
#' return all packages associated with the `maintainer@bioconductor.org` email.
#' `hasBiocMaint` returns a logical vector corresponding to the input character
#' vector of packages indicating whether any package is maintained by the
#' Bioconductor core team.
#'
#' @param main `character(1)` A regex string to search for in the `Maintainer`
#'   column from the `biocPkgList()` output.
#'
#' @inheritParams biocBuildReport
#' @inheritParams biocDownloadStats
#'
#' @return For `biocMaintained`: a `tibble` of packages associated with the
#'   maintainer.
#'
#' @examples
#'
#' biocMaintained()
#'
#' @export
biocMaintained <-
    function(
        main = "maintainer@bioconductor.org",
        version = BiocManager::version(),
        pkgType = c(
            "software", "data-experiment",
            "workflows", "data-annotation"
        )
    )
{
    pkgType <- match.arg(pkgType, several.ok = TRUE)
    repos <- .matchGetShortName(pkgType, "biocmanager.names")
    pkgs <- biocPkgList(version = version, repo = repos)
    biocmaint <- vapply(
        pkgs$Maintainer,
        function(x)
            any(grepl(main, x, ignore.case = TRUE)),
        logical(1L)
    )
    filtered <- pkgs[biocmaint, ]
    attr(filtered, "version") <- version
    attr(filtered, "maintainer") <- main
    attr(filtered, "pkgType") <- pkgType
    filtered
}

#' @rdname biocMaintained
#'
#' @param pkg `character(1)` A vector of package names (case sensitive).
#'
#' @param repo `character()` A vector of Bioconductor repositories to search
#'   through. By default, it will search through all Bioconductor repositories.
#'
#' @inheritParams biocPkgList
#'
#' @return For `hasBiocMaint`: a logical vector indicating whether the
#'   package is maintained by Bioconductor.
#'
#' @examples
#' ## maintained by Hervé and not maintainer at bioconductor dot org
#' hasBiocMaint("BiocGenerics")
#'
#' @export
hasBiocMaint <- function(
    pkg,
    version = BiocManager::version(),
    main = "maintainer@bioconductor\\.org",
    repo = c("BioCsoft", "BioCexp", "BioCworkflows", "BioCann")
) {
    pkgs <- biocPkgList(
        version = version, repo = repo, addBiocViewParents = FALSE
    )
    match_idx <- match(pkg, pkgs[["Package"]])
    pkgs <- pkgs[match_idx, c("Package", "Maintainer")]
    missing_pkgs <- is.na(pkgs[["Package"]])
    if (any(missing_pkgs))
        pkgs[missing_pkgs, "Package"] <- pkg[missing_pkgs]

    res <- vapply(
        pkgs[["Maintainer"]],
        function(maint) {
            any(grepl(main, maint))
        },
        logical(1L)
    )
    names(res) <- pkgs[["Package"]]
    res
}

#' Generate a DOI for a Bioconductor package
#'
#' This function makes calls out to the DataCite REST API described
#' here: \url{https://support.datacite.org/docs/api-create-dois}. The
#' function creates a new DOI for a Bioconductor package (cannot already
#' exist). The target URL for the DOI is the short Bioconductor
#' package URL.
#'
#' The login information for the "real" Bioconductor account
#' should be stored in the environment variables "DATACITE_USERNAME"
#' and "DATACITE_PASSWORD
#'
#' The GUI is available here: \url{https://doi.datacite.org/}.
#'
#' @param pkg `character(1)` package name
#' @param authors `character()` vector of authors (will be "pasted" together)
#' @param pubyear `integer(1)` publication year
#' @param event Either "hide", "register", or publish". Typically, we use
#'     "publish" to make the DOI findable.
#' @param testing `logical(1)` If true, will use the apitest
#'     user with the password apitest. These DOIs will expire.
#'     The same apitest:apitest combination can be used to
#'     login to the website for doing things using the
#'     web interface. If false, the Bioconductor-specific
#'     user credentials should be in the correct environment
#'     variables
#'
#' @return The DOI as a `character(1)` vector.
#'
#' @importFrom httr VERB content_type content add_headers message_for_status
#' @importFrom jsonlite toJSON fromJSON
#' @importFrom stringr str_to_upper
#'
#' @keywords Internal
#'
#' @examples
#' \dontrun{
#'   x = generateBiocPkgDOI('RANDOM_TEST_PACKAGE','Sean Davis',1972)
#' }
generateBiocPkgDOI <- function(pkg, authors, pubyear, event = "publish", testing = TRUE) {

  username <- Sys.getenv("DATACITE_USERNAME")
  password <- Sys.getenv("DATACITE_PASSWORD")

  if (!is.element(event, c("hide", "register", "publish")))
    stop("event must be 'hide', 'register', or 'publish'.")

  if (testing) {
    # View results at: https://doi.test.datacite.org
    bioc_prefix <- "10.82962"
    base_url <- "https://api.test.datacite.org/dois"
  } else {
    bioc_prefix <- "10.18129"
    base_url <- "https://api.datacite.org/dois"
  }

  bioc_doi_namespace <- "B9.bioc"
  pkg_doi <- paste0(bioc_prefix, "/", bioc_doi_namespace, ".", pkg)
  encode <- "raw"
  payload <- list("data" = list("id" = paste0("https://doi.org/", pkg_doi),
                                "doi" = stringr::str_to_upper(pkg_doi),
                                "attributes" = list("doi" = pkg_doi,
                                                    "event" = event,
                                                    "prefix" = bioc_prefix,
                                                    "suffix" = paste(bioc_doi_namespace, pkg, sep = "."),
                                                    "identifiers" = list("identifier" = pkg_doi,
                                                                         "identifierType" = "DOI"),
                                                    "creators" = list("name" = paste(authors, collapse = ", ")),
                                                    "titles" = list("title" = pkg),
                                                    "url" = paste0("https://bioconductor.org/packages/", pkg),
                                                    "publisher" = "Bioconductor",
                                                    "publicationYear" = pubyear,
                                                    "types" = list("resourceTypeGeneral" = "Software")
                                                   )
                                )
                 )

  authorization <- jsonlite::base64_enc(paste(username, password, sep = ":"))
  response <- httr::VERB("POST",
                         base_url,
                         body = jsonlite::toJSON(payload, auto_unbox = TRUE),
                         httr::add_headers(Authorization = paste("Basic", authorization, sep = " ")),
                         httr::content_type("application/vnd.api+json"),
                         encode = encode)

  if (response$status_code >= 200 && response$status_code < 300) {
    response_text <- httr::content(response, "text")
    response_json <- jsonlite::fromJSON(response_text)
    return(response_json$data$id)
  } else {
    httr::message_for_status(response)
  }
}

## private function from itdepends/R/utils.R
.flat_map_lst <- function(x, f, ...) {
  if (!length(x)) {
    NULL
  } else {
    if (typeof(x) == "closure" && !inherits(x, "function")) {
      class(x) <- c(class(x), "function")
    }
    unlist(lapply(x, f, ...), recursive = FALSE, use.names = FALSE)
  }
}

## private function from itdepends/R/utils.R
.char_or_sym <- function(x) {
  if (is.character(x)) {
    x
  } else if (is.symbol(x)) {
    as.character(x)
  } else {
    character()
  }
}

## Private function from itdepends/R/dep_locate.R
#' @importFrom rlang is_syntactic_literal is_symbol is_pairlist is_call
.dep_usage_lang <- function(x) {
  f <- function(x) {
    ## currently, for constants such as Biostrings::DNA_BASES
    ## is_symbol(x) == TRUE and therefore won't reported as
    ## functionality calls
    if (is_syntactic_literal(x) || is_symbol(x)) {
      return(NULL)
    }

    if (is_pairlist(x) || is.expression(x)) {
      return(.flat_map_lst(x, f))
    }

    if (is_call(x, c("::", ":::"))) {
      return(list(pkg = .char_or_sym(x[[2]]), fun = .char_or_sym(x[[3]])))
    }

    if (is_call(x) && length(x[[1]]) == 1) {
      return(
        c(
          list(pkg = NA, fun = .char_or_sym(x[[1]])),
          .flat_map_lst(x, f)
          )
        )
    }

    .flat_map_lst(x, f)
  }

  res <- f(x)
  if (length(res) > 0) {
    data.frame(
      pkg = as.character(res[seq(1, length(res), 2)]),
      fun = as.character(res[seq(2, length(res), 2)]), stringsAsFactors = FALSE)
  }
}

## Private function to check arguments for pkgDep* functions
.pkgDepCheckArgs <- function(pkg, depdf) {
  if (!is.data.frame(depdf) ||
      any(!c("Package", "dependency", "edgetype") %in%
           colnames(depdf)))
    stop("argument 'depdf' must be a 'data.frame' with columns ",
         "'Package', 'dependency' and 'edgetype'.")

  if (!pkg %in% depdf$Package)
    stop(sprintf("Package %s not in the package dependency 'data.frame'.",
                 pkg))
}

#' Calculate the 'dependency gain' from excluding one or more direct
#' dependencies
#'
#' Calculate the difference between the total number of dependencies of a
#' package and the number of dependencies that would remain if one or more
#' of the direct dependencies were removed.
#'
#' @param g Package dependency graph
#' @param pkg Character string representing the package of interest
#' @param depsToRemove Character vector representing the dependencies
#'   to remove
#'
#' @author Charlotte Soneson
#'
#' @return The 'dependency gain' that would be achieved by excluding the
#'   indicated direct dependencies
#'
#' @keywords internal
#'
#' @importFrom igraph degree delete_vertices V delete_edges
#'
.getDepGain <- function(g, pkg, depsToRemove) {
  ## First make sure that there are no vertices with in-degree 0 in the graph
  ## (these will anyway be removed below, but they are not dependencies
  ## of pkg)
  while (sum(igraph::degree(g, mode = "in") == 0) > 1) {
    g <- igraph::delete_vertices(
      g, setdiff(names(which(igraph::degree(g, mode = "in") == 0)), pkg)
    )
  }

  ## Get number of vertices
  nVertices <- length(igraph::V(g))

  ## For each package in depsToRemove, remove the edge from pkg to
  ## the dependency
  for (dtr in depsToRemove) {
    g <- igraph::delete_edges(g, paste(pkg, dtr, sep = "|"))
  }

  ## Iteratively remove vertices with in-degree 0
  ## (there's nothing left that depends on it)
  ## pkg will always have in-degree 0, but will be retained
  while (sum(igraph::degree(g, mode = "in") == 0) > 1) {
    g <- igraph::delete_vertices(
      g, setdiff(names(which(igraph::degree(g, mode = "in") == 0)), pkg)
    )
  }

  ## Return the dependency gain
  nVertices - length(igraph::V(g))
}

#' Report package imported functionality
#'
#' Function adapted from 'itdepends::dep_usage_pkg' at https://github.com/r-lib/itdepends
#' to obtain the functionality imported and used by a given package.
#'
#' @importFrom tibble tibble
#' @importFrom stats setNames
#'
#' @param pkg `character()` name of the package for which we want
#' to obtain the functionality calls imported from its dependencies
#' and used within the package.
#'
#' @details
#' Certain imported elements, such as built-in constants, will not
#' be identified as imported functionality by this function.
#'
#' @return A tidy data frame with two columns:
#'   * `pkg`: name of the package dependency.
#'   * `fun`: name of the functionality call imported from the
#'            the dependency in the column `pkg` and used within
#'            the analyzed package.
#'
#' @author Robert Castelo
#'
#' @examples
#' pkgDepImports('BiocPkgTools')
#'
#' @export
#' @md
pkgDepImports <- function(pkg) {

  ## fetch all imported functionality
  imp <- getNamespaceImports(pkg)
  mask <- sapply(imp, isTRUE)
  if (any(mask)) {
    imp[mask] <- lapply(names(imp)[mask], function(pkg2) {
      ns <- getNamespace(pkg2)
      exports <- getNamespaceExports(ns)
      nms <- intersect(exports, ls(envir=ns, all.names=TRUE, sorted=FALSE))
      setNames(nms, nms)
    })
  }
  imp <- lapply(imp, grep, pattern="^.__|^-.", invert=TRUE, value=TRUE)

  ## 'getNamespaceImports()' returns a list with imported functionality
  ## by package but the imported functionality of a package may be scattered
  ## throughout more than one list entry with the same name. here we glue
  ## together those entries to have a single one per package
  fun_to_imp <- setNames(rep(names(imp), lengths(imp)),
                         unlist(imp, use.names=FALSE))
  imp <- split(names(fun_to_imp), fun_to_imp)

  ## not all imported functionality may be actually used in a package,
  ## e.g., 'import(IRanges)' and using 'findOverlaps()' only. here we
  ## fetch all calls made from the package to know what is actually
  ## being used.
  pkg_funs <- mget(ls(envir=asNamespace(pkg), all.names=TRUE,
                      sorted=FALSE),
                   envir=asNamespace(pkg), mode="function",
                   inherits=TRUE, ifnotfound=NA)

  ## according to https://cran.r-project.org/doc/manuals/r-devel/R-ints.html#S4-objects
  ## names starting with .__C__classname correspond to classes and .__T__generic:package
  ## correspond to methods. for these two entries, 'pkg_funs' is NA and we should
  ## treat them separately
  mask <- sapply(pkg_funs, function(x) is_syntactic_literal(x) && is.na(x))
  pkg_calls <- do.call(rbind, c(lapply(pkg_funs[!mask], .dep_usage_lang),
                                lapply(strsplit(sub("^\\.__.__", "",
                                                    names(pkg_funs[mask])), ":"),
                                       function(x) c(pkg=x[2], fun=x[1])),
                                make.row.names=FALSE,
                                stringsAsFactors=FALSE))

  ## remove calls to functionality from 'pkg' itself. a call to a functionality is
  ## from the 'pkg' itself ("ours") if either is annotated to 'pkg' OR is not imported
  ## AND either is not annotated either to 'pkg' or has a missing package annotation
  pkg_calls$pkg[pkg_calls$pkg == "NA"] <- NA_character_
  missing_pkg <- is.na(pkg_calls$pkg)
  our_pkg <- !is.na(pkg_calls$pkg) & pkg_calls$pkg == pkg
  ours <- our_pkg | ((!pkg_calls$fun %in% names(fun_to_imp)) & (!our_pkg | missing_pkg))
  pkg_calls <- pkg_calls[!ours, ]
  pkg_calls$pkg[is.na(pkg_calls$pkg)] <- "__otherpkgs__"

  ## group calls to functionality by imported package
  pkg_calls <- split(pkg_calls$fun, pkg_calls$pkg)

  ## now filter out from the imported functionality the part
  ## that is not actually being called from the package
  imp <- lapply(as.list(setNames(names(imp), names(imp))),
                function(nm, implst, pkgcallslst) {
                  res <- intersect(implst[[nm]], pkgcallslst[["__otherpkgs__"]])
                  if (nm %in% names(pkgcallslst))
                    res <- intersect(implst[[nm]], pkgcallslst[[nm]])
                  res
                  }, imp, pkg_calls)
  ## remove functionality from 'base' R
  imp$base <- NULL

  res <- tibble(pkg=rep(names(imp), lengths(imp)),
                fun=unlist(imp, use.names=FALSE))
  res
}

#' Calculate dependency gain achieved by excluding combinations of packages
#'
#' @param pkg character, the name of the package for which we want
#'   to estimate the dependency gain
#' @param depdf a tidy data frame with package dependency information
#'   obtained through the function [buildPkgDependencyDataFrame()]
#' @param maxNbr numeric, the maximal number of direct dependencies to leave
#'   out simultaneously
#'
#' @export
#'
#' @author Charlotte Soneson
#'
#' @return A data frame with three columns: ExclPackages (the excluded direct
#'   dependencies), NbrExcl (the number of excluded direct dependencies),
#'   DepGain (the dependency gain from excluding these direct dependencies)
#'
#' @examples
#' depdf <- buildPkgDependencyDataFrame(
#'   dependencies=c("Depends", "Imports"),
#'   repo=c("BioCsoft", "CRAN")
#' )
#' pcd <- pkgCombDependencyGain('GEOquery', depdf, maxNbr = 3L)
#' head(pcd[order(pcd$DepGain, decreasing = TRUE), ])
#'
#' @importFrom igraph induced_subgraph subcomponent ego
#' @importFrom utils combn
#'
pkgCombDependencyGain <- function(pkg, depdf, maxNbr = 3L) {

  ## check arguments
  .pkgDepCheckArgs(pkg, depdf)

  ## fetch dependency graph
  g <- buildPkgDependencyIgraph(depdf)

  ## exclude 'R', 'base' and 'methods'
  excludedpkgs <- c("R", "base", "methods")
  g <- igraph::induced_subgraph(g, setdiff(names(V(g)), excludedpkgs))

  ## get all reachable dependencies
  deppkgs <- igraph::subcomponent(g, pkg, mode="out")

  ## get the induced subgraph of dependencies for 'pkg'
  g.pkg <- igraph::induced_subgraph(g, deppkgs)

  ## fetch first level dependencies
  dep1pkgs <- names(igraph::ego(g.pkg, nodes=pkg, mode="out", mindist=1)[[1]])

  ## exclude each combination of dependencies and calculate the
  ## dependency gain
  allcombs <- do.call(rbind, lapply(seq_len(min(maxNbr, length(dep1pkgs))),
                                    function(i) {
    combs <- utils::combn(dep1pkgs, i)
    do.call(rbind, apply(combs, 2, function(w) {
      data.frame(Packages = paste(w, collapse = ", "),
                 NbrExcl = length(w),
                 DepGain = .getDepGain(g = g.pkg, pkg = pkg,
                                       depsToRemove = w),
                 stringsAsFactors = FALSE)
    }))
  }))
  allcombs
}

#' Report package dependency burden
#'
#' Elaborate a report on the dependency burden of a given package.
#'
#' @importFrom igraph subcomponent ego
#'
#' @param pkg `character()` name of the package for which we want
#' to obtain metrics on its dependency burden.
#'
#' @param depdf a tidy data frame with package dependency information
#' obtained through the function [buildPkgDependencyDataFrame()].
#'
#' @return A tidy data frame with different metrics on the
#'         package dependency burden. More concretely, the following columns:
#'  * `ImportedAndUsed`: number of functionality calls imported and used in
#'    the package.
#'  * `Exported`: number of functionality calls exported by the dependency.
#'  * `Usage`: (`ImportedAndUsed`x 100) / `Exported`. This value provides an
#'    estimate of what fraction of the functionality of the dependency is
#'    actually used in the given package.
#'  * `DepOverlap`: Similarity between the dependency graph structure of the
#'    given package and the one of the dependency in the corresponding row,
#'    estimated as the [Jaccard index](https://en.wikipedia.org/wiki/Jaccard_index)
#'    between the two sets of vertices of the corresponding graphs. Its values
#'    goes between 0 and 1, where 0 indicates that no dependency is shared, while
#'    1 indicates that the given package and the corresponding dependency depend
#'    on an identical subset of packages.
#'  * `DepGainIfExcluded`: The 'dependency gain' (decrease in the total number
#'    of dependencies) that would be obtained if this package was excluded
#'    from the list of direct dependencies.
#'
#'  The reported information is ordered by the `Usage` column to facilitate the
#'  identification of dependencies for which the analyzed package is using a small
#'  fraction of their functionality and therefore, it could be easier remove them.
#'  To aid in that decision, the column `DepOverlap` reports the overlap of the
#'  dependency graph of each dependency with the one of the analyzed package. Here
#'  a value above, e.g., 0.5, could, albeit not necessarily, imply that removing
#'  that dependency could substantially lighten the dependency burden of the analyzed
#'  package.
#'
#'  An `NA` value in the `ImportedAndUsed` column indicates that the function
#'  `pkgDepMetrics()` could not identify what functionality calls in the analyzed
#'  package are made to the dependency.
#'
#' @author Robert Castelo
#' @author Charlotte Soneson
#'
#' @examples
#' depdf <- buildPkgDependencyDataFrame(
#'   dependencies=c("Depends", "Imports"),
#'   repo=c("BioCsoft", "CRAN")
#' )
#' pkgDepMetrics('BiocPkgTools', depdf)
#'
#' @export
#' @md
pkgDepMetrics <- function(pkg, depdf) {

  ## check arguments
  .pkgDepCheckArgs(pkg, depdf)

  ## fetch dependency graph
  g <- buildPkgDependencyIgraph(depdf)

  ## exclude 'R', 'base' and 'methods'
  excludedpkgs <- c("R", "base", "methods", "rorcid")
  g <- induced_subgraph(g, setdiff(names(V(g)), excludedpkgs))

  ## get all reachable dependencies
  deppkgs <- subcomponent(g, pkg, mode="out")

  ## get the induced subgraph of dependencies for 'pkg'
  g.pkg <- induced_subgraph(g, deppkgs)

  ## fetch first level dependencies
  dep1pkgs <- names(ego(g.pkg, nodes=pkg, mode="out", mindist=1)[[1]])

  ## fetch imported functionality
  du <- pkgDepImports(pkg)
  du <- sapply(split(du$fun, du$pkg), unique)
  du <- lengths(du)

  ifun <- efun <- dgain <- integer(length(dep1pkgs))
  names(ifun) <- names(efun) <- names(dgain) <- dep1pkgs
  ifun[dep1pkgs] <- du[dep1pkgs]

  depov <- numeric(length(dep1pkgs))
  names(depov) <- dep1pkgs

  gvtx <- names(V(g.pkg))
  for (p in dep1pkgs) {
    deppkgs <- subcomponent(g, p, mode="out")
    gdep <- induced_subgraph(g, deppkgs)
    gdepvtx <- names(V(gdep))
    depov[p] <- length(intersect(gvtx, gdepvtx)) / length(union(gvtx, gdepvtx))
    expfun <- getNamespaceExports(p)
    efun[p] <- length(expfun[grep("^\\.", expfun, invert=TRUE)])
    dgain[p] <- .getDepGain(g = g.pkg, pkg = pkg, depsToRemove = p)
  }

  res <- data.frame(ImportedAndUsed=ifun,
                    Exported=efun,
                    Usage=round(100*ifun/efun, digits=2),
                    DepOverlap=round(depov, digits=2),
                    DepGainIfExcluded=dgain,
                    row.names=dep1pkgs,
                    stringsAsFactors=FALSE)
  res[order(res$Usage), ]
}

#' Download and parse the build status information for Bioconductor packages
#'
#' @description This function downloads and parses the build status information
#'  for Bioconductor packages. The build status information is available for
#'  the current release and the previous release. Other versions may be
#'  available.
#'
#' @inheritParams biocBuildReport
#'
#' @return A `data.frame` with the following columns:
#'  * pkg: The name of the package
#'  * node: The builder on which the package was built
#'  * stage: The stage of the build, e.g., 'install', 'buildsrc', 'checksrc',
#'    etc.
#'  * result: The status of the build, e.g., 'OK', 'ERROR', 'WARNINGS', etc.
#'
#' @export
biocBuildStatusDB <- function(
    version = BiocManager::version(),
    pkgType = c(
        "software", "data-experiment", "data-annotation", "workflows"
    )
) {
    if (version %in% c("release", "devel"))
        version <- BiocManager:::.version_bioc(version)

    version <- try(as.package_version(version), silent = TRUE)

    if (is(version, "try-error"))
        stop(
            "'version' is not 'release', 'devel', or a valid 'package_version'"
        )

    pkgType <- match.arg(pkgType, several.ok = TRUE)

    if (version < package_version("3.14"))
        pkgType <- "software"

    pkgType <- .matchGetShortName(pkgType, "stat.url")

    urls <- get_build_status_db_url(version, pkgType)
    names(urls) <- pkgType
    urls <- Filter(.url_exists, urls)

    url_list <- lapply(
        urls,
        function(url) {
            file <- .cache_url_file(url)
            dat <- readLines(file)
            sdat <- strsplit(dat, "#|:\\s")
        }
    )
    sdat <- do.call(
        rbind, unlist(url_list, recursive = FALSE)
    )
    sdat <- as.data.frame(sdat)
    names(sdat) <- c("pkg", "node", "stage", "result")
    attr(sdat, "BioCversion") <- version
    attr(sdat, "retrieved") <- Sys.time()
    sdat
}

#' Extract funder names from a formatted \code{Author} field string
#'
#' @param author_string character(1) a formatted \code{Author} field value as
#'   found in a Bioconductor \code{VIEWS} file, e.g.
#'   \code{"Jane Doe [aut, cre], Big Org [fnd]"}.  May contain embedded
#'   newlines.  Note: funder names that themselves contain commas will not be
#'   extracted correctly, as commas are used as entry separators in the
#'   formatted \code{Author} field.
#'
#' @return character() of funder names, or \code{NA_character_} when no funder
#'   is found or the input is missing/empty
#'
#' @keywords internal
.extract_fnd <- function(author_string) {
    if (is.na(author_string) || !nzchar(trimws(author_string)))
        return(NA_character_)
    ## Normalise line-continuation whitespace
    s <- gsub("\\s*[\n\r]+\\s*", " ", author_string)
    ## Match author entries whose role bracket contains the "fnd" role
    ## Format: "Some Name [role1, fnd, role2]"
    m <- gregexpr(
        "([^,\\[]+?)\\s*\\[([^\\]]*\\bfnd\\b[^\\]]*)\\]",
        s, perl = TRUE
    )
    matches <- regmatches(s, m)[[1L]]
    if (!length(matches))
        return(NA_character_)
    ## Strip the trailing "[roles]" to recover just the name
    names_only <- trimws(sub("\\s*\\[[^\\]]*\\]$", "", matches))
    ## Remove any stray leading comma (can occur when the entry is not first)
    names_only <- trimws(sub("^,\\s*", "", names_only))
    names_only <- names_only[nzchar(names_only)]
    if (!length(names_only)) NA_character_ else names_only
}

#' @import biocViews
#' @importFrom RBGL transitive.closure
.computeBiocViewsTransitiveClosure <- function() {
  data("biocViewsVocab", package = "biocViews", envir = environment())
  return(RBGL::transitive.closure(biocViewsVocab))
}

#' @importFrom graph edges nodes
.biocChildToParent <- function() {
  biocViewsTC = .computeBiocViewsTransitiveClosure()
  biocViewsTCEdges = edges(biocViewsTC)
  df = data.frame(child = unlist(biocViewsTCEdges, use.names = FALSE),
                  parent = rep(names(biocViewsTCEdges), sapply(biocViewsTCEdges, length)),
                  stringsAsFactors = FALSE)
  # Remove BiocViews, as this is simply the root
  df = df[df$parent!='BiocViews',]
  return(split(df$parent, df$child))
}

#' Get full Bioconductor software package listing, with details
#'
#' The BiocViews-generated `VIEWS` file is available
#' for Bioconductor release and devel repositories. It
#' contains quite a bit more information from the
#' package `DESCRIPTION` files than the `PACKAGES`
#' file. In particular, it contains `biocViews` annotations
#' and URLs for vignettes and developer URLs.
#'
#' Since packages are annotated with the most specific
#' views, the default functionality here is to add parent terms
#' for all views for each package. For example, in the bioCsoft
#' repository, all packages will have at least "Software" added
#' to their biocViews. If one wants to stick to only the most
#' specific terms, set `addBiocViewParents` to `FALSE`.
#'
#' @inheritParams biocBuildReport
#'
#' @param repo `character(1)` The requested Bioconductor repository. The default
#'   is to pull from the "BioCsoft" repository. Possible repositories include
#'   "BioCsoft", "BioCexp", "BioCworkflows", "BioCann", and "CRAN". Note that
#'   not all repos are available for all versions, particularly older versions.
#'
#' @param addBiocViewParents `logical(1)` whether to add all biocViews
#'    parents to biocViews annotations.
#'
#' @return An object of class `tbl_df` with one row per package. The result
#'   always includes a `fnd` list-column whose elements are character vectors
#'   of funder names extracted from persons with role `"fnd"` in the
#'   formatted `Author` field. Elements are `NA_character_` for packages
#'   that declare no funder (or whose `Author` field contains no `[fnd]`
#'   role tag).
#'
#' @importFrom BiocManager repositories version
#' @importFrom stringr str_split str_replace_all str_remove_all str_squish
#' @importFrom tibble as_tibble
#'
#' @examples
#' bpkgl <- biocPkgList(repo = "BioCsoft")
#' bpkgl
#' unlist(bpkgl[1,'Depends'], use.names = FALSE)
#'
#' # Get a list of all packages that
#' # import "GEOquery"
#' library(dplyr)
#' bpkgl |>
#'   filter(Package == 'GEOquery') |>
#'   pull('importsMe') |>
#'   unlist()
#'
#' @export
biocPkgList <- function(
    version = BiocManager::version(),
    repo = c("BioCsoft", "BioCexp", "BioCworkflows", "BioCann", "CRAN"),
    addBiocViewParents = TRUE
) {
    if(!is.logical(addBiocViewParents) ||
       !(addBiocViewParents %in% c(TRUE,FALSE)))
      stop('addBiocViewParents must be a single logical.')
    if (missing(repo))
        repo <- "BioCsoft"
    else
        repo <- match.arg(repo, several.ok = TRUE)
    if (!is.character(repo))
      stop('repo must be a character vector.')

    if (version %in% c("release", "devel"))
        version <- BiocManager:::.version_bioc(version)

    version <- try(as.package_version(version), silent = TRUE)

    if (is(version, "try-error"))
        stop(
            "'version' is not 'release', 'devel', or a valid 'package_version'"
        )
    repos <- BiocManager:::.repositories(
        site_repository = character(), version = version
    )
    mask <- !repo %in% names(repos)
    if (any(mask))
      warning(sprintf("repo %s not found in known repositories. Valid values for repo can be shown by running names(BiocManager::repositories()).",
                      paste(repo[mask], collapse=", ")))
    if (all(mask))
      stop("No repo was found in known repositories.")
    repo <- repo[!mask]

    ret <- lapply(as.list(repo),
                  function(r) {
                    if (r != "CRAN") {
                      ret <- .get_VIEWS(version = version, type = r)
                    } else {
                      ret <- get_CRAN_pkg_rds()
                      ## to increase the overlap of the information from
                      ## CRAN with Bioconductor, replace CRAN column names
                      ## to Bioconductor column names for 'reverse imports',
                      ## 'depends' and 'suggests'
                      biocrevnames <- c(importsMe="Reverse imports",
                                        dependsOnMe="Reverse depends",
                                        suggestsMe="Reverse suggests")
                      mt <- match(biocrevnames, colnames(ret))
                      mask <- !is.na(mt)
                      colnames(ret)[mt[mask]] <- names(biocrevnames)[mask]
                      ret$biocViews<- NA_character_
                    }

                    # convert comma-delimited text columns into
                    # list columns
                    commaCols = c('Depends', 'Suggests', 'dependsOnMe', 'Imports', 'importsMe',
                                'Enhances', 'vignettes', 'vignetteTitles', 'suggestsMe',
                                'Maintainer', 'biocViews', 'Archs', 'linksToMe', 'LinkingTo',
                                'Rfiles')
                    commaCols = intersect(commaCols, colnames(ret))
                    for(commaCol in commaCols) {
                        ret[[commaCol]] = str_split(ret[[commaCol]],'\\s?,\\s?')
                    }
                    if(r != "CRAN" && addBiocViewParents) {
                      child2parentMap = .biocChildToParent()
                      tmp = lapply(ret$biocViews, function(views) {
                        return(unique(c(unlist(child2parentMap[views]),views)))
                      })
                      ret$biocViews = tmp
                    }

                    ## Extract funder names from the raw (unprocessed) Author
                    ## field while role brackets are still present
                    ret[["fnd"]] <- lapply(ret[["Author"]], .extract_fnd)

                    ret[["Author"]] = ret[["Author"]] |>
                        str_replace_all("\n", " ") |>
                        str_remove_all("\\[.*?\\]") |>
                        str_remove_all("<.*?>") |>
                        str_remove_all("\\(.*?\\)") |>
                        str_squish() |>
                        str_replace_all("\\w* contributions ?\\w*", ", ") |>
                        str_replace_all("\\sand\\s", ", ") |>
                        str_replace_all(",\\s+,", ",") |>
                        str_replace_all(",+", ",")

                    ret[["Author"]] = lapply(
                        str_split(ret[["Author"]], ","),
                        str_squish
                    )
                    ret
                  })

    ## build a table with the common columns provided by the different
    ## repositories. if no common columns exist, then prompt an error
    commonCols <- Reduce(intersect, lapply(ret, colnames))
    if (length(commonCols) == 0)
      stop(sprintf("No common columns between repos in %s",
                   paste(repo, collapse=", ")))

    ret <- lapply(ret, function(x, cnames) x[, cnames], commonCols)
    ret <- do.call("rbind", ret)

    ret = as_tibble(ret)
    class(ret) = c("biocPkgList", class(ret))
    ret
}

stripVersionString <- function(s) {
    sub('\\s?\\(.*\\)\\s?','',s)
}

CRAN_pkg_rds_url <- function() {
    paste0(.CRAN_WEB_URL, "/web/packages/packages.rds")
}

read_CRAN_url <- function(url) {
    con <- url(url)
    on.exit(close(con))
    res <- try(readRDS(gzcon(con)), silent = TRUE)
    if (inherits(res, "try-error"))
        stop("'packages.rds' not found in CRAN 'repo'", call. = FALSE)
    else
        as.data.frame(res, stringsAsFactors = FALSE)
}

get_CRAN_pkg_rds <- function() {
    cran_url <- CRAN_pkg_rds_url()
    read_CRAN_url(cran_url)
}

context("biocPkgList")
library(BiocPkgTools)

bpkgl = biocPkgList()


test_that("pkg list has correct classes", {
    expect_is(bpkgl, 'tbl_df')
    expect_is(bpkgl, 'tbl')
    expect_is(bpkgl, 'data.frame')
})

test_that("ncol is approximately correct", {
  expect_gt(ncol(bpkgl), 35)
})

test_that("nrow is approximately correct", {
  expect_gt(nrow(bpkgl), 1000)
})

test_that("fnd column is present and is a list", {
  expect_true("fnd" %in% colnames(bpkgl))
  expect_type(bpkgl[["fnd"]], "list")
})

.get_VIEWS_url <- function(version, type) {
    bioc_repos <- BiocManager:::.repositories_bioc(version = version)
    paste0(bioc_repos[type], "/VIEWS")
}

.read_VIEWS_url <- function(url) {
    file <- .cache_url_file(url)
    tryCatch({
        read.dcf(file) |>
            as.data.frame()
    }, error = function(e) {
        stop(
            "Error reading VIEWS file from URL: ",
            url,
            "\n",
            conditionMessage(e)
        )
    })
}

.get_VIEWS <- function(version, type) {
    views_url <- .get_VIEWS_url(version = version, type = type)
    .read_VIEWS_url(views_url)
}

#' Parse and return the VIEWS file for a Bioconductor Release
#'
#' The function parses and returns the `VIEWS` file for a specified Bioconductor
#' version, either "release" or "devel". The `VIEWS` file contains metadata
#' about Bioconductor packages, including information about their categories,
#' topics, and other details.
#'
#' @inheritParams biocBuildReport
#'
#' @examplesIf interactive()
#' biocVIEWSdb(pkgType = "software")
#' @export
biocVIEWSdb <- function(
    version = BiocManager::version(),
    pkgType = c("software", "data-experiment", "data-annotation", "workflows")
) {
    if (version %in% c("release", "devel"))
        version <- BiocManager:::.version_bioc(version)

    version <- try(as.package_version(version), silent = TRUE)

    if (is(version, "try-error"))
        stop(
            "'version' is not 'release', 'devel', or a valid 'package_version'"
        )

    pkgType <- match.arg(pkgType, several.ok = TRUE)
    pkgType <- .matchGetShortName(pkgType, "biocmanager.names")

    views_list <- structure(
        vector("list", length(pkgType)),
        .Names = pkgType
    )
    for (type in pkgType) {
        views_df <- .get_VIEWS(version = version, type = type)
        if (nrow(views_df))
            views_list[[type]] <- views_df
    }
    dplyr::bind_rows(views_list)
}

test_that("biocVIEWSdb returns a data.frame with rows", {
    pkgTypes <- c("software", "data-experiment", "data-annotation", "workflows")
    for (type in pkgTypes) {
        df <- biocVIEWSdb(pkgType = type)
        expect_s3_class(df, "data.frame")
        expect_gt(nrow(df), 0L)
    }
})

test_that("biocVIEWSdb handles invalid pkgType", {
    expect_error(
        biocVIEWSdb(pkgType = "invalid-type"),
        "'arg' should be one of"
    )
})

test_that("biocVIEWSdb handles invalid version", {
    expect_error(
        biocVIEWSdb(version = "invalid-version"),
        "'version' is not 'release', 'devel', or a valid 'package_version'"
    )
})

.COLS_OF_INTEREST <- c(
    "employment-summary.source.source-name.value",
    "employment-summary.organization.name",
    "employment-summary.source.source-orcid.path",
    "employment-summary.source.source-orcid.uri",
    "employment-summary.put-code",
    "employment-summary.path",
    "employment-summary.organization.address.city",
    "employment-summary.organization.address.country"
)

.get_orcid_token <- function(req) {
    orcid_app <-  oauth_client(
        id = "APP-OSPJCJPZ6P4I3DF0",
        secret = obfuscated(
            paste0(
                "zx3gSdnxRyInqsP0f_72jT0JO54IMQA-",
                "Vgx2hblcIEGEcDDt2VpWaLgeHO8LHRVffOUY4w"
            )
        ),
        token_url = "https://orcid.org/oauth/token",
        name = "biocreporting_orcid"
    )
    oauth_flow_client_credentials(
        client = orcid_app,
        scope = "/read-public"
    )
}

.get_orcid_endpoint <- function(orcid_id, endpoint, token) {
    request("https://pub.orcid.org/") |>
        req_template(
            "v3.0/{orcid}/{endpoint}",
            orcid = orcid_id,
            endpoint = endpoint
        ) |>
        req_headers(
            Accept = "application/vnd.orcid+json"
        ) |>
        req_auth_bearer_token(token$access_token) |>
        req_perform() |>
        resp_body_json()
}

.bind_employments <- function(orcid_id, token) {
    result <- .get_orcid_endpoint(orcid_id, "employments", token)
    affiliation_groups <- lapply(
        result[["affiliation-group"]],
        function(affiliation_group) {
            resframe <- unlist(affiliation_group[["summaries"]][[1L]]) |>
                t() |>
                as.data.frame()
            resframe[, .COLS_OF_INTEREST]
        }
    )
    do.call(rbind.data.frame, affiliation_groups)
}

#' @title Obtain employment data from ORCID
#'
#' @description Get a `data.frame` of employment info from ORCID
#'
#' @importFrom httr2 request req_headers req_body_form req_method req_perform
#'   resp_body_json req_template req_auth_bearer_token
#'   oauth_flow_client_credentials oauth_client obfuscated
#'
#' @param orcids `character()` A vector of ORCID identifiers
#'
#' @returns a `data.frame` of employment info using the ORCID API
#'
#' @examples
#' if (interactive()) {
#'     orcid_table(
#'         orcids = c(
#'             "0000-0002-3242-0582",
#'             "0000-0003-4046-0063",
#'             "0000-0003-2725-0694"
#'         )
#'     )
#' }
#' @export
orcid_table <- function(orcids) {
    token <- .get_orcid_token()
    results <- lapply(
        orcids,
        function(orcid, token) {
            .bind_employments(orcid, token)
        }, token = token
    )
    do.call(rbind.data.frame, results)
}

#' get the ORCID id from cre field of Authors@R in packageDescription result
#' @param pkgname character(1)
.get_cre_orcid <- function(pkgname) {
    stopifnot(length(pkgname) == 1)
    de <- packageDescription(pkgname)
    nde <- names(de)
    if (!("Authors@R" %in% nde)) {
        return(NA_character_)
    }
    au <- de[["Authors@R"]]
    vec <- eval(parse(text = au))
    cre <- grep("\\[.*cre.*\\]", vec, value = TRUE)
    hasorc <- grep("orcid.org", cre)
    if (length(hasorc) == 0) {
        return(NA_character_)
    }
    gsub(".*orcid.org/(.*)>.*", "\\1", cre)
}

#' get ORCID ids from cre fields of Authors@R in packageDescription results
#' @param pkgnames `character()` must be installed
#' @note returns NA if no ORCID provided in Authors@R for package description
#' @examples
#' get_cre_orcids(c("BiocPkgTools", "utils"))
#' @export
get_cre_orcids <- function(pkgnames) {
    vapply(pkgnames, .get_cre_orcid, character(1))
}

#' The Bioconductor datacite.org XML generator
#'
#' This function is used internally to generate XML elements from the
#' datacite.org website for incoming Bioconductor packages.
#'
#' @param pkg The name of a Bioconductor package
#'
#' @import xml2
#'
#' @return An xml_document object from the `xml2` package.
#'
#' @seealso \preformatted{?xml2::`xml_document-class`}
#'
#' @keywords internal
dataciteXMLGenerate <- function(pkg) {
    bioc_shoulder <- "doi:10.5072/FK2"
    bioc_doi_namespace <- ".bioc."
    pkg_doi <- paste0(bioc_shoulder,bioc_doi_namespace,pkg)
    x <- xml_new_root(
        'resource',
        xmlns="http://datacite.org/schema/kernel-3",
        `xmlns:xsi`="http://www.w3.org/2001/XMLSchema-instance",
        `xsi:schemaLocation`="http://datacite.org/schema/kernel-3 http://schema.datacite.org/meta/kernel-3/metadata.xsd"
    )
    xml_add_child(x,'identifier',identifierType='DOI',pkg_doi)
    descriptions <- xml_add_child(x,'descriptions')
    xml_add_child(descriptions,'description','this is the description')
    return(x)
}

#' Get data from Bioconductor
#'
#' @return A JSON string containing Bioconductor package details
#'
#' @importFrom jsonlite toJSON
#'
#' @export
#'
#' @examples
#' bioc_data <- get_bioc_data()
get_bioc_data <- function() {
    # if (has_cached_data() && cache_last_update_days() < 7) {
    #     return(get_cached_data())
    # }

    message("Downloading package data...")
    full_data <- process_data(
        pkg_list = BiocPkgTools::biocPkgList(),
        raw_dl_stats = BiocPkgTools::biocDownloadStats()
    )

    full_data <- full_data |>
        dplyr::filter(!is.na(.data$tags))

    message("Package data download complete")

    json_data <- jsonlite::toJSON(full_data)
    # write_to_cache(json_data)

    return(json_data)
}

# process retrieved data into required data.frame columns
process_data <- function(pkg_list, raw_dl_stats) {
    dl_stats <- summarise_dl_stats(raw_dl_stats)

    pkg_link <- function(pkg) {
        vapply(
            pkg,
            function(x) {
                stringr::str_interp("https://bioconductor.org/packages/release/bioc/html/${x}.html")
            },
            character(1)
        )
    }

    # convert from factor to character to avoid inner_join warning
    pkg_list$Package <- as.character(pkg_list$Package)
    dl_stats$Package <- as.character(dl_stats$Package)

    full_data <- dplyr::inner_join(pkg_list, dl_stats, by = "Package") |>
        dplyr::select(
            .data$Author,
            .data$Package,
            .data$License,
            .data$biocViews,
            .data$Description,
            .data$downloads_month,
            .data$downloads_total
        ) |>
        dplyr::mutate(
            page = pkg_link(.data$Package)
        ) |>
        dplyr::rename(
            authors = .data$Author,
            name = .data$Package,
            license = .data$License,
            tags = .data$biocViews,
            description = .data$Description
        )

    full_data$authors <- author_list_to_string(full_data$authors)

    full_data
}

# summarise download stats into monthly and total lifetime downloads
summarise_dl_stats <- function(dl_stats) {
    dl_stats |>
        dplyr::group_by(.data$Package) |>
        dplyr::summarise(
            downloads_month = dplyr::first(.data$Nb_of_downloads),
            downloads_total = sum(.data$Nb_of_downloads)
        )
}

# collapse list of names into a comma separated string
# final two authors separated by 'and'
author_list_to_string <- function(authors) {
    collapse_list <- function(x) unlist(x) |> paste(collapse = ", ")
    sapply(authors, collapse_list) |>
        stringr::str_replace(" and ", ", ") |>
        stringr::str_replace(",([^,]*)$", " and\\1")
}

# get days since cached data was last updated
cache_last_update_days <- function() {
    cached_data_path <- system.file(
        "extdata",
        "data.Rds",
        package = "BiocExplorer"
    )
    cache_last_mod <- file.info(cached_data_path)$mtime

    difftime(Sys.time(), cache_last_mod, units = "days")
}

# get cached json data
get_cached_data <- function() {
    message("Getting data from cache")
    cached_data_path <- system.file(
        "extdata",
        "data.Rds",
        package = "BiocExplorer"
    )

    readRDS(cached_data_path)
}

# write json to cache location
write_to_cache <- function(json_data) {
    message("Writing data to cache")
    cached_data_path <- system.file(
        "extdata",
        package = "BiocExplorer"
    ) |>
        file.path("data.Rds")

    saveRDS(json_data, cached_data_path)
}

# check if cached data is available
has_cached_data <- function() {
    cached_data_path <- system.file(
        "extdata",
        "data.Rds",
        package = "BiocExplorer"
    )

    return(cached_data_path != "")
}

#' Download a Bioconductor vignette
#'
#' The actual vignette path is available using [biocPkgList()].
#'
#' @param vignettePath `character(1)` the additional path information to get to
#'   the vignette
#' @param destfile `character(1)` the file location to store the vignette
#' @param version `character(1)` such as "3.7", defaults to user version
#'
#' @importFrom utils download.file
#'
#' @return `character(1)` The filename of the downloaded vignette
#'
#' @examples
#' x = biocPkgList()
#' tmp = getBiocVignette(x$vignettes[[1]][1])
#' tmp
#'
#' \dontrun{
#' library(pdftools)
#' y = pdf_text(tmp)
#' y = paste(y,collapse=" ")
#' library(tm)
#' v = VCorpus(VectorSource(y))
#'
#' v <- v |>
#'     tm_map(stripWhitespace) |>
#'     tm_map(content_transformer(tolower)) |>
#'     tm_map(removeWords, stopwords("english")) |>
#'     tm_map(stemDocument)
#' dtm = DocumentTermMatrix(v)
#' inspect(DocumentTermMatrix(v,
#'     list(dictionary = as.character(x$Package))))
#'}
#' @export
getBiocVignette <- function(
    vignettePath,
    destfile = tempfile(),
    version = BiocManager::version()
) {
    stopifnot(is.character(vignettePath) & length(vignettePath) == 1)
    p = sprintf(
        'https://bioconductor.org/packages/%s/bioc/%s',
        version,
        vignettePath
    )
    download.file(p, destfile, mode = "wb")
    destfile
}

---
title: "%s Bioconductor Package Deprecation"
author:
- name: "%s"
  affiliation:
  - Roswell Park Comprehensive Cancer Center, Buffalo, NY, USA
date: "%s"
output:
  BiocStyle::html_document:
    toc: false
    number_sections: false
---

## Package: %s

---


Hi %s,

We are contacting you regarding the `%s` package in Bioconductor
%s, for which you are listed as the maintainer.

We have attempted to reach you multiple times with deprecation warnings.
Unfortunately, we have decided to proceed with the deprecation of your package.

The package will be marked deprecated until the package is fixed so that
install, build, and check all pass on the Bioconductor Build System **AND**
that you formally request undeprecation by sending an email to
bioc-devel@bioconductor.org.

If the package remains marked for deprecation by the
[next release of Bioconductor](https://www.bioconductor.org/developers/release-schedule/),
the package will be removed from Bioconductor and it will have to be
resubmitted and undergo review to be reinstated.

%s

If you have any questions, feel free to contact contact us on the bioc-devel
mailing list.

Best regards,

%s


%s

---

<footer>
<p>Bioconductor Core Team
<br>Roswell Park Comprehensive Cancer Center
<br>Dept. of Biostatistics & Bioinformatics
<br>Elm & Carlton Streets
<br>Buffalo, New York 14263
</p>
</footer>

---
title: '`r sprintf("Package Report: %s", params$package)`'
date: "`r Sys.Date()`"
output: 
  BiocStyle::html_document:
    toc: false
params:
  # change this to YOUR package name
  package: "GEOquery"
---

```{r include=FALSE}
library(knitr)
library(BiocStyle)
opts_chunk$set(message = FALSE, warning = FALSE, cache = TRUE)
```

```{r include=FALSE}
library(BiocPkgTools)
dls <- biocDownloadStats()
pkl <- biocPkgList()
pkl_me <- pkl[pkl$Package==params$package,]
```

# Downloads and Usage

```{r echo=FALSE}
#| fig.cap = paste0(
#|  "Package download counts. Bioconductor tracks the number of unique IP",
#|  " addresses that download individual packages. These numbers can be used",
#|  " as a proxy for the number of users of a software package.",
#| )
library(ggplot2)
library(dplyr)
downloads <- dls |>
    filter(Package == params$package) |>
    filter(!is.na(Date))
ggplot(downloads, aes(y=Nb_of_distinct_IPs, x=Date)) +
    geom_bar(stat='identity') +
    ylab("Distinct IP Downloads") +
    ggtitle("Number of Downloads")
```

# Dependencies

The `r params$package` package depends on `r length(pkl_me$Depends[1][[1]])`
packages and suggests `r length(pkl_me$Suggests[1][[1]])` packages.

```{r, echo=FALSE}
library(pkgndep)
pkg_report <-  pkgndep(params$package)
pkg_report
```

```{r,echo=FALSE}
rmd_dir <- dirname(knitr::current_input(dir = TRUE))
## NOTE. this command saves a file in the directory of the Rmd file
heatmap_file <- file.path(rmd_dir, "heatmap.png")
dependency_heatmap(pkg_report, file = heatmap_file)
```

```{r, echo=FALSE, results="asis"}
cat(paste0(
'<div style="overflow-x: auto; width: 100%; border: 1px solid #ddd;">\n',
'<img src="', heatmap_file,
'" alt="Dependency Heatmap" style="display: block; max-width: none;">\n',
'</div>'
))
```

---
title: "%s CRAN Package Report"
date: "%s"
output:
  BiocStyle::html_document:
    toc: false
    number_sections: false
---

## Package: %s

---

Please be advised that the `%s` CRAN package has at least one
error in the build results. Please see the status table below:

```{r echo = FALSE, results='asis', out.width="\\maxwidth"}
kableExtra::kable_styling(
    knitr::kable(%s, caption = "CRAN Build Report",
        align = "l", escape = FALSE
    ),
    bootstrap_options = c("hover", "striped", "responsive"),
    full_width = FALSE
)
```

%s

---

---
title: "Overview of BiocPkgTools"
author:
- name: Shian Su
  affiliation: Walter and Eliza Hall Institute, Melbourne, Australia
- name: Vince Carey
  affiliation: Channing Lab, Brigham and Womens Hospital, Harvard University, Boston, MA USA
- name: Marcel Ramos
  affiliation: CUNY School of Public Health, New York, NY USA
- name: Lori Shepherd
  affiliation: Roswell Park Comprehensive Cancer Center, Buffalo, NY USA
- name: Martin Morgan
  affiliation: Roswell Park Comprehensive Cancer Center, Buffalo, NY USA
- name: Sean Davis
  affiliation: National Cancer Institute, National Institutes of Health, Bethesda, MD USA
  email: seandavi@gmail.com
package: BiocPkgTools
output:
  BiocStyle::html_document:
    toc: false
abstract: |
  Bioconductor has a rich ecosystem of metadata around packages, usage, and build status. This package is a simple collection of functions to access that metadata from R in a tidy data format. The goal is to expose metadata for data mining and value-added functionality such as package searching, text mining, and analytics on packages. 
vignette: |
  %\VignetteIndexEntry{Overview of BiocPkgTools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

Bioconductor has a rich ecosystem of metadata around packages, usage, and build
status. This package is a simple collection of functions to access that metadata
from R in a tidy data format. The goal is to expose metadata for data mining and
value-added functionality such as package searching, text mining, and analytics
on packages. 

Functionality includes access to :

- Download statistics
- General package listing
- Build reports
- Package dependency graphs
- Vignettes

```{r init, include=FALSE}
library(knitr)
opts_chunk$set(warning = FALSE, message = FALSE, cache=FALSE)
```

```{r style, echo = FALSE, results = 'asis'}
BiocStyle::markdown()
```

# Build reports

The Bioconductor build reports are available online as HTML pages. 
However, they are not very computable.
The `biocBuildReport` function does some heroic parsing of the HTML
to produce a *tidy* data.frame for further processing in R. 

```{r}
library(BiocPkgTools)
head(biocBuildReport())
```

## Personal build report

Because developers may be interested in a quick view of their own packages,
there is a simple function, `problemPage`, to produce an HTML report of the
build status of packages matching a given author *regex* supplied to the
`authorPattern` argument. The default is to report only "problem" build statuses
(ERROR, WARNING).

```{r eval=FALSE}
problemPage(authorPattern = "V.*Carey")
```

In similar fashion, maintainers of packages that have many downstream packages
that depend on them may wish to check that a change they introduced hasn't
suddenly broken a large number of these.  You can use the `dependsOn` argument
to produce the summary report of those packages that "depend on" the given
package.

```{r eval=FALSE}
problemPage(dependsOn = "limma")
```

When run in an interactive environment, the `problemPage` function 
will open a browser window for user interaction. Note that if you want
to include all your package results, not just the broken ones, simply 
specify `includeOK = TRUE`.

# Download statistics

Bioconductor supplies download stats for all packages. The `biocDownloadStats`
function grabs all available download stats for all packages in all
Experiment Data, Annotation Data, and Software packages. The results
are returned as a tidy data.frame for further analysis.

```{r}
head(biocDownloadStats())
```

The download statistics reported are for ***all available versions*** of a
package. There are no separate, publicly available statistics broken down by
version.
The majority of Bioconductor Software packages are also available through other
channels such as Anaconda, who also provided download statistics for packages
installed from their repositories.  Access to these counts is provided by the
`anacondaDownloadStats` function:

```{r}
head(anacondaDownloadStats())
```

Note that Anaconda do not provide counts for distinct IP addresses, but this
column is included for compatibility with the Bioconductor count tables.

# Package details

The R `DESCRIPTION` file contains a plethora of information regarding package
authors, dependencies, versions, etc. In a repository such as Bioconductor,
these details are available in bulk for all included packages. The `biocPkgList`
returns a data.frame with a row for each package. Tons of information are
available, as evidenced by the column names of the results.

```{r}
bpi = biocPkgList()
colnames(bpi)
```

Some of the variables are parsed to produce `list` columns. 

```{r}
head(bpi)
```

As a simple example of how these columns can be used, extracting
the `importsMe` column to find the packages that import the 
`r Biocpkg("GEOquery")` package.


```{r}
require(dplyr)
bpi = biocPkgList()
bpi |> 
    filter(Package=="GEOquery") |>
    pull(importsMe) |>
    unlist()
```

# Package Explorer

For the end user of Bioconductor, an analysis often starts with finding a
package or set of packages that perform required tasks or are tailored 
to a specific operation or data type. The `biocExplore()` function
implements an interactive bubble visualization with filtering based on 
biocViews terms. Bubbles are sized based on download statistics. Tooltip
and detail-on-click capabilities are included. To start a local session:

```{r biocExplore}
biocExplore()
```

# Dependency graphs

The Bioconductor ecosystem is built around the concept of interoperability
and dependencies. These interdependencies are available as part of the
`biocPkgList()` output. The `BiocPkgTools` provides some convenience
functions to convert package dependencies to R graphs. A modular approach leads
to the following workflow.

1. Create a `data.frame` of dependencies using `buildPkgDependencyDataFrame`.
2. Create an `igraph` object from the dependency data frame using
  `buildPkgDependencyIgraph`
3. Use native `igraph` functionality to perform arbitrary network operations. 
  Convenience functions, `inducedSubgraphByPkgs` and `subgraphByDegree` are
  available.
4. Visualize with packages such as `r CRANpkg("visNetwork")`.


## Working with dependency graphs

A dependency graph for all of Bioconductor is a starting place.

```{r}
dep_df = buildPkgDependencyDataFrame()
g = buildPkgDependencyIgraph(dep_df)
g
library(igraph)
head(V(g))
head(E(g))
```

See `inducedSubgraphByPkgs` and `subgraphByDegree` to produce
subgraphs based on a subset of packages.

See the igraph documentation for more detail on graph analytics, setting
vertex and edge attributes, and advanced subsetting.

## Graph visualization

The visNetwork package is a nice interactive visualization tool
that implements graph plotting in a browser. It can be integrated
into shiny applications. Interactive graphs can also be included in
Rmarkdown documents (see vignette)

```{r}
igraph_network = buildPkgDependencyIgraph(buildPkgDependencyDataFrame())
``` 

The full dependency graph is really not that informative to look at, though
doing so is possible. A common use case is to visualize the graph of
dependencies "centered" on a package of interest. In this case, I will 
focus on the `r Biocpkg("GEOquery")` package. 

```{r}
igraph_geoquery_network = subgraphByDegree(igraph_network, "GEOquery")
```

The `subgraphByDegree()` function returns all nodes and connections within
`degree` of the named package; the default `degree` is `1`.

The visNework package can plot `igraph` objects directly, but more flexibility
is offered by first converting the graph to visNetwork form.

```{r}
library(visNetwork)
data <- toVisNetworkData(igraph_geoquery_network)
```

The next few code chunks highlight just a few examples of the visNetwork
capabilities, starting with a basic plot.

```{r }
visNetwork(nodes = data$nodes, edges = data$edges, height = "500px")
```

For fun, we can watch the graph stabilize during drawing, best viewed
interactively.

```{r}
visNetwork(nodes = data$nodes, edges = data$edges, height = "500px") |>
    visPhysics(stabilization=FALSE)
```

Add arrows and colors to better capture dependencies.

```{r}
data$edges$color='lightblue'
data$edges[data$edges$edgetype=='Imports','color']= 'red'
data$edges[data$edges$edgetype=='Depends','color']= 'green'

visNetwork(nodes = data$nodes, edges = data$edges, height = "500px") |>
    visEdges(arrows='from') 
```

Add a legend.

```{r}
ledges <- data.frame(color = c("green", "lightblue", "red"),
  label = c("Depends", "Suggests", "Imports"), arrows =c("from", "from", "from"))
visNetwork(nodes = data$nodes, edges = data$edges, height = "500px") |>
  visEdges(arrows='from') |>
  visLegend(addEdges=ledges)
```

## Integration with `r Biocpkg("BiocViews")`

[Work in progress]

The `r Biocpkg("biocViews")` package is a small ontology of terms describing
Bioconductor packages. This is a work-in-progress section, but here is a small
example of plotting the biocViews graph.

```{r biocViews}
library(biocViews)
data(biocViewsVocab)
biocViewsVocab
library(igraph)
g = graph_from_graphnel(biocViewsVocab)
library(visNetwork)
gv <- toVisNetworkData(g)
visNetwork(gv$nodes, gv$edges, width="100%") |>
    visIgraphLayout(layout = "layout_as_tree", circular=TRUE) |>
    visNodes(size=20) |>
    visPhysics(stabilization=FALSE)
```

# Dependency burden

The dependency burden of a package, namely the amount of functionality that
a given package is importing, is an important parameter to take into account
during package development. A package may break because one or more of its
dependencies have changed the part of the API our package is importing or
this part has even broken. For this reason, it may be useful for package
developers to quantify the dependency burden of a given package. To do that
we should first gather all dependency information using the function
`buildPkgDependencyDataFrame()` but setting the arguments to work with
packages in Bioconductor and CRAN and dependencies categorised as `Depends`
or `Imports`, which are the ones installed by default for a given package.

```{r}

depdf <- buildPkgDependencyDataFrame(repo=c("BioCsoft", "CRAN"),
                                     dependencies=c("Depends", "Imports"))
dim(depdf)
head(depdf)  # too big to show all
```

Finally, we call the function `pkgDepMetrics()` to obtain different metrics
on the dependency burden of a package we want to analyze, in the case below,
the package `BiocPkgTools` itself:

```{r}
pkgDepMetrics("BiocPkgTools", depdf)
```

In this resulting table, rows correspond to dependencies and columns provide
the following information:

  * `ImportedAndUsed`: number of functionality calls imported and used in
    the package.
  * `Exported`: number of functionality calls exported by the dependency.
  * `Usage`: (`ImportedAndUsed`x 100) / `Exported`. This value provides an
    estimate of what fraction of the functionality of the dependency is
    actually used in the given package.
  * `DepOverlap`: Similarity between the dependency graph structure of the
    given package and the one of the dependency in the corresponding row,
    estimated as the
    [Jaccard index](https://en.wikipedia.org/wiki/Jaccard_index)
    between the two sets of vertices of the corresponding graphs. Its values
    goes between 0 and 1, where 0 indicates that no dependency is shared, while
    1 indicates that the given package and the corresponding dependency depend
    on an identical subset of packages.
  * `DepGainIfExcluded`: The 'dependency gain' (decrease in the total number 
    of dependencies) that would be obtained if this package was excluded 
    from the list of direct dependencies. 

The reported information is ordered by the `Usage` column to facilitate the
identification of dependencies for which the analyzed package is using a small
fraction of their functionality and therefore, it could be easier remove them.
To aid in that decision, the column `DepOverlap` reports the overlap of the
dependency graph of each dependency with the one of the analyzed package. Here
a value above, e.g., 0.5, could, albeit not necessarily, imply that removing
that dependency could substantially lighten the dependency burden of the
analyzed package.

An `NA` value in the `ImportedAndUsed` column indicates that the function
`pkgDepMetrics()` could not identify what functionality calls in the analyzed
package are made to the dependency. This may happen because `pkgDepMetrics()`
has failed to identify the corresponding calls, as it happens with imported
built-in constants such as `DNA_BASES` from `Biostrings`, or that although the
given package is importing that dependency, none of its functionality is
actually being used. In such a case, this dependency could be safely removed
without any further change in the analyzed package.

We can find out what actually functionality calls are we importing as follows:

```{r}
imp <- pkgDepImports("BiocPkgTools")
imp |> filter(pkg == "DT")
```

# Identities of maintainers

It is important to be able to identify the maintainer of a package in
a reliable way.  The DESCRIPTION file for a package can include an `Authors@R`
field.  This field can capture metadata about maintainers and contributors
in a programmatically accessible way.  Each element of the role
field of a `person` (see `?person`) in the `Authors@R` field comes
from a subset of the 
[relations vocabulary of the Library of Congress](https://www.loc.gov/marc/relators/relaterm.html).  

Metadata about maintainers can be extracted from DESCRIPTION in various ways.
As of October 2022, we focus on the use of the ORCID field which is
an optional `comment` component in a `person` element.  For example,
in the DESCRIPTION for the AnVIL package we have
```
Authors@R:
    c(person(
        "Martin", "Morgan", role = c("aut", "cre"),
        email = "mtmorgan.bioc@gmail.com",
        comment = c(ORCID = "0000-0002-5874-8148")
     ),
```
This convention is used for a fair number of Bioconductor and CRAN packages.

We'll demonstrate the use of `get_cre_orcids` with some packages.
```{r demoorc}
inst = rownames(installed.packages())
cands = c("devtools", "evaluate", "ggplot2", "GEOquery", "gert", "utils")
totry = intersect(cands, inst)
oids = get_cre_orcids(totry)
oids
```

We use the ORCID API to tabulate metadata about the holders of these IDs.
We'll avoid evaluating this because a token must be refreshed for the query
to succeed.

```{r lktab,eval=FALSE}
orcid_table(oids)
```
In October 2022 the result is
```
> orcid_table(.Last.value)
                        name                                            org
devtools      Jennifer Bryan                                        RStudio
evaluate           Yihui Xie                                  RStudio, Inc.
ggplot2  Thomas Lin Pedersen                                        RStudio
GEOquery          Sean Davis University of Colorado Anschutz Medical Campus
gert             Jeroen Ooms            Berkeley Institute for Data Science
utils                   <NA>                                           <NA>
               city   region country               orcid
devtools     Boston       MA      US 0000-0002-6983-2759
evaluate    Elkhorn       NE      US 0000-0003-0645-5666
ggplot2  Copenhagen     <NA>      DK 0000-0002-5147-4711
GEOquery     Aurora Colorado      US 0000-0002-8991-6458
gert       Berkeley       CA      US 0000-0002-4035-0289
utils          <NA>     <NA>    <NA>                <NA>
```

# Provenance

```{r}
sessionInfo()
```

---
title: "Package(s) Deprecation Notification"
author:
- name: "%s"
  affiliation:
  - Roswell Park Comprehensive Cancer Center, Buffalo, NY, USA
date: "%s"
output:
  BiocStyle::html_document:
    toc: false
    number_sections: false
---

## Deprecated Package(s): %s

---


Hi Maintainers,

Unfortunately, due to the deprecation of the %s package(s), your packages
have been affected. Please remove the facilities of the deprecated
packages from your packages. This includes the removal of imported functions
and the package from your `DESCRIPTION` file.

Here is the full list of affected Bioconductor packages:

%s

We apologize for the inconvenience and if you have any questions please
direct them to the mailing list at bioc-devel@bioconductor.org.

%s

Best regards,

%s

---

<footer>
<p>Bioconductor Core Team
<br>Roswell Park Comprehensive Cancer Center
<br>Dept. of Biostatistics & Bioinformatics
<br>Elm & Carlton Streets
<br>Buffalo, New York 14263
</p>
</footer>

---
title: "%s Bioconductor Package Notification"
author:
- name: "%s"
  affiliation:
  - Roswell Park Comprehensive Cancer Center, Buffalo, NY, USA
date: "%s"
output:
  BiocStyle::html_document:
    toc: false
    number_sections: false
---

## Package: %s

---


Hi %s,

I am contacting you concerning the `%s` package, on the %s
branch of Bioconductor, for which you are listed as the maintainer.

The latest build/check report shows that your package is failing on one or more
platforms.
Please resolve the issues before the next [Bioconductor release](https://www.bioconductor.org/developers/release-schedule/)
for your package's continued inclusion in Bioconductor.

%s

Please let me know if you have any questions.

Regards,

%s


%s

---

<footer>
<p>Bioconductor Core Team
<br>Roswell Park Comprehensive Cancer Center
<br>Dept. of Biostatistics & Bioinformatics
<br>Elm & Carlton Streets
<br>Buffalo, New York 14263
</p>
</footer>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildPkgDependencyGraph.R
\name{inducedSubgraphByPkgs}
\alias{inducedSubgraphByPkgs}
\title{Return a minimal subgraph based on package name(s)}
\usage{
inducedSubgraphByPkgs(g, pkgs, pkg_color = "red")
}
\arguments{
\item{g}{an igraph graph, typically created by
\code{\link[=buildPkgDependencyIgraph]{buildPkgDependencyIgraph()}}}

\item{pkgs}{\code{character()} vector of packages to
include. Package names not included in
the graph are ignored.}

\item{pkg_color}{\code{character(1)} giving color of named
packages. Other packages in the graph that fall in
connecting paths will be colored as the igraph default.}
}
\description{
Find the subgraph induced by including
specific packages. The induced subgraph is
the graph that includes the named packages
and all edges connecting them. This is useful
for a developer, for example, to examine her packages
and their intervening dependencies.
}
\examples{
library(igraph)
g <- buildPkgDependencyIgraph(buildPkgDependencyDataFrame())
## subgraph of only the first 10 packages maintained by Bioconductor
biocmaintained <- head(biocMaintained()[["Package"]], 10L)
g2 <- inducedSubgraphByPkgs(g, pkgs = biocmaintained)
g2
V(g2)

plot(g2)

## subgraph of a package's strong Bioconductor package dependencies
maedeps <- unlist(pkgBiocDeps(
    "MultiAssayExperiment", which = "strong",
    recursive = TRUE, only.bioc = TRUE
), use.names = FALSE)
g3 <- inducedSubgraphByPkgs(g, pkgs = maedeps)
plot(g3)

## same subgraph with networkD3::forceNetwork
library(networkD3)
wt <- cluster_walktrap(g3)
members <- membership(wt)
ndg3 <- igraph_to_networkD3(g3, group = members)
forceNetwork(
    Links = ndg3$links, Nodes = ndg3$nodes, Source = 'source',
    Target = 'target', NodeID = 'name', Group = 'group', zoom = TRUE,
    linkDistance = 200, fontSize = 20, opacity = 0.9, opacityNoHover = 0.9
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgBiocDeps.R
\name{pkgBiocRevDeps}
\alias{pkgBiocRevDeps}
\alias{summary.biocrevdeps}
\title{Obtain all the reverse dependencies for a Bioconductor package}
\usage{
pkgBiocRevDeps(
  pkg,
  pkgType = c("software", "data-experiment", "workflows", "data-annotation"),
  which = "strong",
  only.bioc = TRUE,
  version = BiocManager::version(),
  recursive = FALSE
)

\method{summary}{biocrevdeps}(object, ...)
}
\arguments{
\item{pkg}{\code{character(1)} The package for which to look up dependencies.}

\item{pkgType}{\code{character()} Any of 'software', 'data-experiment',
'workflows', and / or 'data-annotation' (defaults to all)}

\item{which}{a character vector listing the types of
    dependencies, a subset of
    \code{c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")}.
    Character string \code{"all"} is shorthand for that vector,
    character string \code{"most"} for the same vector without
    \code{"Enhances"},
    character string \code{"strong"} (default) for the first three
    elements of that vector.
  }

\item{only.bioc}{\code{logical(1)} Whether to only return Bioconductor
dependencies in the list (default \code{TRUE})}

\item{version}{(Optional) \code{character(1)} or \code{package_version}
indicating the \emph{Bioconductor} version (e.g., "3.8") for which
repositories are required.}

\item{recursive}{a logical indicating whether (reverse) dependencies
    of (reverse) dependencies (and so on) should be included, or a
    character vector like \code{which} indicating the type of (reverse)
    dependencies to be added recursively.}

\item{object}{an object for which a summary is desired.}

\item{...}{additional arguments affecting the summary produced.}
}
\value{
A \code{biocrevdeps} list class object
}
\description{
The function returns a slightly upgraded list with dependency types as
elements and package names in each of those elements, if any. The
types of dependencies can be seen in the \code{which} argument documentation.
}
\details{
The summary method of the \code{biocrevdeps} class given by
\code{pkgBiocRevDeps} provides a tally in each dependency field.
}
\examples{
rdeps <- pkgBiocRevDeps("MultiAssayExperiment", which = "all")
rdeps
summary(rdeps)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocPkgList.R
\name{.extract_fnd}
\alias{.extract_fnd}
\title{Extract funder names from a formatted \code{Author} field string}
\usage{
.extract_fnd(author_string)
}
\arguments{
\item{author_string}{character(1) a formatted \code{Author} field value as
found in a Bioconductor \code{VIEWS} file, e.g.
\code{"Jane Doe [aut, cre], Big Org [fnd]"}.  May contain embedded
newlines.  Note: funder names that themselves contain commas will not be
extracted correctly, as commas are used as entry separators in the
formatted \code{Author} field.}
}
\value{
character() of funder names, or \code{NA_character_} when no funder
is found or the input is missing/empty
}
\description{
Extract funder names from a formatted \code{Author} field string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocDownloadStats.R
\name{activitySince}
\alias{activitySince}
\title{What are the issues, pulls, commits created since a date?}
\usage{
activitySince(
  gh_repo,
  activity = c("issues", "pulls", "commits"),
  status = c("closed", "open", "all"),
  Date,
  issue_metadata = c("created_at", "number", "title"),
  token = NULL
)
}
\arguments{
\item{gh_repo}{\code{character(1)} The GitHub repository location including the
username / organization and the repository name, e.g.,
"Bioconductor/S4Vectors"}

\item{activity}{\code{character(1)} The type of repository activity to pull from the
GitHub API. It can be one of "issues" (default), "pulls", or "commits".}

\item{status}{\code{character(1)} One of 'closed', 'open', or 'all' corresponding to
the issue state desired from the GitHub API (Default: "closed"). This
argument is ignored for the "commits" activity report.}

\item{Date}{\code{character(1)} The date cutoff from which to analyze closed issues
in the YYYY-MM-DD or YYYY-MM-DDTHH:MM:SSZ format (ISO 8601).}

\item{issue_metadata}{\code{character()} The metadata labels to extract from the
\code{gh::gh} response. See \code{?gh::gh} for more details. Defaults to
'created_at', 'number', and 'title'. This argument is ignored for the
"commits" activity report.}

\item{token}{\code{character(1)} For big requests, e.g., commit history, you may be
prompted to use a GitHub Personal Access Token. Enter the token as plain
text.}
}
\value{
A \code{tibble} with three columns corresponding to issue metadata (i.e.,
"created_at", "number", "title")
}
\description{
This function uses the \code{gh} package to get a list of either issues, pull
requests, or GitHub commits since the specified date for a particular GitHub
repository. The repository must have both the username / organization and the
name, e.g., "Bioconductor/S4Vectors".
}
\details{
The \code{tibble} returned by the commits activity report contains five
columns:
\itemize{
\item 'committer_date'
\item 'commit' - hash
\item 'parents' - hash of parent for merge commits
\item 'author'
\item 'message'
}

For information on other columns, refer to the GitHub API under repository
issues or pulls (e.g., \verb{/repos/:repo/issues}).
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
activitySince("Bioconductor/S4Vectors", "issues", "closed", "2021-05-01")
activitySince("Bioconductor/S4Vectors", "issues", "open", "2022-05-01")
activitySince("Bioconductor/S4Vectors", "commits", Date = "2022-05-01")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDependencyMetrics.R
\name{pkgDepImports}
\alias{pkgDepImports}
\title{Report package imported functionality}
\usage{
pkgDepImports(pkg)
}
\arguments{
\item{pkg}{\code{character()} name of the package for which we want
to obtain the functionality calls imported from its dependencies
and used within the package.}
}
\value{
A tidy data frame with two columns:
\itemize{
\item \code{pkg}: name of the package dependency.
\item \code{fun}: name of the functionality call imported from the
the dependency in the column \code{pkg} and used within
the analyzed package.
}
}
\description{
Function adapted from 'itdepends::dep_usage_pkg' at https://github.com/r-lib/itdepends
to obtain the functionality imported and used by a given package.
}
\details{
Certain imported elements, such as built-in constants, will not
be identified as imported functionality by this function.
}
\examples{
pkgDepImports('BiocPkgTools')

}
\author{
Robert Castelo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocDownloadStats.R
\name{pkgDownloadRank}
\alias{pkgDownloadRank}
\title{What is a package's download rank?}
\usage{
pkgDownloadRank(
  pkg,
  pkgType = c("software", "data-experiment", "workflows", "data-annotation"),
  version = BiocManager::version()
)
}
\arguments{
\item{pkg}{\code{character(1)} The name of a Bioconductor package}

\item{pkgType}{\code{character(1)} One of 'software', 'data-experiment',
'workflows', or 'data-annotation' (defaults to 'software')}

\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}
}
\value{
The package's percentile rank, in terms of download statistics, and
proportion in the name
}
\description{
This function uses \code{available.packages} to calculate the download rank
\emph{percentile} of a given package. It approximates what is observed
in the Bioconductor landing page.
}
\examples{
## Percentile rank for BiocGenerics (top 1\%)
pkgDownloadRank("BiocGenerics", "software")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getYearsInBioc.R
\name{getPkgYearsInBioc}
\alias{getPkgYearsInBioc}
\title{Calculate the years in Bioconductor}
\usage{
getPkgYearsInBioc(pkglist = NULL)
}
\arguments{
\item{pkglist}{List of packages to retrieve information. If default NULL,
returns a tibble of all Bioconductor packages.}
}
\value{
'tibble' with the following columns:
\itemize{
\item package: name of Bioconductor package
\item category: bioc, data/experiment, data/annotation, workflow
\item first_version_available: Bioconductor version (e.g. 1.9, 3.21) the
package first became available
\item first_version_release_date: Equivalent calendar date of given
Bioconductor release
\item approx_years_in: Numeric indicator of years in Bioconductor. If empty,
indicates package was removed. See final three columns for more
information.
\item last_version_available: If package was removed from Bioconductor, the
last Bioconductor version (e.g. 1.9, 3.21) the package was able to be
installed
\item last_version_release_date: Equivalent calendar date of given Bioconductor
release
\item years_before_rm: If removed, how many years it was in Bioconductor
}
}
\description{
This function determines the number of years a package has been
in Bioconductor. Available information includes first Bioconductor version
a package appeared and the current length of time in Bioconductor. If a
package has been removed from Bioconductor, information on the last
Bioconductor version and approximate time in Bioconductor before removal is
available.
}
\examples{
\dontrun{
   ## full table all Bioconductor packages
   tbl <- getPkgYearsInBioc()

   ## example of package list. Packages active in Bioconductor
   tbl <- getPkgYearsInBioc(c("BiocFileCache", "BiocPkgTools"))

   ## example of a package that has been removed from Bioconductor
   tbl <- getPkgYearsInBioc("ensemblVEP")
}

}
\author{
Lori Shepherd Kern, Robert Shear
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/githubUtils.R
\name{githubDetails}
\alias{githubDetails}
\title{Get package details from GitHub}
\usage{
githubDetails(pkgs, sleep = 0)
}
\arguments{
\item{pkgs}{a \code{character()} vector of username/repo
for one or more GitHub repos, such as \code{seandavi/GEOquery}.}

\item{sleep}{\code{numeric()} denoting the number of seconds to
sleep between GitHub API calls. Since GitHub rate limits
its APIs, it might be necessary to either use small
chunks of packages iteratively or to supply a non-zero
argument here. See the \code{details} section for a better
solution using GitHub tokens.}
}
\description{
For packages that live on GitHub, we can mine
further details. This function returns the
GitHub details for the listed packages.
}
\details{
The \code{\link[gh:gh]{gh::gh()}} function is used to
do the fetching. If the number of packages supplied
to this function is large (>40 or so), it is possible
to run into problems with API rate limits. The \code{gh}
package uses the environment variable "GITHUB_PAT"
(for personal access token) to authenticate and then
provide higher rate limits. If you run into problems
with rate limits, set sleep to some small positive
number to slow queries. Alternatively, create a Personal
Access Token on GitHub and register it. See the \code{gh}
package for details.
}
\examples{
pkglist = biocPkgList()

# example of "pkgs" format.
head(pkglist$URL)

gh_list = githubURLParts(pkglist$URL)
gh_list = gh_list[!is.null(gh_list$user_repo),]

head(gh_list$user_repo)

ghd = githubDetails(gh_list$user_repo[1:5])
lapply(ghd, '[[', "stargazers")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDependencyMetrics.R
\name{.getDepGain}
\alias{.getDepGain}
\title{Calculate the 'dependency gain' from excluding one or more direct
dependencies}
\usage{
.getDepGain(g, pkg, depsToRemove)
}
\arguments{
\item{g}{Package dependency graph}

\item{pkg}{Character string representing the package of interest}

\item{depsToRemove}{Character vector representing the dependencies
to remove}
}
\value{
The 'dependency gain' that would be achieved by excluding the
indicated direct dependencies
}
\description{
Calculate the difference between the total number of dependencies of a
package and the number of dependencies that would remain if one or more
of the direct dependencies were removed.
}
\author{
Charlotte Soneson
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/problemPage.R
\name{problemPage}
\alias{problemPage}
\title{generate hyperlinked HTML for build reports for Bioc packages}
\usage{
problemPage(
  authorPattern = "V.*Carey",
  dependsOn,
  ver = "devel",
  includeOK = FALSE
)
}
\arguments{
\item{authorPattern}{\code{character(1)} regexp used with grep() to filter author field of package DESCRIPTION for listing}

\item{dependsOn}{\code{character(1)} name of a Bioconductor package. The function will return the status of packages
that directly depend on this package  Can only be used when 'authorPattern' is the empty string.}

\item{ver}{\code{character(1)} version tag for Bioconductor}

\item{includeOK}{\code{logical(1)} include entries from the build report that are listed
as "OK". Default FALSE will result in only those entries that
are in WARNING or ERROR state.}
}
\value{
DT::datatable call; if assigned to a variable, must evaluate to get the page to appear
}
\description{
This is a quick way to get an HTML report of packages maintained by a specific developer
or which depend directly on a specified package. The function is keyed to filter based on either
the maintainer name or by using the 'Depends', 'Suggests' and 'Imports' fields in package descriptions.
}
\examples{
if (interactive()) {
  problemPage()
  problemPage(dependsOn = "limma")
}

}
\author{
Vince Carey, Mike L. Smith
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{BiocPkgTools-cache}
\alias{BiocPkgTools-cache}
\alias{setCache}
\alias{pkgToolsCache}
\title{Manage cache for BiocPkgTools}
\usage{
setCache(
  directory = tools::R_user_dir("BiocPkgTools", "cache"),
  verbose = TRUE,
  ask = interactive()
)

pkgToolsCache(...)
}
\arguments{
\item{directory}{The file location where the cache is located. Once set
future downloads will go to this folder.}

\item{verbose}{Whether to print descriptive messages}

\item{ask}{logical (default TRUE when interactive session) Confirm the file
location of the cache directory}

\item{...}{For \code{pkgToolsCache}, arguments are passed to \code{setCache}}
}
\description{
Managing user data is important to allow use of email functions
such as \code{biocBuildEmail} and made easy with \code{BiocFileCache}.
}
\section{pkgToolsCache}{

Get the directory location of the cache. It will prompt the user to create
a cache if not already created. A specific directory can be used via
\code{setCache}.
}

\section{setCache}{

Specify the directory location of the data cache. By default, it will
got to the user's home/.cache/R and "appname" directory as specified by
\code{tools::R_user_dir} (with package="BiocPkgTools" and which="cache").
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocDownloadStats.R
\name{latestPkgStats}
\alias{latestPkgStats}
\title{Summary of the latest package statistics}
\usage{
latestPkgStats(
  gh_repo,
  Date,
  pkgType = c("software", "data-experiment", "workflows", "data-annotation")
)
}
\arguments{
\item{gh_repo}{\code{character(1)} The GitHub repository location including the
username / organization and the repository name, e.g.,
"Bioconductor/S4Vectors"}

\item{Date}{\code{character(1)} The date cutoff from which to analyze closed issues
in the YYYY-MM-DD or YYYY-MM-DDTHH:MM:SSZ format (ISO 8601).}

\item{pkgType}{\code{character(1)} One of 'software', 'data-experiment',
'workflows', or 'data-annotation' (defaults to 'software')}
}
\description{
The \code{latestPkgStats} function combines outputs from several functions to
generate a table of relevant statistics for a given package.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
latestPkgStats("Bioconductor/BiocGenerics", "2021-05-05")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocBuildStatusDB.R
\name{biocBuildStatusDB}
\alias{biocBuildStatusDB}
\title{Download and parse the build status information for Bioconductor packages}
\usage{
biocBuildStatusDB(
  version = BiocManager::version(),
  pkgType = c("software", "data-experiment", "data-annotation", "workflows")
)
}
\arguments{
\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{pkgType}{\code{character()} The type of packages for which to get
information for. Values can be one or more of the following:
\itemize{
\item \code{software}: Software packages
\item \code{data-experiment}: Experiment data packages
\item \code{data-annotation}: Annotation data packages
\item \code{workflows}: Workflow packages
}}
}
\value{
A \code{data.frame} with the following columns:
\itemize{
\item pkg: The name of the package
\item node: The builder on which the package was built
\item stage: The stage of the build, e.g., 'install', 'buildsrc', 'checksrc',
etc.
\item result: The status of the build, e.g., 'OK', 'ERROR', 'WARNINGS', etc.
}
}
\description{
This function downloads and parses the build status information
for Bioconductor packages. The build status information is available for
the current release and the previous release. Other versions may be
available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/githubUtils.R
\name{githubURLParts}
\alias{githubURLParts}
\title{Extract GitHub user and repo name from GitHub URL}
\usage{
githubURLParts(urls)
}
\arguments{
\item{urls}{\code{character()} A vector of URLs}
}
\value{
A \code{data.frame} with four columns:
\itemize{
\item url: The original GitHub URL
\item user_repo: The GitHub "username/repo", combined
\item user:  The GitHub username
\item repo: The GitHub repo name
}
}
\description{
Extract GitHub user and repo name from GitHub URL
}
\examples{
# find GitHub URL details for
# Bioconductor packages
bpkgl = biocPkgList()
urldetails = githubURLParts(bpkgl$URL)
urldetails = urldetails[!is.na(urldetails$url),]
head(urldetails)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocBuildEmail.R
\name{templatePath}
\alias{templatePath}
\title{Obtain the location of available email templates}
\usage{
templatePath(
  type = c("buildemail", "deprecation", "deprecguide", "cranreport", "revdepnote")
)
}
\arguments{
\item{type}{\code{character(1)} Either one of "buildemail", "deprecation",
"deprecguide", "cranreport", or "revdepnote". See the templates in the
\code{resources} folder.}
}
\description{
These templates are used with \code{biocBuildEmail} to notify maintainers
regarding package errors and final deprecation warning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgBiocDeps.R
\name{pkgBiocDeps}
\alias{pkgBiocDeps}
\title{Look up a package's Bioconductor dependencies}
\usage{
pkgBiocDeps(
  pkg,
  pkgType = c("software", "data-experiment", "workflows", "data-annotation"),
  which = "strong",
  only.bioc = TRUE,
  recursive = FALSE,
  version = BiocManager::version()
)
}
\arguments{
\item{pkg}{\code{character(1)} The package for which to look up dependencies.}

\item{pkgType}{\code{character()} Any of 'software', 'data-experiment',
'workflows', and / or 'data-annotation' (defaults to all)}

\item{which}{a character vector listing the types of
    dependencies, a subset of
    \code{c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")}.
    Character string \code{"all"} is shorthand for that vector,
    character string \code{"most"} for the same vector without
    \code{"Enhances"},
    character string \code{"strong"} (default) for the first three
    elements of that vector.
  }

\item{only.bioc}{\code{logical(1)} Whether to only return Bioconductor
dependencies in the list (default \code{TRUE})}

\item{recursive}{a logical indicating whether (reverse) dependencies
    of (reverse) dependencies (and so on) should be included, or a
    character vector like \code{which} indicating the type of (reverse)
    dependencies to be added recursively.}

\item{version}{(Optional) \code{character(1)} or \code{package_version}
indicating the \emph{Bioconductor} version (e.g., "3.8") for which
repositories are required.}
}
\description{
The function uses the \code{pkgType} argument to restrict the look up to only the
relevant Bioconductor repository. It works for multiple packages of the same
type.
}
\examples{

pkgBiocDeps("MultiAssayExperiment", only.bioc = TRUE)

pkgBiocDeps("MultiAssayExperiment", only.bioc = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BiocPkgDOI_mint_by_REST.R
\name{generateBiocPkgDOI_REST}
\alias{generateBiocPkgDOI_REST}
\title{Generate a DOI for a Bioconductor package}
\usage{
generateBiocPkgDOI_REST(pkg, authors, pubyear, testing = TRUE)
}
\arguments{
\item{pkg}{character(1) package name}

\item{authors}{character vector of authors (will be "pasted" together)}

\item{pubyear}{integer(1) publication year}

\item{testing}{logical(1) If true, will use the apitest
user with the password apitest. These DOIs will expire.
The same apitest:apitest combination can be used to
login to the EZID website for doing things using the
web interface. If false, the Bioconductor-specific
user credentials should be in the correct environment
variables}
}
\value{
The DOI as a character(1) vector.
}
\description{
This function makes calls out to the REST API (v2) described
here: \url{https://support.datacite.org/docs/api-create-dois}.  This
function creates a new DOI for a Bioconductor package (cannot already
exist). The target URL for the DOI is the short Bioconductor
package URL.
}
\details{
The login information for the "real" Bioconductor account
should be stored in the environment variables "EZID_USERNAME"
and "EZID_PASSWORD".

The GUI is available here: \url{https://doi.datacite.org/}.
}
\examples{
\dontrun{
  x = generateBiocPkgDOI('RANDOM_TEST_PACKAGE','Sean Davis',1972)
}
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/condaDownloadStats.R
\name{anacondaDownloadStats}
\alias{anacondaDownloadStats}
\title{Get download statistics for Bioconductor packages distributed via
Anaconda.}
\usage{
anacondaDownloadStats()
}
\value{
A \code{data.frame} of download statistics for
all Bioconductor packages distributed by Anaconda, in tidy format.
Note: Anaconda do not provide counts for unique IP addresses.  This column
is listed as \code{NA} for all packages to provide continuity with data from
Bioconductor.org obtained by \code{\link[=biocDownloadStats]{biocDownloadStats()}}.  The counts are
updated monthly, so do not expect to see counts for the current month.
}
\description{
Get download statistics for Bioconductor packages distributed via
Anaconda.
}
\details{
Anaconda provide daily download counts for all software packages
they distribute.  These are summarised into monthly tables of counts and made
available from https://github.com/grimbough/anaconda-download-stats
This function provides a mechanism to download these monthly counts for
Bioconductor packages distributed through Anaconda.
}
\examples{
anacondaDownloadStats()

}
\author{
Mike L. Smith
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocRevDepEmail.R
\name{biocRevDepEmail}
\alias{biocRevDepEmail}
\title{Notify downstream maintainers of changes in upstream packages}
\usage{
biocRevDepEmail(
  packages,
  which = c("strong", "most", "all"),
  PS = character(1L),
  version = BiocManager::version(),
  dry.run = TRUE,
  cc = NULL,
  emailTemplate = templatePath("revdepnote"),
  core.name = NULL,
  core.email = NULL,
  core.id = NULL,
  textOnly = FALSE,
  verbose = FALSE,
  credFile = "~/.blastula_creds",
  ...,
  pkg
)
}
\arguments{
\item{packages}{\code{character()} A vector of CRAN and/or Bioconductor packages
for whose reverse dependencies are to be checked and notified.}

\item{which}{a character vector listing the types of
    dependencies, a subset of
    \code{c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances")}.
    Character string \code{"all"} is shorthand for that vector,
    character string \code{"most"} for the same vector without
    \code{"Enhances"},
    character string \code{"strong"} (default) for the first three
    elements of that vector.
  }

\item{PS}{\code{character(1)} Postscript, an additional note to the recipient of
the email (i.e., the package maintainer)}

\item{version}{\code{character()} A vector indicating which version of Bioconductor
the package is failing in (either 'release' or 'devel'; defaults to both)}

\item{dry.run}{\code{logical(1)} Display the email without sending to the recipient.
It only works for HTML email reports and ignored when \code{textOnly=TRUE}}

\item{cc}{\code{character()} A vector of email addresses for sending the message
as a carbon copy.}

\item{emailTemplate}{\code{character(1)} The path to the email template Rmd file as
obtained by \code{templatePath()}. A custom template can be provided as file
path.}

\item{core.name}{\code{character(1)} The full name of the core team member}

\item{core.email}{\code{character(1)} The Roswell Park email of the core team
member}

\item{core.id}{\code{character(1)} The internal identifier for the Roswell employee.
This ID usually matches \verb{^[A-Z]\{2\}[0-9]\{5\}} for more recent identifiers.}

\item{textOnly}{\code{logical(1)} Whether to return the text of the email only.
This avoids the use of the 'blastula' package and adds the text to the
system clipboard if the \code{clipr} package is installed (default: FALSE)}

\item{verbose}{\code{logical(1)} Whether to output full email information from
'smtp_send' (when \code{dry.run} is \code{FALSE} and 'blastula' is installed)}

\item{credFile}{\code{character(1)} An optional file generated by the
\code{blastula::create_smtp_creds_file} function containing email
authentication information (default: "~/.blastula_creds"). See
\code{?biocBuildEmail} details.}

\item{pkg}{\code{character(1)} DEPRECATED. The name of a single package whose
reverse dependencies are to be checked and notified.}

\item{\ldots}{Additional inputs to internal functions (not used).}
}
\description{
The \code{biocRevDepEmail} function collects all the emails of the reverse
dependencies and sends a notification that upstream package(s) have been
deprecated or removed. It uses a template found in \code{inst/resources} with the
\code{templatePath()} function.
}
\examples{

biocRevDepEmail(
    "FindMyFriends", version = "3.13", dry.run = TRUE, textOnly = TRUE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildPkgDependencyGraph.R
\name{subgraphByDegree}
\alias{subgraphByDegree}
\title{Subset graph by degree}
\usage{
subgraphByDegree(g, pkg, degree = 1, ...)
}
\arguments{
\item{g}{an igraph graph, typically created by
\code{\link[=buildPkgDependencyIgraph]{buildPkgDependencyIgraph()}}}

\item{pkg}{\code{character(1)} package name from which to
measure degree.}

\item{degree}{integer(1) degree, limit search for
adjacent vertices to this degree.}

\item{...}{passed on to \code{\link[igraph:distances]{igraph::distances()}}}
}
\value{
An igraph graph, with only nodes and their
edges within degree of the named package
}
\description{
While the \code{\link[=inducedSubgraphByPkgs]{inducedSubgraphByPkgs()}}
returns the subgraph with the minimal connections
between named packages, this function takes a vector of
package names, a degree (1 or more) and returns the
subgraph(s) that are within \code{degree} of the
package named.
}
\examples{

g <- buildPkgDependencyIgraph(buildPkgDependencyDataFrame())
g2 <- subgraphByDegree(g, 'GEOquery')
plot(g2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocBuildReport.R
\name{biocBuildReport}
\alias{biocBuildReport}
\title{Tidy Bioconductor build report results}
\usage{
biocBuildReport(
  version = BiocManager::version(),
  pkgType = c("software", "data-experiment", "data-annotation", "workflows"),
  stage.timings = FALSE
)
}
\arguments{
\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{pkgType}{\code{character()} The type of packages for which to get
information for. Values can be one or more of the following:
\itemize{
\item \code{software}: Software packages
\item \code{data-experiment}: Experiment data packages
\item \code{data-annotation}: Annotation data packages
\item \code{workflows}: Workflow packages
}}

\item{stage.timings}{\code{logical(1)} Whether to include the start, end, and
elapsed time for each build, check, install stage from each building in
the result (default: FALSE)}
}
\value{
A \code{tbl_df} object with columns pkg, version,
author, commit, date, node, stage, and result.
}
\description{
The online Bioconductor build reports
are great for humans to look at, but
they are not easily computable. This function
scrapes HTML and text files available
from the build report online pages to generate
a tidy data frame version of the build report.
}
\examples{

# Set the stage--what version of Bioc am I using?
BiocManager::version()

latest_build <- biocBuildReport()
head(latest_build)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocPkgRanges.R
\name{biocPkgRanges}
\alias{biocPkgRanges}
\title{Grab build report results from BUILD_STATUS_DB for a particular package range}
\usage{
biocPkgRanges(
  start,
  end,
  condition = c("ERROR", "WARNINGS"),
  phase = "buildsrc",
  version = c("devel", "release")
)
}
\arguments{
\item{start}{\code{character(1)} alphabetically first package name in range}

\item{end}{\code{character(1)} alphabetically last package name in range}

\item{condition}{\code{character(1)} condition string, typically 'ERROR' or
'WARNING'}

\item{phase}{\code{character(1)} string for phase of event:
'install', 'checksrc', or 'buildsrc' (default)}

\item{version}{\code{character(1)} string indication Bioconductor version, either
'devel' (default) or 'release'}
}
\description{
Grab build report results from BUILD_STATUS_DB for a particular package range
}
\examples{
\dontrun{
biocPkgRanges(
    start = "a4", end = "CMA",
    condition = "ERROR", version = "devel"
)
}
}
\author{
Vincent J. Carey
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDependencyMetrics.R
\name{pkgDepMetrics}
\alias{pkgDepMetrics}
\title{Report package dependency burden}
\usage{
pkgDepMetrics(pkg, depdf)
}
\arguments{
\item{pkg}{\code{character()} name of the package for which we want
to obtain metrics on its dependency burden.}

\item{depdf}{a tidy data frame with package dependency information
obtained through the function \code{\link[=buildPkgDependencyDataFrame]{buildPkgDependencyDataFrame()}}.}
}
\value{
A tidy data frame with different metrics on the
package dependency burden. More concretely, the following columns:
\itemize{
\item \code{ImportedAndUsed}: number of functionality calls imported and used in
the package.
\item \code{Exported}: number of functionality calls exported by the dependency.
\item \code{Usage}: (\code{ImportedAndUsed}x 100) / \code{Exported}. This value provides an
estimate of what fraction of the functionality of the dependency is
actually used in the given package.
\item \code{DepOverlap}: Similarity between the dependency graph structure of the
given package and the one of the dependency in the corresponding row,
estimated as the \href{https://en.wikipedia.org/wiki/Jaccard_index}{Jaccard index}
between the two sets of vertices of the corresponding graphs. Its values
goes between 0 and 1, where 0 indicates that no dependency is shared, while
1 indicates that the given package and the corresponding dependency depend
on an identical subset of packages.
\item \code{DepGainIfExcluded}: The 'dependency gain' (decrease in the total number
of dependencies) that would be obtained if this package was excluded
from the list of direct dependencies.
}

The reported information is ordered by the \code{Usage} column to facilitate the
identification of dependencies for which the analyzed package is using a small
fraction of their functionality and therefore, it could be easier remove them.
To aid in that decision, the column \code{DepOverlap} reports the overlap of the
dependency graph of each dependency with the one of the analyzed package. Here
a value above, e.g., 0.5, could, albeit not necessarily, imply that removing
that dependency could substantially lighten the dependency burden of the analyzed
package.

An \code{NA} value in the \code{ImportedAndUsed} column indicates that the function
\code{pkgDepMetrics()} could not identify what functionality calls in the analyzed
package are made to the dependency.
}
\description{
Elaborate a report on the dependency burden of a given package.
}
\examples{
depdf <- buildPkgDependencyDataFrame(
  dependencies=c("Depends", "Imports"),
  repo=c("BioCsoft", "CRAN")
)
pkgDepMetrics('BiocPkgTools', depdf)

}
\author{
Robert Castelo

Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_orcid.R
\name{.get_cre_orcid}
\alias{.get_cre_orcid}
\title{get the ORCID id from cre field of Authors@R in packageDescription result}
\usage{
.get_cre_orcid(pkgname)
}
\arguments{
\item{pkgname}{character(1)}
}
\description{
get the ORCID id from cre field of Authors@R in packageDescription result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_orcid.R
\name{get_cre_orcids}
\alias{get_cre_orcids}
\title{get ORCID ids from cre fields of Authors@R in packageDescription results}
\usage{
get_cre_orcids(pkgnames)
}
\arguments{
\item{pkgnames}{\code{character()} must be installed}
}
\description{
get ORCID ids from cre fields of Authors@R in packageDescription results
}
\note{
returns NA if no ORCID provided in Authors@R for package description
}
\examples{
get_cre_orcids(c("BiocPkgTools", "utils"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataciteXMLGenerate.R
\name{dataciteXMLGenerate}
\alias{dataciteXMLGenerate}
\title{The Bioconductor datacite.org XML generator}
\usage{
dataciteXMLGenerate(pkg)
}
\arguments{
\item{pkg}{The name of a Bioconductor package}
}
\value{
An xml_document object from the \code{xml2} package.
}
\description{
This function is used internally to generate XML elements from the
datacite.org website for incoming Bioconductor packages.
}
\seealso{
\preformatted{?xml2::`xml_document-class`}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CRANstatus.R
\name{CRANstatus}
\alias{CRANstatus}
\title{Check the CRAN build report page and email a notification}
\usage{
CRANstatus(
  pkg,
  core.name = NULL,
  core.email = NULL,
  core.id = NULL,
  to.mail = "maintainer@bioconductor.org",
  dry.run = TRUE,
  emailTemplate = templatePath("cranreport")
)
}
\arguments{
\item{pkg}{\code{character(1)} The name of the package in trouble}

\item{core.name}{\code{character(1)} The full name of the core team member}

\item{core.email}{\code{character(1)} The Roswell Park email of the core team
member}

\item{core.id}{\code{character(1)} The internal identifier for the Roswell employee.
This ID usually matches \verb{^[A-Z]\{2\}[0-9]\{5\}} for more recent identifiers.}

\item{to.mail}{The email of the CRAN report recipient}

\item{dry.run}{\code{logical(1)} Display the email without sending to the recipient.
It only works for HTML email reports and ignored when \code{textOnly=TRUE}}

\item{emailTemplate}{\code{character(1)} The path to the email template Rmd file as
obtained by \code{templatePath()}. A custom template can be provided as file
path.}
}
\description{
The \code{CRANstatus} function allows users to check the status of a package
and send an email report of any failures.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocDownloadStats.R
\name{firstInBioc}
\alias{firstInBioc}
\title{When did a package enter Bioconductor?}
\usage{
firstInBioc(download_stats)
}
\arguments{
\item{download_stats}{a data.frame from \code{\link[=biocDownloadStats]{biocDownloadStats()}}}
}
\description{
This function uses the biocDownloadStats
data to \emph{approximate} when a package entered
Bioconductor. Note that the download stats
go back only to 2009.
}
\examples{
dls <- biocDownloadStats()
tail(firstInBioc(dls))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocBuildReportDB.R
\name{biocBuildReportDB}
\alias{biocBuildReportDB}
\title{Parse the Build Report tarball for a Bioconductor release}
\usage{
biocBuildReportDB(
  version = BiocManager::version(),
  pkgType = c("software", "data-experiment", "data-annotation", "workflows"),
  stage.timings = FALSE
)
}
\arguments{
\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{pkgType}{\code{character()} The type of packages for which to get
information for. Values can be one or more of the following:
\itemize{
\item \code{software}: Software packages
\item \code{data-experiment}: Experiment data packages
\item \code{data-annotation}: Annotation data packages
\item \code{workflows}: Workflow packages
}}

\item{stage.timings}{\code{logical(1)} Whether to include the start, end, and
elapsed time for each build, check, install stage from each building in
the result (default: FALSE)}
}
\description{
This function parses the Build Report tarball for a Bioconductor
release. By default it will pull all the \code{report.tgz} files for each
Bioconductor package type. The Bioconductor Build System (BBS) Build Report
tarball contains build status information for all packages in a
Bioconductor release. This function is mainly used by \code{\link[=biocBuildReport]{biocBuildReport()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildPkgDependencyGraph.R
\name{buildPkgDependencyDataFrame}
\alias{buildPkgDependencyDataFrame}
\title{Work with Bioconductor package dependencies}
\usage{
buildPkgDependencyDataFrame(dependencies = c("strong", "most", "all"), ...)
}
\arguments{
\item{dependencies}{\code{character()} a vector listing the types of dependencies, a
subset of c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances").
Character string "all" is shorthand for that vector, character string
"most" for the same vector without "Enhances", character string "strong"
(default) for the first three elements of that vector.}

\item{...}{parameters passed along to \code{\link[=biocPkgList]{biocPkgList()}}}
}
\value{
A \code{data.frame} (also a \code{tbl_df}) of
S3 class "biocDepDF" including columns "Package", "dependency",
and "edgetype".
}
\description{
Bioconductor is built using an extensive set of
core capabilities and data structures. This leads
to package developers depending on other packages
for interoperability and functionality. This
function extracts package dependency information
from \code{\link[=biocPkgList]{biocPkgList()}} and returns a tidy
\code{data.frame} that can be used for analysis
and to build graph structures of package dependencies.
}
\note{
This function requires network access.
}
\examples{
# performs a network call, so must be online.
library(BiocPkgTools)
depdf <- buildPkgDependencyDataFrame()
head(depdf)
library(dplyr)
# filter to include only "Imports" type
# dependencies
imports_only <- depdf |> filter(edgetype=='Imports')

# top ten most imported packages
imports_only |> select(dependency) |>
  group_by(dependency) |> tally() |>
  arrange(desc(n))

# The Bioconductor packages with the
# largest number of imports
largest_importers <- imports_only |>
  select(Package) |>
  group_by(Package) |> tally() |>
  arrange(desc(n))

# not sure what these packages do. Join
# to their descriptions
biocPkgList() |> select(Package, Description) |>
  left_join(largest_importers) |> arrange(desc(n)) |>
  head()
}
\seealso{
See \code{\link[=buildPkgDependencyIgraph]{buildPkgDependencyIgraph()}}, \code{\link[=biocPkgList]{biocPkgList()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_orcid.R
\name{orcid_table}
\alias{orcid_table}
\title{Obtain employment data from ORCID}
\usage{
orcid_table(orcids)
}
\arguments{
\item{orcids}{\code{character()} A vector of ORCID identifiers}
}
\value{
a \code{data.frame} of employment info using the ORCID API
}
\description{
Get a \code{data.frame} of employment info from ORCID
}
\examples{
if (interactive()) {
    orcid_table(
        orcids = c(
            "0000-0002-3242-0582",
            "0000-0003-4046-0063",
            "0000-0003-2725-0694"
        )
    )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocPkgTools.R
\docType{package}
\name{BiocPkgTools}
\alias{BiocPkgTools-package}
\alias{BiocPkgTools}
\title{BiocPkgTools: Examine and analyze Bioconductor package metadata}
\description{
Bioconductor has a rich ecosystem of metadata around
packages, usage, and build status. This package is a simple
collection of functions to access that metadata from R. The goal
is to expose metadata for data mining and value-added functionality
such as package searching, text mining, and analytics on
packages.
}
\section{For developers}{

The \code{\link[=biocBuildReport]{biocBuildReport()}} function returns a computable
form of the Bioconductor Build Report.
}

\section{For users}{

The \code{\link[=biocDownloadStats]{biocDownloadStats()}} function gets Bioconductor
download stats, allowing users to quickly find commonly used
packages. The \code{\link[=biocPkgList]{biocPkgList()}} is useful for getting
a complete listing of all Bioconductor packages.
}

\section{Infrastructure}{

Bioconductor packages all have Digital Object Identifiers (DOIs).
This package contains basic infrastructure for creating, updating,
and de-referencing DOIs.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/seandavi/BiocPkgTools}
  \item Report bugs at \url{https://github.com/seandavi/BiocPkgTools/issues/new}
}

}
\author{
\strong{Maintainer}: Sean Davis \email{seandavi@gmail.com}

Authors:
\itemize{
  \item Shian Su \email{su.s@wehi.edu.au} [contributor]
  \item Marcel Ramos \email{marcel.ramos@sph.cuny.edu} (\href{https://orcid.org/0000-0002-3242-0582}{ORCID}) [contributor]
}

Other contributors:
\itemize{
  \item Lori Shepherd \email{Lori.Shepherd@roswellpark.org} [contributor]
  \item Felix G.M. Ernst \email{felix.gm.ernst@outlook.com} [contributor]
  \item Jennifer Wokaty \email{jennifer.wokaty@gmail.com} [contributor]
  \item Charlotte Soneson \email{charlottesoneson@gmail.com} [contributor]
  \item Martin Morgan \email{martin.morgan@roswellpark.org} [contributor]
  \item Vince Carey \email{stvjc@channing.harvard.edu} [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repositoryStats.R
\name{repositoryStats}
\alias{repositoryStats}
\alias{print.repositoryStats}
\title{Bioconductor Binary Repository Statistics}
\usage{
repositoryStats(
  version = BiocManager::version(),
  binary_repository = BiocManager::containerRepository(version),
  local = FALSE
)

\method{print}{repositoryStats}(x, ...)
}
\arguments{
\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{binary_repository}{\code{character(1)} location of binary repository as
given by \code{BiocManager::containerRepository} (default)}

\item{local}{\code{logical(1)} whether to check the local file system for the
\code{PACKAGES} file's last modified date (default: \code{FALSE}).}

\item{x}{the object returned by \code{repositoryStats()}.}

\item{...}{further arguments passed to or from other methods (not used).}
}
\value{
a list of class \code{repositoryStats} with the following fields:
\itemize{
\item container: \code{character(1)} container label, e.g.,
\code{bioconductor_docker}, or NA if not evaluated on a supported container
\item bioconductor_version: \code{package_version} the
Bioconductor version provided by the user.
\item repository_exists: \code{logical(1)} TRUE if a binary repository
exists for the container and Bioconductor_Version version.
\item bioconductor_binary_repository: \code{character(1)} repository
location, if available, or NA if the repository does not exist.
\item n_software_packages: \code{integer(1)} number of software packages
in the Bioconductor source repository.
\item n_binary_packages: \code{integer(1)} number of binary packages
available. When a binary repository exists, this number is likely
to be larger than the number of source software packages, because
it includes the binary version of the source software packages, as
well as the (possibly CRAN) dependencies of the binary packages
\item n_binary_software_packages: \code{integer(1)} number of binary
packages derived from Bioconductor source packages. This number is
less than or equal to \code{n_software_packages}.
\item missing_binaries: \code{integer(1)} the number of Bioconductor
source software packages that are not present in the binary
repository.
\item out_of_date_binaries: \code{integer(1)} the number of Bioconductor
source software packages that are newer than their binary
counterpart. A newer source software package
might occur when the main Bioconductor build system has
updated a package after the most recent run of the binary
build system.
}
}
\description{
Summarize binary packages compatible with the Bioconductor or
Terra container in use.
}
\details{
For local repositories, use the \code{local = TRUE} argument. Local
repositories will typically start with the \verb{file://} URI. The function
checks the \code{mtime} of the output of \code{file.info} on the \code{PACKAGES} file in
the local repository. Otherwise, by default, it will check the
\code{last-modified} header of the \code{PACKAGES} file via \code{httr::HEAD()}.
}
\section{Methods (by generic)}{
\itemize{
\item \code{print(repositoryStats)}: Print a summary of package
availability in binary repositories.

}}
\examples{
stats <- repositoryStats() # obtain statistics
stats                       # display a summary
stats$container             # access an element for further computation

}
\author{
M. Morgan
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocDownloadStats.R
\name{biocDownloadStats}
\alias{biocDownloadStats}
\title{Get Bioconductor download statistics}
\usage{
biocDownloadStats(
  pkgType = c("software", "data-experiment", "workflows", "data-annotation")
)
}
\arguments{
\item{pkgType}{\code{character()} All, some, or one of 'software',
'data-experiment', 'workflows', or 'data-annotation' (defaults to all
types)}
}
\value{
A \code{tibble} of download statistics for all Bioconductor packages
}
\description{
Get Bioconductor download statistics
}
\details{
Note that Bioconductor package download
stats are not version-specific.
}
\examples{
biocDownloadStats()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignetteTools.R
\name{getBiocVignette}
\alias{getBiocVignette}
\title{Download a Bioconductor vignette}
\usage{
getBiocVignette(
  vignettePath,
  destfile = tempfile(),
  version = BiocManager::version()
)
}
\arguments{
\item{vignettePath}{\code{character(1)} the additional path information to get to
the vignette}

\item{destfile}{\code{character(1)} the file location to store the vignette}

\item{version}{\code{character(1)} such as "3.7", defaults to user version}
}
\value{
\code{character(1)} The filename of the downloaded vignette
}
\description{
The actual vignette path is available using \code{\link[=biocPkgList]{biocPkgList()}}.
}
\examples{
x = biocPkgList()
tmp = getBiocVignette(x$vignettes[[1]][1])
tmp

\dontrun{
library(pdftools)
y = pdf_text(tmp)
y = paste(y,collapse=" ")
library(tm)
v = VCorpus(VectorSource(y))

v <- v |>
    tm_map(stripWhitespace) |>
    tm_map(content_transformer(tolower)) |>
    tm_map(removeWords, stopwords("english")) |>
    tm_map(stemDocument)
dtm = DocumentTermMatrix(v)
inspect(DocumentTermMatrix(v,
    list(dictionary = as.character(x$Package))))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/newBiocPkgDOI.R
\name{generateBiocPkgDOI}
\alias{generateBiocPkgDOI}
\title{Generate a DOI for a Bioconductor package}
\usage{
generateBiocPkgDOI(pkg, authors, pubyear, event = "publish", testing = TRUE)
}
\arguments{
\item{pkg}{\code{character(1)} package name}

\item{authors}{\code{character()} vector of authors (will be "pasted" together)}

\item{pubyear}{\code{integer(1)} publication year}

\item{event}{Either "hide", "register", or publish". Typically, we use
"publish" to make the DOI findable.}

\item{testing}{\code{logical(1)} If true, will use the apitest
user with the password apitest. These DOIs will expire.
The same apitest:apitest combination can be used to
login to the website for doing things using the
web interface. If false, the Bioconductor-specific
user credentials should be in the correct environment
variables}
}
\value{
The DOI as a \code{character(1)} vector.
}
\description{
This function makes calls out to the DataCite REST API described
here: \url{https://support.datacite.org/docs/api-create-dois}. The
function creates a new DOI for a Bioconductor package (cannot already
exist). The target URL for the DOI is the short Bioconductor
package URL.
}
\details{
The login information for the "real" Bioconductor account
should be stored in the environment variables "DATACITE_USERNAME"
and "DATACITE_PASSWORD

The GUI is available here: \url{https://doi.datacite.org/}.
}
\examples{
\dontrun{
  x = generateBiocPkgDOI('RANDOM_TEST_PACKAGE','Sean Davis',1972)
}
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocVIEWSdb.R
\name{biocVIEWSdb}
\alias{biocVIEWSdb}
\title{Parse and return the VIEWS file for a Bioconductor Release}
\usage{
biocVIEWSdb(
  version = BiocManager::version(),
  pkgType = c("software", "data-experiment", "data-annotation", "workflows")
)
}
\arguments{
\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{pkgType}{\code{character()} The type of packages for which to get
information for. Values can be one or more of the following:
\itemize{
\item \code{software}: Software packages
\item \code{data-experiment}: Experiment data packages
\item \code{data-annotation}: Annotation data packages
\item \code{workflows}: Workflow packages
}}
}
\description{
The function parses and returns the \code{VIEWS} file for a specified Bioconductor
version, either "release" or "devel". The \code{VIEWS} file contains metadata
about Bioconductor packages, including information about their categories,
topics, and other details.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
biocVIEWSdb(pkgType = "software")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocDownloadStats.R
\name{pkgDownloadStats}
\alias{pkgDownloadStats}
\title{Get Bioconductor download statistics for a package}
\usage{
pkgDownloadStats(
  pkg,
  pkgType = c("software", "data-experiment", "workflows", "data-annotation"),
  years = format(Sys.time(), "\%Y")
)
}
\arguments{
\item{pkg}{\code{character(1)} The name of a Bioconductor package}

\item{pkgType}{\code{character(1)} One of 'software', 'data-experiment',
'workflows', or 'data-annotation' (defaults to 'software')}

\item{years}{numeric(), \code{character()} A vector of years from which to
obtain download statistics (defaults to current year)}
}
\value{
A \code{tibble} of download statistics
}
\description{
Get Bioconductor download statistics for a package
}
\examples{
pkgDownloadStats("GenomicRanges")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocPkgList.R
\name{biocPkgList}
\alias{biocPkgList}
\title{Get full Bioconductor software package listing, with details}
\usage{
biocPkgList(
  version = BiocManager::version(),
  repo = c("BioCsoft", "BioCexp", "BioCworkflows", "BioCann", "CRAN"),
  addBiocViewParents = TRUE
)
}
\arguments{
\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{repo}{\code{character(1)} The requested Bioconductor repository. The default
is to pull from the "BioCsoft" repository. Possible repositories include
"BioCsoft", "BioCexp", "BioCworkflows", "BioCann", and "CRAN". Note that
not all repos are available for all versions, particularly older versions.}

\item{addBiocViewParents}{\code{logical(1)} whether to add all biocViews
parents to biocViews annotations.}
}
\value{
An object of class \code{tbl_df} with one row per package. The result
always includes a \code{fnd} list-column whose elements are character vectors
of funder names extracted from persons with role \code{"fnd"} in the
formatted \code{Author} field. Elements are \code{NA_character_} for packages
that declare no funder (or whose \code{Author} field contains no \verb{[fnd]}
role tag).
}
\description{
The BiocViews-generated \code{VIEWS} file is available
for Bioconductor release and devel repositories. It
contains quite a bit more information from the
package \code{DESCRIPTION} files than the \code{PACKAGES}
file. In particular, it contains \code{biocViews} annotations
and URLs for vignettes and developer URLs.
}
\details{
Since packages are annotated with the most specific
views, the default functionality here is to add parent terms
for all views for each package. For example, in the bioCsoft
repository, all packages will have at least "Software" added
to their biocViews. If one wants to stick to only the most
specific terms, set \code{addBiocViewParents} to \code{FALSE}.
}
\examples{
bpkgl <- biocPkgList(repo = "BioCsoft")
bpkgl
unlist(bpkgl[1,'Depends'], use.names = FALSE)

# Get a list of all packages that
# import "GEOquery"
library(dplyr)
bpkgl |>
  filter(Package == 'GEOquery') |>
  pull('importsMe') |>
  unlist()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classDependencies.R
\name{class-dependencies}
\alias{class-dependencies}
\alias{buildClassDepGraph}
\alias{buildClassDepData}
\alias{buildClassDepFromPackage}
\alias{plotClassDep}
\alias{plotClassDepData}
\alias{plotClassDepGraph}
\title{Retrieve Class relationships}
\usage{
buildClassDepGraph(class, includeUnions = FALSE)

buildClassDepData(class, includeUnions = FALSE)

buildClassDepFromPackage(pkg, includeUnions = FALSE)

plotClassDep(class, includeUnions = FALSE)

plotClassDepData(data)

plotClassDepGraph(g)
}
\arguments{
\item{class}{a single \code{character} value defining a \sQuote{S4} class
name}

\item{includeUnions}{\code{TRUE} or \code{FALSE}: Should union definitions
included in the result? (default: \code{FALSE})}

\item{pkg}{a single \code{character} value defining a package name}

\item{data}{a \code{data.frame} with compatible columns. See output of
\code{buildClassDepData}}

\item{g}{an \code{igraph} object with compatible edge attributes. See output
of \code{buildClassDepGraph}}
}
\description{
As the title says it should do something with class relationships
}
\examples{
library("SummarizedExperiment")
depData <- buildClassDepData("RangedSummarizedExperiment")
depData
g <- buildClassDepGraph("RangedSummarizedExperiment")
plotClassDepGraph(g)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDependencyMetrics.R
\name{pkgCombDependencyGain}
\alias{pkgCombDependencyGain}
\title{Calculate dependency gain achieved by excluding combinations of packages}
\usage{
pkgCombDependencyGain(pkg, depdf, maxNbr = 3L)
}
\arguments{
\item{pkg}{character, the name of the package for which we want
to estimate the dependency gain}

\item{depdf}{a tidy data frame with package dependency information
obtained through the function \code{\link[=buildPkgDependencyDataFrame]{buildPkgDependencyDataFrame()}}}

\item{maxNbr}{numeric, the maximal number of direct dependencies to leave
out simultaneously}
}
\value{
A data frame with three columns: ExclPackages (the excluded direct
dependencies), NbrExcl (the number of excluded direct dependencies),
DepGain (the dependency gain from excluding these direct dependencies)
}
\description{
Calculate dependency gain achieved by excluding combinations of packages
}
\examples{
depdf <- buildPkgDependencyDataFrame(
  dependencies=c("Depends", "Imports"),
  repo=c("BioCsoft", "CRAN")
)
pcd <- pkgCombDependencyGain('GEOquery', depdf, maxNbr = 3L)
head(pcd[order(pcd$DepGain, decreasing = TRUE), ])

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocMaintained.R
\name{biocMaintained}
\alias{biocMaintained}
\alias{hasBiocMaint}
\title{Bioconductor Maintained Packages}
\usage{
biocMaintained(
  main = "maintainer@bioconductor.org",
  version = BiocManager::version(),
  pkgType = c("software", "data-experiment", "workflows", "data-annotation")
)

hasBiocMaint(
  pkg,
  version = BiocManager::version(),
  main = "maintainer@bioconductor\\\\.org",
  repo = c("BioCsoft", "BioCexp", "BioCworkflows", "BioCann")
)
}
\arguments{
\item{main}{\code{character(1)} A regex string to search for in the \code{Maintainer}
column from the \code{biocPkgList()} output.}

\item{version}{\code{character(1)} or \code{package_version} indicating the
\emph{Bioconductor} version (e.g., "3.22"). Defaults to the "current version"
provided by \code{BiocManager::version()}. The keywords "release" and "devel"
are also accepted.}

\item{pkgType}{\code{character()} The type of packages for which to get
information for. Values can be one or more of the following:
\itemize{
\item \code{software}: Software packages
\item \code{data-experiment}: Experiment data packages
\item \code{data-annotation}: Annotation data packages
\item \code{workflows}: Workflow packages
}}

\item{pkg}{\code{character(1)} A vector of package names (case sensitive).}

\item{repo}{\code{character()} A vector of Bioconductor repositories to search
through. By default, it will search through all Bioconductor repositories.}
}
\value{
For \code{biocMaintained}: a \code{tibble} of packages associated with the
maintainer.

For \code{hasBiocMaint}: a logical vector indicating whether the
package is maintained by Bioconductor.
}
\description{
List all the packages associated with a maintainer. By default, it will
return all packages associated with the \code{maintainer@bioconductor.org} email.
\code{hasBiocMaint} returns a logical vector corresponding to the input character
vector of packages indicating whether any package is maintained by the
Bioconductor core team.
}
\examples{

biocMaintained()

## maintained by Hervé and not maintainer at bioconductor dot org
hasBiocMaint("BiocGenerics")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPackageInfo.R
\name{getPackageInfo}
\alias{getPackageInfo}
\title{Generate needed information to create DOI from a package
directory.}
\usage{
getPackageInfo(dir)
}
\arguments{
\item{dir}{\code{character(1)} Path to package}
}
\value{
A \code{data.frame}
}
\description{
Generate needed information to create DOI from a package
directory.
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getData.R
\name{get_bioc_data}
\alias{get_bioc_data}
\title{Get data from Bioconductor}
\usage{
get_bioc_data()
}
\value{
A JSON string containing Bioconductor package details
}
\description{
Get data from Bioconductor
}
\examples{
bioc_data <- get_bioc_data()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocExplore.R
\name{biocExplore}
\alias{biocExplore}
\title{Explore Bioconductor packages interactively}
\usage{
biocExplore(top = 500L, ...)
}
\arguments{
\item{top}{maximum number of packages displayed in any biocView}

\item{...}{parameters passed to \code{htmlwidgets::createWidget()}}
}
\value{
A bubble plot of Bioconductor packages
}
\description{
Explore Bioconductor packages through an interactive bubble
plot. Click on bubbles to bring up additional information about the
package. Size and proximity to center of a bubble is based on the
downloads the package has in the past month.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildPkgDependencyGraph.R
\name{buildPkgDependencyIgraph}
\alias{buildPkgDependencyIgraph}
\title{Work with package dependencies as a graph}
\usage{
buildPkgDependencyIgraph(pkgDepDF)
}
\arguments{
\item{pkgDepDF}{a tidy data frame. See description for
details.}
}
\value{
An igraph directed graph. See the igraph
package for details of what can be done.
}
\description{
Package dependencies represent a directed
graph (though Bioconductor dependencies are
not an acyclic graph). This function simply
returns an igraph graph from the package
dependency data frame from a call to
\code{\link[=buildPkgDependencyDataFrame]{buildPkgDependencyDataFrame()}} or
any tidy data frame with rows of (Package, dependency)
pairs. Additional columns are added as igraph edge
attributes (see \code{\link[igraph:graph_from_data_frame]{igraph::graph_from_data_frame()}}).
}
\examples{

library(igraph)

pkg_dep_df = buildPkgDependencyDataFrame()

# at this point, filter or join to manipulate
# dependency data frame as you see fit.

g = buildPkgDependencyIgraph(pkg_dep_df)
g

# Look at nodes and edges
head(V(g)) # vertices
head(E(g)) # edges

# subset graph by attributes


head(sort(degree(g, mode='in'), decreasing=TRUE))
head(sort(degree(g, mode='out'), decreasing=TRUE))

}
\seealso{
See \code{\link[=buildPkgDependencyDataFrame]{buildPkgDependencyDataFrame()}},
\code{\link[igraph:graph_from_data_frame]{igraph::graph_from_data_frame()}},
\code{\link[=inducedSubgraphByPkgs]{inducedSubgraphByPkgs()}}, \code{\link[=subgraphByDegree]{subgraphByDegree()}},
\link[igraph:igraph-es-indexing]{igraph::igraph-es-indexing},
\link[igraph:igraph-vs-indexing]{igraph::igraph-vs-indexing}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biocBuildEmail.R
\name{biocBuildEmail}
\alias{biocBuildEmail}
\alias{sentHistory}
\title{Create and copy e-mail package notification template to clipboard}
\usage{
biocBuildEmail(
  pkg,
  version = c("release", "devel"),
  PS = character(1L),
  dry.run = TRUE,
  to = NULL,
  cc = NULL,
  bcc = NULL,
  emailTemplate = templatePath(),
  core.name = NULL,
  core.email = NULL,
  core.id = NULL,
  textOnly = FALSE,
  resend = FALSE,
  verbose = FALSE,
  credFile = "~/.blastula_creds"
)

sentHistory()
}
\arguments{
\item{pkg}{\code{character(1)} The name of the package in trouble}

\item{version}{\code{character()} A vector indicating which version of Bioconductor
the package is failing in (either 'release' or 'devel'; defaults to both)}

\item{PS}{\code{character(1)} Postscript, an additional note to the recipient of
the email (i.e., the package maintainer)}

\item{dry.run}{\code{logical(1)} Display the email without sending to the recipient.
It only works for HTML email reports and ignored when \code{textOnly=TRUE}}

\item{to}{\code{character()} A vector of email addresses serving as primary
recipients for the message. For secondary recipients, use the \code{cc} and
\code{bcc} arguments.}

\item{cc, bcc}{\code{character()} A vector of email addresses for sending the message
as a carbon copy or blind carbon copy.}

\item{emailTemplate}{\code{character(1)} The path to the email template Rmd file as
obtained by \code{templatePath()}. A custom template can be provided as file
path.}

\item{core.name}{\code{character(1)} The full name of the core team member}

\item{core.email}{\code{character(1)} The Roswell Park email of the core team
member}

\item{core.id}{\code{character(1)} The internal identifier for the Roswell employee.
This ID usually matches \verb{^[A-Z]\{2\}[0-9]\{5\}} for more recent identifiers.}

\item{textOnly}{\code{logical(1)} Whether to return the text of the email only.
This avoids the use of the 'blastula' package and adds the text to the
system clipboard if the \code{clipr} package is installed (default: FALSE)}

\item{resend}{\code{logical(1)} Whether to force a resend of the email}

\item{verbose}{\code{logical(1)} Whether to output full email information from
'smtp_send' (when \code{dry.run} is \code{FALSE} and 'blastula' is installed)}

\item{credFile}{\code{character(1)} An optional file generated by the
\code{blastula::create_smtp_creds_file} function containing email
authentication information (default: "~/.blastula_creds"). See
\code{?biocBuildEmail} details.}
}
\value{
A character string of the email
}
\description{
The \code{biocBuildEmail} function provides a template for notifying
maintainers of errors in the Bioconductor Build System (BBS). This
convenience function returns the body of the email from a template within the
package and provides a copy in the clipboard.
}
\details{
The \code{credFile} argument is a convenience for avoiding password entry
at every instance an email is sent. If the default file
\verb{~/.blastula_creds} does not exist, the user will be prompted for
authorization information. Currently it is configured to emails for the
core-team:
\preformatted{
        blastula::create_smtp_creds_file(
            file = "~/.blastula_creds",
            user = "user.email@domain.org",
            host = "smtp.office365.com",
            port = 587,
            use_ssl = TRUE
        )
    }
}
\section{sentHistory}{
 Check the history of emails sent
}
